/* $Id: ddl_dense.c,v 1.7 1995/08/25 09:47:46 tim Exp $ */

/* Copyright 1994 IASC */

/* dense DDL routines
   This file contains routines for the dense distributed vector DDL_Vec */

/* Procedures and parameter lists:
   The DDL is continually evolving and this often results in new procedures
   or changes to parameter lists, and sometimes the removal of procedures.
   Each procedure is marked to indicate its longevity and use:
     1-4: indicates longevity of procedure interface 
        1: won't change (ever, honest!)
	2: unlikely to change, but possible
	3: might well be changed
	4: very likely to be changed or removed 
     HL: indicates expected use of procedure
        H: high level procedure, probably documented in written docs,
	   for user programs and library procedures
	L: low level procedure, not documented in written docs,
           for use in Library procedures mainly
   eg. DDL_Create_sparse: H1.
   High level procedures are usually unlikely to change. 
   Parameters are documented:
     IN:      input only parameter; is not changed by procedure
     OUT:     output parameter; any value on entry; changed on exit
     INOUT:   both input and output; read on input; may be changed on exit */

#include <stdio.h>
#include <mpi.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include "ddl.h"

/* low level routine to create DDLR_Vector structure called by
   DDL_Create_vector or after DDLR_Prop is created by property func.
   Parameters:
     size: IN       global size of vector
     type: IN       type of individual elements using MPI datatypes
     block: IN      smallest block size for distribution; see docs
     comm: IN       MPI communicator for object gives processes to use
     vec: OUT       pointer to DDL_Vector structure
   Status: L3
   might introduce format parameter in future similar to DDL_Create_sparse */
int DDLL_Create_vector(int size, MPI_Datatype type,
		       int block, MPI_Comm comm, DDL_Vector *vec)
{
  int ierr = DDL_SUCCESS;
  DDL_Image im;
  DDL_Vector v;
  int dim[2], nproc;

  MPI_Comm_size(comm, &nproc);
  if (size % block != 0) {
    ierr = DDL_Error(DDL_ERR_BLOCK,DDL_ERR_HARD,DDL_CREATE_VECTOR);
    if (ierr != DDL_SUCCESS) return ierr;
  }
  if (DDL_Find_proc(size, block, nproc) != nproc) {
    ierr = DDL_Error(DDL_ERR_DIST, DDL_ERR_SOFT, DDL_CREATE_VECTOR);
    if (ierr != DDL_SUCCESS) return ierr;
  }
  
  v = malloc(sizeof(struct DDLR_Vector));

  dim[0] = size;
  dim[1] = 1;
  ierr = DDL_Image_create(dim, type, 0, block, comm, &im);
  v->image = DDL_Image_find(im);
  v->format = DDL_VEC_BLK;
  v->block = block;
  *vec = v;
  return ierr;
}

/* high level user routine to create dense distributed vector
   Parameters:
     size: IN       global size of vector
     type: IN       type of individual elements using MPI datatypes
     block: IN      smallest block size for distribution; see docs
     comm: IN       MPI communicator for object gives processes to use
     vec: OUT       handle on created vector
   Status: H2
   might introduce format parameter in future similar to DDL_Create_sparse */
int DDL_Create_vector(int size, MPI_Datatype type,
		      int block, MPI_Comm comm, DDL_Object *vec)
{
  int ierr = DDL_SUCCESS;
  int dim[2], nproc;
  DDL_Vector v;
  DDL_Prop pr;

  MPI_Comm_size(comm, &nproc);
  if (size % block != 0) {
    ierr = DDL_Error(DDL_ERR_BLOCK,DDL_ERR_HARD,DDL_CREATE_VECTOR);
    if (ierr != DDL_SUCCESS) return ierr;
  }
  if (DDL_Find_proc(size, block, nproc) != nproc) {
    ierr = DDL_Error(DDL_ERR_DIST, DDL_ERR_SOFT, DDL_CREATE_VECTOR);
    if (ierr != DDL_SUCCESS) return ierr;
  }

  /* allocate space for DDL_Prop */
  pr = malloc(sizeof(struct DDLR_Prop));
  if (pr == NULL) {
    ierr = DDL_Error(DDL_ERR_MEMORY, DDL_ERR_HARD, DDL_CREATE_VECTOR);
    if (ierr > 0) return ierr;} /* cleanup and exit with error */
  pr->type = DDL_PROP_VECTOR;
  pr->tag = DDL_VECTOR; /* high level sparse object */
  pr->prop = NULL;
  
  /* create DDLR_Vector structure */
  ierr = DDLL_Create_vector(size, type, block, comm, &v);
  if (ierr != DDL_SUCCESS) {
    free(pr);
    return ierr; /* cleanup and exit with error */
  }
  pr->data = v;

  *vec = pr;
  return ierr;
}

/* free object memory
   Parameters:
     vec: INOUT     object to delete; NULL on exit
   Status: H1 */
int DDL_Free_vector(DDL_Object vec)
{
  int ierr = DDL_SUCCESS;
  DDL_Vector vecr;
  DDL_Image im;
  vecr = (DDL_Vector)(vec->data);
  im = DDL_Vectortoimage(vecr);
  ierr = DDL_Image_free(im);
  ierr = DDL_Free_allprop(vec);
  free(vecr);
  free(vec);
  return ierr;
}

/* get pointer to raw data values for any property
   so user can update elements of vector or sparse matrix
   Parameters:
     obj: IN        object to manipulate
     ptr: OUT       on exit points to raw data
   Status: H1 */
int DDL_Get_vector(DDL_Object obj, void *ptr)
{ 
  int ierr = DDL_SUCCESS;
  if (obj == NULL) {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_GET_VECTOR);
    return ierr;
  }
  switch (obj->type) {
  case DDL_PROP_VECTOR:
    *(void**)ptr = DDL_Raw_vector(((DDL_Vector)(obj->data))); 
    return ierr;
  default:
    ierr = DDL_Error(DDL_ERR_DDLTYPE, DDL_ERR_HARD, DDL_GET_VECTOR);
    return ierr;
  }
} 

/* get MPI communicator for object
   Parameters:
     obj: IN        object to manipulate
     comm: OUT      MPI communicator
   Status: H1 */
int DDL_Comm_vector(DDL_Object vec, MPI_Comm *comm)
{
  int ierr;
  DDL_Image im;
  im = DDL_Vectortoimage(((DDL_Vector)(vec->data)));
  ierr = DDL_Image_comm(im, comm);
  return ierr;
}

/* get format of vector object; not used since only one format
   is currently supported
   Parameters:
     vec: IN        object to manipulate
     format: OUT    format
   Status: H3 */
int DDL_Format_vector(DDL_Object vec, int *format)
{
  int ierr = DDL_SUCCESS;
  DDL_Image im;
  *format = ((DDL_Vector)(vec->data))->format;
  return ierr;
}

/* get data type of individual elements of object
   Parameters:
     vec: IN        object to manipulate
     type: OUT      data type
   Status: H1 */
int DDL_Type_vector(DDL_Object vec, MPI_Datatype *type)
{
  int ierr;
  DDL_Image im;
  im = DDL_Vectortoimage(((DDL_Vector)(vec->data)));
  ierr = DDL_Image_type(im, type);
  return ierr;
}

/* get global size of vector
   Parameters:
     vec: IN        object to manipulate
     size: OUT      global size
   Status: H1 */
int DDL_Gsize_vector(DDL_Object vec, int *size)
{
  int ierr;
  int dim[2];
  DDL_Image im;
  im = DDL_Vectortoimage(((DDL_Vector)(vec->data)));
  ierr = DDL_Image_global_size(im, dim);
  *size = dim[0];
  return ierr;
}

/* get local size of vector; number of elements on calling process
   Parameters:
     vec: IN        object to manipulate
     size: OUT      local size
   Status: H1 */
int DDL_Size_vector(DDL_Object vec, int *size)
{
  int ierr;
  int dim[2];
  DDL_Image im;
  im = DDL_Vectortoimage(((DDL_Vector)(vec->data)));
  ierr = DDL_Image_local_size(im, dim);
  *size = dim[0];
  return ierr;
}

/* get offset of local block of vector
   Parameters:
     vec: IN        object to manipulate
     offset: OUT    local offset
   Status: H1 */
int DDL_Offset_vector(DDL_Object vec, int *offset)
{
  int ierr;
  int off[2];
  DDL_Image im;
  im = DDL_Vectortoimage(((DDL_Vector)(vec->data)));
  ierr = DDL_Image_offset(im, off);
  *offset = off[0];
  return ierr;
}

/* get block size of vector
   Parameters:
     vec: IN        object to manipulate
     block: OUT     block size used for distribution
   Status: H1 */
int DDL_Block_vector(DDL_Object vec, int *block)
{
  int ierr = DDL_SUCCESS;
  if (vec != NULL) {
    *block = ((DDL_Vector)(vec->data))->block;
    return ierr;}
  else {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_BLOCK_VECTOR);
    return ierr;}
}

/* dense vector I/O for DDL

   data file format uses ascii text, so is not efficient

   format supports vectors and sparse matrices currently

   each object in file has a header which identifies data to follow.
   for vector:
     format: INTEGER  specifies whether object is vector or sparse matrix  

     n: INTEGER       global size of vector
     type: INTEGER    data type of individual elements of vector in memory
                      supports REAL, DOUBLE, INT types

   After the header the data values follow 1 element per line.
   (see include file ddl_dense.h for constant values defining types and
   formats) */

/* write distributed vector object to file 
   Parameters:
     vec: IN        object to output
     file: INOUT    file handle
   Status: H1 */
int DDL_Write_vector(DDL_Object vec, FILE *file)
{
  int ierr;
  int fformat, ftype;
  MPI_Aint typesize;
  int n, nproc, proc, i, j, count;
  MPI_Comm comm;
  MPI_Datatype type;
  MPI_Status status;
  void *data;
  void *tmp;
  int own[2], *owned;
  
  /* get object details */
  ierr = DDL_SUCCESS;
  ierr = DDL_Type_vector(vec, &type);
  ierr = DDL_Gsize_vector(vec, &n);

  fformat = DDL_VEC_FMT;

  /* get communication info */
  DDL_Comm_vector(vec, &comm);
  MPI_Comm_size(comm, &nproc);
  MPI_Comm_rank(comm, &proc);

  if (DDL_Ishost())
    if (file == NULL) {
      ierr = DDL_Error(DDL_ERR_FILE, DDL_ERR_HARD, DDL_WRITE_VECTOR);
    }

  /* tell all nodes whether to continue */
  MPI_Bcast(&ierr, 1, MPI_INT, 0, comm);

  /* exit if error */
  if (ierr != DDL_SUCCESS) return ierr;

  /* OK, we can continue */

  /* decide which elements this node owns */
  DDL_Offset_vector(vec, own);
  DDL_Size_vector(vec, &own[1]);

  /* find size of element */
  DDL_DDLtype(type, &ftype);
  MPI_Type_size(type, &typesize);
  
  if (DDL_Ishost()) {
    owned = malloc(sizeof(int)*2*nproc);
    /* allocate temporary space for vector */
    tmp = malloc(DDL_COMM_BLKSIZE);
  }
  /* get the information about which bit of the object each node owns */
  MPI_Gather(own, 2, MPI_INT, owned, 2, MPI_INT, 0, comm);

  /* get pointers to raw vector data */
  ierr = DDL_Get_vector(vec, &data);

  if (DDL_Ishost()) {
    /* Write out the information heading */
    fprintf(file, "%10d\n", fformat);
    fprintf(file, "%10d%10d\n", n, ftype);
  }

  for (i=0;i<nproc;i++) {
    if (proc == i) {
      if (DDL_Ishost()) {
	/* just output directly */
	count = own[1];
	ierr = DDL_Write_vectorblk(file, data, type, &count);
      }
      else {
	/* send block to host */
	j = 0;
	while (j<own[1]) {
	  /* send data */
	  count = DDL_MIN(DDL_COMM_BLKSIZE/typesize, own[1]-j);
	  ierr = MPI_Send(&count, 1, MPI_INT, 0, 0, comm);
	  ierr = MPI_Send((char *)data+j*typesize, count, type, 0, 0, comm);
	  j = j + count;
	}
      }
    }
    else if (DDL_Ishost()) {
      j = 0;
      while (j<owned[2*i+1]) {
	/* receive data */
	ierr = MPI_Recv(&count, 1, MPI_INT, i, 0, comm, &status);
	ierr = MPI_Recv(tmp, count, type, i, 0, comm, &status);
	/* output block to file */
	ierr = DDL_Write_vectorblk(file, tmp, type, &count);
	j = j + count;
      }
    }
  }
  if (DDL_Ishost()) {
    free(owned);
    free(tmp);
  }
  return ierr;
}

/* write block of vector data to file; called from DDL_Write_vector
   probably no use anywhere else
   Parameters:
     file: INOUT    file handle
     data: IN       pointer to data values
     type: IN       data type
     count: IN      number of values to write
   Status: L2 */
int DDL_Write_vectorblk(FILE *file, void *data, MPI_Datatype type, int *count)
{
  int i, ierr;
  if (type == MPI_FLOAT) {
    for (i=0;i<*count;i++)
      fprintf(file, " %16.7e\n", ((float *)data)[i]);
  }
  else if (type == MPI_DOUBLE) {
    for (i=0;i<*count;i++)
      ierr = fprintf(file, " %16.7le\n", ((double *)data)[i]);
  }
  else if (type == MPI_INT) {
    for (i=0;i<*count;i++)
      fprintf(file, " %10d\n", ((int *)data)[i]);
  }

  return DDL_SUCCESS;
}

/* read distributed vector object from file 
   Parameters:
     vec: OUT       object to fill with values from file
     file: INOUT    file handle
   Status: H1 */
int DDL_Read_vector(DDL_Object vec, FILE *file)
{
  int ierr;
  int ftype, format, fformat;
  MPI_Aint typesize;
  int n, nproc, proc, i, j, count;
  MPI_Comm comm;
  MPI_Datatype type;
  MPI_Status status;
  void *data;
  void *tmp;
  int own[2], *owned;
  int fn;

  format = DDL_VEC_BLK;
  ierr = DDL_Type_vector(vec, &type);
  ierr = DDL_Gsize_vector(vec, &n);

  /* check file parameters */
  ierr = DDL_SUCCESS;
  if (DDL_Ishost()) {
    if (file == NULL) {
      ierr = DDL_Error(DDL_ERR_FILE, DDL_ERR_HARD, DDL_READ_VECTOR);
    }
    /* check file contains vector in correct format */
    else {
      fscanf(file, "%10d\n", &fformat);
      if (fformat != DDL_VEC_FMT) {
	ierr = DDL_Error(DDL_ERR_FILEFORMAT, DDL_ERR_HARD, DDL_READ_VECTOR);
      }
      else {
	/* get vector details from file */
	fscanf(file, "%10d%10d\n", &fn, &ftype);
	/* compare with object */
	if (n < fn) {
	  ierr = DDL_Error(DDL_ERR_TOOSMALL, DDL_ERR_HARD, DDL_READ_VECTOR);
	}
      }
    }
  }
 
  /* get communication info */
  DDL_Comm_vector(vec, &comm);
  MPI_Comm_size(comm, &nproc);
  MPI_Comm_rank(comm, &proc);

  /* tell all nodes whether to continue */
  MPI_Bcast(&ierr, 1, MPI_INT, 0, comm);

  /* exit if error */
  if (ierr != DDL_SUCCESS) return ierr;

  /* OK, we can continue */

  /* decide which elements this node owns */
  DDL_Offset_vector(vec, own);
  DDL_Size_vector(vec, &own[1]);

  /* find size of element */
  MPI_Type_size(type, &typesize);

  if (DDL_Ishost()) {
    owned = malloc(sizeof(int)*2*nproc);
    /* allocate temporary space for vector */
    tmp = malloc(DDL_COMM_BLKSIZE);
  }
  /* get the information about which bit of the object each node owns */
  MPI_Gather(own, 2, MPI_INT, owned, 2, MPI_INT, 0, comm);

  /* get pointer to raw vector data */
  ierr = DDL_Get_vector(vec, &data);

  for (i=0;i<nproc;i++) {
    if (proc == i) {
      if (DDL_Ishost()) {
	/* just input directly */
	count = own[1];
	ierr = DDL_Read_vectorblk(file, data, type, &count);
      }
      else {
	/* receive block from host */
	j = 0;
	while (j<own[1]) {
	  /* receive data */
	  ierr = MPI_Recv(&count, 1, MPI_INT, 0, 0, comm, &status);
	  ierr = MPI_Recv((char *)data+j*typesize, count, type, 0, 0, comm, &status);
	  j = j + count;
	}
      }
    }
    else if (DDL_Ishost()) {
      j = 0;
      while (j<owned[2*i+1]) {
	/* read block from file */
	count = DDL_MIN(DDL_COMM_BLKSIZE/typesize, owned[2*i+1]-j);
	ierr = DDL_Read_vectorblk(file, tmp, type, &count);
	/* send data */
	ierr = MPI_Send(&count, 1, MPI_INT, i, 0, comm);
	ierr = MPI_Send(tmp, count, type, i, 0, comm);
	j = j + count;
      }
    }
  }
  if (DDL_Ishost()) {
    free(owned);
    free(tmp);
  }
  return DDL_SUCCESS;
}

/* read block of vector data to file; called from DDL_Read_vector
   probably no use anywhere else
   Parameters:
     file: INOUT    file handle
     data: OUT      pointer to buffer for data values
     type: IN       data type
     count: IN      number of values to read
   Status: L2 */
int DDL_Read_vectorblk(FILE *file, void *data, MPI_Datatype type, int *count)
{
  int i;
  int ierr;

  if (type == MPI_FLOAT) {
    for (i=0;i<*count;i++)
      fscanf(file, " %e", &(((float *)data)[i]));
  }
  else if (type == MPI_DOUBLE) {
    for (i=0;i<*count;i++)
      fscanf(file, "%le", &(((double *)data)[i]));
  }
  else if (type == MPI_INT) {
    for (i=0;i<*count;i++)
      fscanf(file, " %d", &(((int *)data)[i]));
  }
  return DDL_SUCCESS;
}

/* dot: level 1 BLAS routine; 
   single and double precision 
   result res is assumed to be of type of vectors 
   return result to all nodes, not just root,
   this fits in better with the hostless programming model,
   although it does incur communication overheads.

   AVOID TYPE SPECIFIC CODE WHEREVER POSSIBLE
   The following two techniques can help with this.

   We use a pointer to local temporary **scalar** results to try to
   avoid duplicating code for different element types (ie. float & double).
   Instead of using temporary variables directly in statements,
   use the pointer to the variable, coerced suitably from void, 
   if necessary. 
   In this example routine, the code is actually longer than the version
   using variables directly. Hopefully more complicated routines will
   benefit from this technique.

   Using void pointers for **vectors** of data of different type is a 
   definite win and should be used where possible.

   Pass all paramters to FORTRAN BLAS routines as *pointers*. 

   supports REAL and DOUBLE 
   Parameters:
     x: IN          first vector object
     y: in          second vector object
     dot: OUT       on exit, global dot product
   Status: H1 */
int DDL_Dot(DDL_Object x, DDL_Object y, void *dot)
{
  int ierr; /* return value indicating error status of routine */
  int ret; /* special return value for comparing communicators */
  int one = 1;
  int x_size, x_itype;
  MPI_Comm x_comm;
  MPI_Datatype x_type;
  void *x_data;
  int y_size, y_itype;
  MPI_Comm y_comm;
  MPI_Datatype y_type;
  void *y_data;
  float flt;     /* temporary variable for float results */
  double dbl;    /* temporary variable for double results */
  void *ptr;     /* pointer to temporary variable */

  ierr = DDL_SUCCESS;

  /* check input parameters */
  ierr = DDL_Size_vector(x, &x_size);
  ierr = DDL_Size_vector(y, &y_size);
  ierr = DDL_Type_vector(x, &x_type);
  ierr = DDL_Type_vector(y, &y_type);
  ierr = DDL_Comm_vector(x, &x_comm);
  ierr = DDL_Comm_vector(y, &y_comm);

  if (x_type != y_type) {
    ierr = DDL_Error(DDL_ERR_ITYPE, DDL_ERR_HARD, DDL_DOT);
    return ierr;}
  if (x_size != y_size) {
    ierr = DDL_Error(DDL_ERR_ISIZE, DDL_ERR_HARD, DDL_DOT);
    return ierr;}
  /* check communicators are **identical**
     same number of processes and ranking of processes */
  ierr = MPI_Comm_compare(x_comm, y_comm, &ret);
  if (ret != MPI_IDENT) {
    ierr = DDL_Error(DDL_ERR_ICOMM, DDL_ERR_HARD, DDL_DOT);
    return ierr;}
  
  /* get pointers to raw data values */
  ierr = DDL_Get_vector(x, &x_data);
  ierr = DDL_Get_vector(y, &y_data);
  
  /* local dot product calling FORTRAN blas routine */
  if (x_type == MPI_FLOAT) {
    ptr = &flt;
    flt = F77(sdot)(&x_size, x_data, &one, y_data, &one);
  }
  else if (x_type == MPI_DOUBLE) {
    ptr = &dbl;
    dbl = F77(ddot)(&x_size, x_data, &one, y_data, &one);
  }

  /* call MPI collective */
  ierr = MPI_Allreduce(ptr, dot, 1, x_type, MPI_SUM, x_comm);
  return ierr;
}   
    
/* scal: level 1 BLAS routine to scale elements of vector
   supports REAL and DOUBLE 
   Parameters:
     x: INOUT       vector object
     alpha: IN      scaling value
   Status: H2 
   might reorder parameters? */
int DDL_Scal(double alpha, DDL_Object x)
{
  int ierr; 
  int one = 1;
  int x_size, x_itype;
  MPI_Comm x_comm;
  MPI_Datatype x_type;
  void *x_data;
  float salpha;

  ierr = DDL_SUCCESS;

  /* check input parameters */
  ierr = DDL_Size_vector(x, &x_size);
  ierr = DDL_Type_vector(x, &x_type);

  /* get pointers to raw data values */
  ierr = DDL_Get_vector(x, &x_data);
  
  /* local scal calling FORTRAN blas routine */
  if (x_type == MPI_FLOAT) {
    salpha = (float)alpha;
    F77(sscal)(&x_size, &salpha, x_data, &one);}
  else if (x_type == MPI_DOUBLE)
    F77(dscal)(&x_size, &alpha, x_data, &one);
  return ierr;
}   
    
/* fill vector with random floating point values between 0-1
   supports DOUBLE only
   Parameters:
     x: OUT         vector object
     alpha: IN      seed
   Status: H2 
   might reorder parameters? */
int DDL_Rand(double alpha, DDL_Object x)
{
  int ierr; 
  int one = 1;
  int x_size, x_itype;
  MPI_Comm x_comm;
  MPI_Datatype x_type;
  void *x_data;
  float salpha;

  ierr = DDL_SUCCESS;

  /* check input parameters */
  ierr = DDL_Size_vector(x, &x_size);
  ierr = DDL_Type_vector(x, &x_type);

  /* get pointers to raw data values */
  ierr = DDL_Get_vector(x, &x_data);
  
  /* local rand calling FORTRAN routine */
  if (x_type == MPI_FLOAT) {
  }
  else if (x_type == MPI_DOUBLE)
    F77(drandn)(&x_size, x_data, &alpha);
  return ierr;
}   
    
/* nrm2: level 1 BLAS routine to calculate 2-norm of vector
   supports REAL and DOUBLE 
   Parameters:
     x: IN          vector object
     nrm2: OUT      2-norm
   Status: H1 */
int DDL_Nrm2(DDL_Object x, void *nrm2)
{
  int ierr; 
  int x_size, x_itype;
  MPI_Comm x_comm;
  MPI_Datatype x_type;
  void *x_data;
  float flt;
  double dbl;

  ierr = DDL_SUCCESS;

  ierr = DDL_Type_vector(x, &x_type);

  if (x_type == MPI_FLOAT) {
    ierr = DDL_Dot(x, x, &flt);
    dbl = sqrt((double)flt);
    *(float*)nrm2 = (float)dbl;
  }
  else if (x_type == MPI_DOUBLE) {
    ierr = DDL_Dot(x, x, &dbl);
    *(double*)nrm2 = sqrt(dbl);
  }

  return ierr;
}   

/* axpy: level 1 BLAS routine 
   supports REAL and DOUBLE 
   Parameters:
     alpha: IN      scale factor
     x: IN          vector object
     y: INOUT       vector object
   Status: H2
   might reorder parameters? */
int DDL_Axpy(double alpha, DDL_Object x, DDL_Object y)
{
  int ierr;
  int ret;
  int one = 1;
  int x_size, x_itype;
  MPI_Comm x_comm;
  MPI_Datatype x_type;
  void *x_data;
  int y_size, y_itype;
  MPI_Comm y_comm;
  MPI_Datatype y_type;
  void *y_data;
  float salpha;

  ierr = DDL_SUCCESS;

  /* check input parameters */
  ierr = DDL_Size_vector(x, &x_size);
  ierr = DDL_Size_vector(y, &y_size);
  ierr = DDL_Type_vector(x, &x_type);
  ierr = DDL_Type_vector(y, &y_type);
  ierr = DDL_Comm_vector(x, &x_comm);
  ierr = DDL_Comm_vector(y, &y_comm);

  if (x_type != y_type) {
    ierr = DDL_Error(DDL_ERR_ITYPE, DDL_ERR_HARD, DDL_AXPY);
    return ierr;}
  if (x_size != y_size) {
    ierr = DDL_Error(DDL_ERR_ISIZE, DDL_ERR_HARD, DDL_AXPY);
    return ierr;}
  /* check communicators are **identical**
     same number of processes and ranking of processes */
  ierr = MPI_Comm_compare(x_comm, y_comm, &ret);
  if (ret != MPI_IDENT) {
    ierr = DDL_Error(DDL_ERR_ICOMM, DDL_ERR_HARD, DDL_AXPY);
    return ierr;}
  
  /* get pointers to raw data values */
  ierr = DDL_Get_vector(x, &x_data);
  ierr = DDL_Get_vector(y, &y_data);
  
  /* local axpy calling FORTRAN blas routine */
  if (x_type == MPI_FLOAT) {
    salpha = (float)alpha;
    F77(saxpy)(&x_size, &salpha, x_data, &one, y_data, &one);
  }
  else if (x_type == MPI_DOUBLE) {
    F77(daxpy)(&x_size, &alpha, x_data, &one, y_data, &one);
  }
  return ierr;
}   
    
/* copy one vector to another
   vectors must currently be identical in size and distribution
   supports REAL and DOUBLE and INT
   Parameters:
     x: IN          vector object
     y: OUT         vector object
   Status: H1 */
int DDL_Copy_vector(DDL_Object x, DDL_Object y)
{
  int ierr, ret; 
  int one = 1;
  int x_size, x_itype;
  MPI_Comm x_comm;
  MPI_Datatype x_type;
  void *x_data;
  int y_size, y_itype;
  MPI_Comm y_comm;
  MPI_Datatype y_type;
  void *y_data;

  ierr = DDL_SUCCESS;

  /* check input parameters */
  ierr = DDL_Size_vector(x, &x_size);
  ierr = DDL_Size_vector(y, &y_size);
  ierr = DDL_Type_vector(x, &x_type);
  ierr = DDL_Type_vector(y, &y_type);
  ierr = DDL_Comm_vector(x, &x_comm);
  ierr = DDL_Comm_vector(y, &y_comm);

  if (x_type != y_type) {
    ierr = DDL_Error(DDL_ERR_ITYPE, DDL_ERR_HARD, DDL_COPY_VECTOR);
    return ierr;}
  if (x_size != y_size) {
    ierr = DDL_Error(DDL_ERR_ISIZE, DDL_ERR_HARD, DDL_COPY_VECTOR);
    return ierr;}
  /* check communicators are **identical**
     same number of processes and ranking of processes */
  ierr = MPI_Comm_compare(x_comm, y_comm, &ret);
  if (ret != MPI_IDENT) {
    ierr = DDL_Error(DDL_ERR_ICOMM, DDL_ERR_HARD, DDL_COPY_VECTOR);
    return ierr;}
  
  /* get pointers to raw data values */
  ierr = DDL_Get_vector(x, &x_data);
  ierr = DDL_Get_vector(y, &y_data);
  
  /* local copy calling FORTRAN blas routine */
  if (x_type == MPI_FLOAT)
    F77(scopy)(&x_size, x_data, &one, y_data, &one);
  else if (x_type == MPI_DOUBLE)
    F77(dcopy)(&x_size, x_data, &one, y_data, &one);
  return ierr;
}   
    
/* fill vector with constant value
   supports REAL, DOUBLE, INT
   Parameters:
     x: OUT         vector object
     val: IN        constant value
   Status: H1 */
int DDL_Fill_vector(DDL_Object x, double val)
{
  int ierr, ret; 
  int i;
  int x_size, x_itype;
  MPI_Comm x_comm;
  MPI_Datatype x_type;
  void *x_data;
  float fltval;
  int intval;

  ierr = DDL_SUCCESS;

  ierr = DDL_Size_vector(x, &x_size);
  ierr = DDL_Type_vector(x, &x_type);

  /* get pointers to raw data values */
  ierr = DDL_Get_vector(x, &x_data);
  
  if (x_type == MPI_FLOAT) {
    fltval = (float)val;
    for (i=0;i<x_size;i++)
      ((float*)x_data)[i] = fltval;
  }
  else if (x_type == MPI_DOUBLE)
    for (i=0;i<x_size;i++)
      ((double*)x_data)[i] = val;
  else if (x_type == MPI_INT) {
    intval = (int)val;
    for (i=0;i<x_size;i++)
      ((int*)x_data)[i] = intval;
  }
  return ierr;
}   
   
/* vector communication ops. */
 
/* gather distributed vector object to single node vector
   supports REAL, DOUBLE, INT
   Parameters:
     vec: IN        vector to gather
     dest: OUT      destination for vector (only changed on root process)
     size: IN       max size of destination buffer in elements
     root: IN       root process for gather ie. destination process
   Status: L1 */
int DDL_Gather_vector(DDL_Object vec, void *dest, int size, int root)
{
  int ierr;
  int n, nproc, proc, i;
  MPI_Comm comm;
  MPI_Datatype type;
  void *data;
  int *cnt, *disp;
  int own[2], *owned;
  
  /* get object details */
  ierr = DDL_SUCCESS;
  ierr = DDL_Type_vector(vec, &type);
  ierr = DDL_Gsize_vector(vec, &n);

  /* get communication info */
  DDL_Comm_vector(vec, &comm);
  MPI_Comm_size(comm, &nproc);
  MPI_Comm_rank(comm, &proc);

  /* check space for data on root */
  if (proc == root)
    if (size < n) {
      ierr = DDL_Error(DDL_ERR_TOOBIG, DDL_ERR_HARD, DDL_GATHER_VECTOR);
    }

  /* tell all nodes whether to continue */
  MPI_Bcast(&ierr, 1, MPI_INT, 0, comm);

  /* exit if error */
  if (ierr != DDL_SUCCESS) return ierr;

  /* OK, we can continue */

  /* decide which elements this node owns */
  DDL_Offset_vector(vec, own);
  DDL_Size_vector(vec, &own[1]);

  if (proc == root) {
    owned = malloc(sizeof(int)*2*nproc);
    cnt = malloc(sizeof(int)*nproc);
    disp = malloc(sizeof(int)*nproc);
  }
  /* get the information about which bit of the object each node owns */
  MPI_Gather(own, 2, MPI_INT, owned, 2, MPI_INT, root, comm);
  /* reorder information for MPI_Gatherv */
  if (proc == root)
    for (i=0;i<nproc;i++) {
      disp[i] = owned[i*2];
      cnt[i] = owned[i*2+1];
    }

  /* get pointers to raw vector data */
  ierr = DDL_Get_vector(vec, &data);

  /* gather vector to root */
  MPI_Gatherv(data, own[1], type, dest, cnt, disp, type, root, comm);

  if (proc == root) {
    free(owned);
    free(cnt);
    free(disp);
  }

  return ierr;
}

/* scatter single node vector to distributed vector object
   supports REAL, DOUBLE, INT
   Parameters:
     vec: OUT       vector to scatter data into
     src: IN        source for vector (only read on root process)
     size: IN       size of source buffer in elements
     root: IN       root process for scatter ie. source process
   Status: L1 */
int DDL_Scatter_vector(DDL_Object vec, void *src, int size, int root)
{
  int ierr;
  int n, nproc, proc, i;
  MPI_Comm comm;
  MPI_Datatype type;
  void *data;
  int *cnt, *disp;
  int own[2], *owned;
  
  /* get object details */
  ierr = DDL_SUCCESS;
  ierr = DDL_Type_vector(vec, &type);
  ierr = DDL_Gsize_vector(vec, &n);

  /* get communication info */
  DDL_Comm_vector(vec, &comm);
  MPI_Comm_size(comm, &nproc);
  MPI_Comm_rank(comm, &proc);

  /* check space for data on nodes */
  if (proc == root)
    if (size > n) {
      ierr = DDL_Error(DDL_ERR_TOOSMALL, DDL_ERR_HARD, DDL_SCATTER_VECTOR);
    }

  /* tell all nodes whether to continue */
  MPI_Bcast(&ierr, 1, MPI_INT, 0, comm);

  /* exit if error */
  if (ierr != DDL_SUCCESS) return ierr;

  /* OK, we can continue */

  /* decide which elements this node owns */
  DDL_Offset_vector(vec, own);
  DDL_Size_vector(vec, &own[1]);

  if (proc == root) {
    owned = malloc(sizeof(int)*2*nproc);
    cnt = malloc(sizeof(int)*nproc);
    disp = malloc(sizeof(int)*nproc);
  }
  /* get the information about which bit of the object each node owns */
  MPI_Gather(own, 2, MPI_INT, owned, 2, MPI_INT, root, comm);
  /* reorder information for MPI_Scatterv */
  if (proc == root)
    for (i=0;i<nproc;i++) {
      disp[i] = owned[i*2];
      cnt[i] = owned[i*2+1];
    }

  /* get pointers to raw vector data */
  ierr = DDL_Get_vector(vec, &data);

  /* scatter vector to nodes */
  MPI_Scatterv(src, cnt, disp, type, data, own[1], type, root, comm);

  if (proc == root) {
    free(owned);
    free(cnt);
    free(disp);
  }

  return ierr;
}

/* check relative error between two vectors
   supports REAL, DOUBLE, INT
   Parameters:
     x: IN          vector 1
     true: IN       vector 2
     err: OUT       relative error
   Status: H3 */
int DDL_Check_err(DDL_Object x, DDL_Object true, double *err)
{
  void *xv, *tv;
  double fe = 0;
  int i, n, ie = 0;
  int ierr = DDL_SUCCESS;
  MPI_Comm comm;
  MPI_Datatype type;
  ierr = DDL_Type_vector(x, &type);
  ierr = DDL_Comm_vector(x, &comm);
  ierr = DDL_Get_vector(x, &xv);
  ierr = DDL_Get_vector(true, &tv);
  ierr = DDL_Size_vector(x, &n);
  if (type == MPI_INT) {
    for (i=0;i<n;i++)
      if (abs(((int*)xv)[i]-((int*)tv)[i])/abs(((int*)tv)[i]) > ie)
	ie = abs(((int*)xv)[i]-((int*)tv)[i])/abs(((int*)tv)[i]);
    ierr = MPI_Allreduce(&ie, err, 1, type, MPI_MAX, comm);
    *err = *(int*)err;
  }
  else if (type == MPI_FLOAT) {
    for (i=0;i<n;i++)
      if (fabs(((float*)xv)[i]-((float*)tv)[i])/fabs(((float*)tv)[i]) > fe)
	fe = fabs(((float*)xv)[i]-((float*)tv)[i])/fabs(((float*)tv)[i]);
    ierr = MPI_Allreduce(&fe, err, 1, type, MPI_MAX, comm);
  }
  else if (type == MPI_DOUBLE) {
    for (i=0;i<n;i++)
      if (fabs(((double*)xv)[i]-((double*)tv)[i])/fabs(((double*)tv)[i]) > fe)
	fe = fabs(((double*)xv)[i]-((double*)tv)[i])/fabs(((double*)tv)[i]);
    ierr = MPI_Allreduce(&fe, err, 1, type, MPI_MAX, comm);
  }

  return ierr;
}



