/* $Id: ddl_generic.c,v 1.4 1995/08/29 11:24:11 tim Exp $ */

/* Copyright 1994 IASC */

/* generic DDL routines
   This file contains generic routines for both vectors and sparse matrices
   These procedures should generally be used in preference to the 
   object specific procedures. */

/* Procedures and parameter lists:
   The DDL is continually evolving and this often results in new procedures
   or changes to parameter lists, and sometimes the removal of procedures.
   Each procedure is marked to indicate its longevity and use:
     1-4: indicates longevity of procedure interface 
        1: won't change (ever, honest!)
	2: unlikely to change, but possible
	3: might well be changed
	4: very likely to be changed or removed 
     HL: indicates expected use of procedure
        H: high level procedure, probably documented in written docs,
	   for user programs and library procedures
	L: low level procedure, not documented in written docs,
           for use in Library procedures mainly
   eg. DDL_Create_sparse: H1.
   High level procedures are usually unlikely to change. 
   Parameters are documented:
     IN:      input only parameter; is not changed by procedure
     OUT:     output parameter; any value on entry; changed on exit
     INOUT:   both input and output; read on input; may be changed on exit */

#include <stdio.h>
#include <mpi.h>
#include <stdlib.h>
#include <malloc.h>
#include "ddl.h"

/* declare external error numbers and procedure numbers */

/* DDL error severity */
int DDL_ERR_HARD               = 1;
int DDL_ERR_SOFT               = 2;

/* DDL error numbers */
#define DDL_ERR_BASE                 100

int DDL_SUCCESS=                  0;
int DDL_ERR=                      DDL_ERR_BASE+  1;
int DDL_ERR_UNALLOCATED=          DDL_ERR_BASE+  2;
int DDL_ERR_MEMORY=               DDL_ERR_BASE+  3;
int DDL_ERR_SIZE=                 DDL_ERR_BASE+  4;
int DDL_ERR_TYPE=                 DDL_ERR_BASE+  5;
int DDL_ERR_FORMAT=               DDL_ERR_BASE+  6;
int DDL_ERR_BLOCK=                DDL_ERR_BASE+  7;
int DDL_ERR_COMMTOP=              DDL_ERR_BASE+  8;
int DDL_ERR_COMMDIM=              DDL_ERR_BASE+  9;
int DDL_ERR_USED=                 DDL_ERR_BASE+  10;
int DDL_ERR_FILE=                 DDL_ERR_BASE+  11;
int DDL_ERR_NOTIMPLEMENTED=       DDL_ERR_BASE+  12;
int DDL_ERR_TOOSMALL=             DDL_ERR_BASE+  13;
int DDL_ERR_FILEFORMAT=           DDL_ERR_BASE+  15;
int DDL_ERR_ISIZE=                DDL_ERR_BASE+  16;
int DDL_ERR_ITYPE=                DDL_ERR_BASE+  17;
int DDL_ERR_IFORMAT=              DDL_ERR_BASE+  18;
int DDL_ERR_ICOMM=                DDL_ERR_BASE+  19;
int DDL_ERR_DDLTYPE=              DDL_ERR_BASE+  20;
int DDL_ERR_TOOBIG=               DDL_ERR_BASE+  21;
int DDL_ERR_IOBJ=                 DDL_ERR_BASE+  22;
int DDL_ERR_OVER=                 DDL_ERR_BASE+  23;
int DDL_ERR_DIST=                 DDL_ERR_BASE+  24;
int DDL_ERR_IPROPTAG=             DDL_ERR_BASE+  25;
int DDL_ERR_NOTEXIST=             DDL_ERR_BASE+  26;
int DDL_ERR_NOCOMMTAG=            DDL_ERR_BASE+  27;

/* DDL dense library routine numbers */
#define DDL_BASE_DENSE               200

int DDL_CREATE_VECTOR          = DDL_BASE_DENSE+  1;
int DDL_FREE_VECTOR            = DDL_BASE_DENSE+  2;
int DDL_WRITE_VECTOR           = DDL_BASE_DENSE+  3;
int DDL_READ_VECTOR            = DDL_BASE_DENSE+  4;
int DDL_SIZE_VECTOR            = DDL_BASE_DENSE+  5;
int DDL_TYPE_VECTOR            = DDL_BASE_DENSE+  6;
int DDL_FORMAT_VECTOR          = DDL_BASE_DENSE+  7;
int DDL_OFFSET_VECTOR          = DDL_BASE_DENSE+  8;
int DDL_IN_VECTOR              = DDL_BASE_DENSE+  9;
int DDL_OUT_VECTOR             = DDL_BASE_DENSE+  10;
int DDL_FILEFORMAT             = DDL_BASE_DENSE+  11;
int DDL_DOT                    = DDL_BASE_DENSE+  12;
int DDL_GET_VECTOR             = DDL_BASE_DENSE+  13;
int DDL_FILEFORMAT_HOST        = DDL_BASE_DENSE+  14;
int DDL_COMM_VECTOR            = DDL_BASE_DENSE+  15;
int DDL_GSIZE_VECTOR           = DDL_BASE_DENSE+  16;
int DDL_COPY_VECTOR            = DDL_BASE_DENSE+  17;
int DDL_AXPY                   = DDL_BASE_DENSE+  18;
int DDL_FILL_VECTOR            = DDL_BASE_DENSE+  19;
int DDL_GATHER_VECTOR          = DDL_BASE_DENSE+  20;
int DDL_SCATTER_VECTOR         = DDL_BASE_DENSE+  21;
int DDL_BLOCK_VECTOR           = DDL_BASE_DENSE+  22;

/* generic DDL library routine numbers */
#define DDL_BASE_GENERIC             DDL_GENERIC

int DDL_FREE=                     DDL_BASE_GENERIC+  1;
int DDL_READ=                     DDL_BASE_GENERIC+  2;
int DDL_WRITE=                    DDL_BASE_GENERIC+  3;
int DDL_SIZE=                     DDL_BASE_GENERIC+  4;
int DDL_OFFSET=                   DDL_BASE_GENERIC+  5;
int DDL_TYPE=                     DDL_BASE_GENERIC+  6;
int DDL_FORMAT=                   DDL_BASE_GENERIC+  7;
int DDL_COMM=                     DDL_BASE_GENERIC+  8;
int DDL_ERROR=                    DDL_BASE_GENERIC+  9;
int DDL_INIT=                     DDL_BASE_GENERIC+  10;
int DDL_FINALIZE=                 DDL_BASE_GENERIC+  11;
int DDL_COPY=                     DDL_BASE_GENERIC+  12;
int DDL_FILL=                     DDL_BASE_GENERIC+  13;
int DDL_GSIZE=                    DDL_BASE_GENERIC+  14;
int DDL_SKIP_HOST=                DDL_BASE_GENERIC+  15;
int DDL_BLOCK=                    DDL_BASE_GENERIC+  16;
int DDL_OBJECTTYPE=               DDL_BASE_GENERIC+  17;
int DDL_TAG=                      DDL_BASE_GENERIC+  18;

/* DDL library routine numbers */
#define DDL_BASE_SPARSE              300

int DDL_CREATE_SPARSE          = DDL_BASE_SPARSE+  1;
int DDL_FREE_GLOBAL            = DDL_BASE_SPARSE+  2;
int DDL_FREE_SPARSE            = DDL_BASE_SPARSE+  3;
int DDL_TYPE_SPARSE            = DDL_BASE_SPARSE+  4;
int DDL_SIZE_SPARSE            = DDL_BASE_SPARSE+  5;
int DDL_FORMAT_SPARSE          = DDL_BASE_SPARSE+  6;
int DDL_INIT_SPARSE            = DDL_BASE_SPARSE+  7;
int DDL_FINALIZE_SPARSE        = DDL_BASE_SPARSE+  9;
int DDL_INFO_SPARSE            = DDL_BASE_SPARSE+  11;
int DDL_INHB                   = DDL_BASE_SPARSE+  13;
int DDL_IN_SPARSE              = DDL_BASE_SPARSE+  14;
int DDL_OUT_SPARSE             = DDL_BASE_SPARSE+  15;
int DDL_COLTOTRIAD             = DDL_BASE_SPARSE+  16;
int DDL_OFFSET_SPARSE          = DDL_BASE_SPARSE+  17;
int DDL_COMM_SPARSE            = DDL_BASE_SPARSE+  18;
int DDL_GSIZE_SPARSE           = DDL_BASE_SPARSE+  19;
int DDL_READ_SPARSE            = DDL_BASE_SPARSE+  20;
int DDL_USED_SPARSE            = DDL_BASE_SPARSE+  21;
int DDL_SETUSED_SPARSE         = DDL_BASE_SPARSE+  22;
int DDL_SETFORMAT_SPARSE       = DDL_BASE_SPARSE+  23;
int DDL_WRITE_SPARSE           = DDL_BASE_SPARSE+  24;
int DDL_GUSED_SPARSE           = DDL_BASE_SPARSE+  25;
int DDL_GET_IA                 = DDL_BASE_SPARSE+  26;
int DDL_GET_JA                 = DDL_BASE_SPARSE+  27;
int DDL_GET_A                  = DDL_BASE_SPARSE+  28;
int DDL_SMV_COMM               = DDL_BASE_SPARSE+  29;
int DDL_SMV                    = DDL_BASE_SPARSE+  30;
int DDL_COPY_SPARSE            = DDL_BASE_SPARSE+  31;

int DDL_TRISOLVE               = DDL_BASE_SPARSE+  32;
int DDL_SETSUBTYPE_SPARSE      = DDL_BASE_SPARSE+  33;
int DDL_SUBTYPE_SPARSE         = DDL_BASE_SPARSE+  34;
int DDL_CHECK_PERM             = DDL_BASE_SPARSE+  35;
int DDL_COMPARE_VECVEC         = DDL_BASE_SPARSE+  36;
int DDL_COMPARE_MATVEC         = DDL_BASE_SPARSE+  37;
int DDL_BLOCK_SPARSE           = DDL_BASE_SPARSE+  38;
int DDL_CREATE_PROP            = DDL_BASE_SPARSE+  39;
int DDL_FREE_PROP              = DDL_BASE_SPARSE+  40;
int DDL_FIND_PROP              = DDL_BASE_SPARSE+  41;
int DDL_FREE_ALLPROP           = DDL_BASE_SPARSE+  42;
int DDL_SETPERM_SPARSE         = DDL_BASE_SPARSE+  43;
int DDL_TRISOLVE_COMM          = DDL_BASE_SPARSE+  44;
int DDL_ADDPERM_VECTOR         = DDL_BASE_SPARSE+  47;
int DDL_PERMUTE_VECTOR         = DDL_BASE_SPARSE+  48;
int DDL_INVPERM                = DDL_BASE_SPARSE+  49;
int DDL_ILU_COMM               = DDL_BASE_SPARSE+  50; 
int DDL_ILU_FORM               = DDL_BASE_SPARSE+  51;

/* DDL library routine numbers */
#define DDL_BASE_SUPPORT             DDL_SUPPORT

int DDL_OPEN_HOST              = DDL_BASE_SUPPORT+  1;
int DDL_CLOSE_HOST             = DDL_BASE_SUPPORT+  2;
int DDL_SKIP_HOST_SPARSE       = DDL_BASE_SUPPORT+  3;
int DDL_SKIP_HOST_VECTOR       = DDL_BASE_SUPPORT+  4;

/* image routine */
#define DDL_BASE_IMAGE               1000
int DDL_IMAGE_CREATE           = DDL_BASE_IMAGE+  1;

/* free object memory
   Parameters:
     obj: INOUT     object to delete; NULL on exit
   Status: H1 */
int DDL_Free(DDL_Object obj)
{
  int ierr;

  /* check parameters */
  if (obj == NULL) {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_FREE);
    return ierr;
  }
  switch (obj->type) {
  case DDL_PROP_SPARSE:
    ierr = DDL_Free_sparse(obj);
    break;
  case DDL_PROP_VECTOR:
    ierr = DDL_Free_vector(obj);
    break;
  case DDL_PROP_DUMMY:
    ierr = DDL_Free_dummy(obj);
    break;
  default:
    ierr = DDL_Error(DDL_ERR_DDLTYPE, DDL_ERR_HARD, DDL_FREE);
  }
  return ierr;
}

/* DDLR_Prop info. retrieval functions */

/* find object type ie DDL_PROP_SPARSE, VECTOR etc (see documentation)
   Parameters:
     obj: IN        object to access
     type: OUT      object type
   Status: H1 */
int DDL_Objecttype(DDL_Object obj, int *type)
{
  int ierr = DDL_SUCCESS;

  /* check parameters */
  if (obj == NULL) {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_OBJECTTYPE);
    return ierr;
  }
  *type = obj->type;

  return ierr;
}

/* find object tag (see documentation)
   Parameters:
     obj: IN        object to access
     tag: OUT       object tag
   Status: H1 */
int DDL_Tag(DDL_Object obj, int *tag)
{
  int ierr = DDL_SUCCESS;

  /* check parameters */
  if (obj == NULL) {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_TAG);
    return ierr;
  }
  *tag = obj->tag;

  return ierr;
}

/* find data type of object elements
   Parameters:
     obj: IN        object to access
     type: OUT      type of individual elements
   Status: H1 */
int DDL_Type(DDL_Object obj, MPI_Datatype *type)
{
  int ierr;

  /* check parameters */
  if (obj == NULL) {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_TYPE);
    return ierr;
  }
  switch (obj->type) {
  case DDL_PROP_SPARSE:
    ierr = DDL_Type_sparse(obj, type);
    break;
  case DDL_PROP_VECTOR:
    ierr = DDL_Type_vector(obj, type);
    break;
  default:
    ierr = DDL_Error(DDL_ERR_DDLTYPE, DDL_ERR_HARD, DDL_TYPE);
  }
  return ierr;
}

/* find local size of object
   Parameters:
     obj: IN        object to access
     size: OUT      vector giving size in each dimension of object
                      vectors have single element: total size
		      sparse matrices have 2:
		        0: number of rows (columns) held locally
			1: total number of elements that can be stored
   Status: H1 */
int DDL_Size(DDL_Object obj, int *size)
{
  int ierr;

  /* check parameters */
  if (obj == NULL) {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_SIZE);
    return ierr;
  }
  switch (obj->type) {
  case DDL_PROP_SPARSE:
    ierr = DDL_Size_sparse(obj, size);
    break;
  case DDL_PROP_VECTOR:
    ierr = DDL_Size_vector(obj, size);
    break;
  default:
    ierr = DDL_Error(DDL_ERR_DDLTYPE, DDL_ERR_HARD, DDL_SIZE);
  }
  return ierr;
}

/* get global size of object
   Parameters:
     obj: IN        object to access
     size: OUT      vector giving global size in each dimension of object
                      vectors have single element: global size
		      sparse matrices have 3:
		        0: total number of rows
		        1: total number of columns
			2: total number of nonzeros that can be stored
   Status: H1 */
int DDL_Gsize(DDL_Object obj, int *size)
{
  int ierr;

  /* check parameters */
  if (obj == NULL) {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_SIZE);
    return ierr;
  }
  switch (obj->type) {
  case DDL_PROP_SPARSE:
    ierr = DDL_Gsize_sparse(obj, size);
    break;
  case DDL_PROP_VECTOR:
    ierr = DDL_Gsize_vector(obj, size);
    break;
  default:
    ierr = DDL_Error(DDL_ERR_DDLTYPE, DDL_ERR_HARD, DDL_GSIZE);
  }
  return ierr;
}

/* find global offset of local block of object
   Parameters:
     obj: IN        object to access
     offset: OUT    vector giving global offset in each dimension of object
                      vectors have single element: offset
		      sparse matrices have 1:
		        0: offset of first row (column) held locally
   Status: H1 */
int DDL_Offset(DDL_Object obj, int *offset)
{
  int ierr;

  /* check parameters */
  if (obj == NULL) {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_OFFSET);
    return ierr;
  }
  switch (obj->type) {
  case DDL_PROP_SPARSE:
    ierr = DDL_Offset_sparse(obj, offset);
    break;
  case DDL_PROP_VECTOR:
    ierr = DDL_Offset_vector(obj, offset);
    break;
  default:
    ierr = DDL_Error(DDL_ERR_DDLTYPE, DDL_ERR_HARD, DDL_OFFSET);
  }
  return ierr;
}

/* find blocking factor used in distribution
   Parameters:
     obj: IN        object to access
     block: OUT     block factor
   Status: H1 */
int DDL_Block(DDL_Object obj, int *block)
{
  int ierr;

  /* check parameters */
  if (obj == NULL) {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_BLOCK);
    return ierr;
  }
  switch (obj->type) {
  case DDL_PROP_SPARSE:
    ierr = DDL_Block_sparse(obj, block);
    break;
  case DDL_PROP_VECTOR:
    ierr = DDL_Block_vector(obj, block);
    break;
  default:
    ierr = DDL_Error(DDL_ERR_DDLTYPE, DDL_ERR_HARD, DDL_BLOCK);
  }
  return ierr;
}

/* find format of object
   Parameters:
     obj: IN        object to access
     format: OUT    format
   Status: H1 */
int DDL_Format(DDL_Object obj, int *format)
{
  int ierr;

  /* check parameters */
  if (obj == NULL) {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_FORMAT);
    return ierr;
  }
  switch (obj->type) {
  case DDL_PROP_SPARSE:
    ierr = DDL_Format_sparse(obj, format);
    break;
  case DDL_PROP_VECTOR:
    ierr = DDL_Format_vector(obj, format);
    break;
  default:
    ierr = DDL_Error(DDL_ERR_DDLTYPE, DDL_ERR_HARD, DDL_FORMAT);
  }
  return ierr;
}

/* find MPI communicator of object
   Parameters:
     obj: IN        object to access
     comm: OUT      format
   Status: H1 */
int DDL_Comm(DDL_Object obj, MPI_Comm *comm)
{
  int ierr;

  /* check parameters */
  if (obj == NULL) {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_COMM);
    return ierr;
  }
  switch (obj->type) {
  case DDL_PROP_SPARSE:
    ierr = DDL_Comm_sparse(obj, comm);
    break;
  case DDL_PROP_VECTOR:
    ierr = DDL_Comm_vector(obj, comm);
    break;
  default:
    ierr = DDL_Error(DDL_ERR_DDLTYPE, DDL_ERR_HARD, DDL_COMM);
  }
  return ierr;
}

/* skip over object data values in input file
   to be called only on the host process
   Parameters:
     file: INOUT     input file
   Status: H1 */
int DDL_Skip_host(FILE *file)
{
  int ierr, format;

#ifndef HAVE_FPOS_T /* gcc compiler not know fpos_t */
  long fpos;
#else
  fpos_t fpos;
#endif

  /* read in the format */
#ifndef HAVE_FPOS_T /* gcc 2.5.7 compiler not know fgetpos? */
    fpos = ftell(file);
#else
    fgetpos(file, &fpos);
#endif
  fscanf(file, "%10d\n", &format);
#ifndef HAVE_FPOS_T /* gcc 2.5.7 compiler not know fsetpos? */
    fseek(file, fpos, 0); /* should be SEEK_SET */
#else
    fsetpos(file, &fpos);
#endif

  switch (format) {
  case DDL_TRIAD_ROW:
  case DDL_TRIAD_COL:
    ierr = DDL_Skip_host_sparse(file);
    break;
  case DDL_VEC_FMT:
    ierr = DDL_Skip_host_vector(file);
    break;
  default:
    ierr = DDL_Error(DDL_ERR_DDLTYPE, DDL_ERR_HARD, DDL_SKIP_HOST);
  }
  return ierr;
}

/* skip over object data values in input file
   to be called by all processes
   Parameters:
     file: INOUT     input file
     comm: IN        MPI communicator giving processes involved
   Status: H1 */
int DDL_Skip(FILE *file, MPI_Comm comm)
{
  if (DDL_Ishost()) 
      DDL_Skip_host(file);
}


/* read object data from input file
   to be called by all processes
   Parameters:
     obj: OUT        object to fill with data
     file: INOUT     input file
   Status: H1 */
int DDL_Read(DDL_Object obj, FILE *file)
{
  int ierr;

  /* check parameters */
  if (obj == NULL) {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_READ);
    return ierr;
  }
  switch (obj->type) {
  case DDL_PROP_SPARSE:
    ierr = DDL_Read_sparse(obj, file);
    break;
  case DDL_PROP_VECTOR:
    ierr = DDL_Read_vector(obj, file);
    break;
  default:
    ierr = DDL_Error(DDL_ERR_DDLTYPE, DDL_ERR_HARD, DDL_READ);
  }
  return ierr;
}

/* write object data to output file
   to be called by all processes
   Parameters:
     obj: IN         object to output
     file: INOUT     output file
   Status: H1 */
int DDL_Write(DDL_Object obj, FILE *file)
{
  int ierr;

  /* check parameters */
  if (obj == NULL) {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_WRITE);
    return ierr;
  }
  switch (obj->type) {
  case DDL_PROP_SPARSE:
    ierr = DDL_Write_sparse(obj, file);
    break;
  case DDL_PROP_VECTOR:
    ierr = DDL_Write_vector(obj, file);
    break;
  default:
    ierr = DDL_Error(DDL_ERR_DDLTYPE, DDL_ERR_HARD, DDL_WRITE);
  }
  return ierr;
}

/* copy data from one object to another
   objects must be *identical* in size and distribution
   Parameters:
     src: IN         source object
     dest: OUT       destination object
   Status: H1 */
int DDL_Copy(DDL_Object src, DDL_Object dest)
{
  int ierr;

  /* check parameters */
  if ((src == NULL) || (dest == NULL)) {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_COPY);
    return ierr;
  }
  switch (src->type) {
  case DDL_PROP_SPARSE:
    ierr = DDL_Copy_sparse(src, dest);
    break;
  case DDL_PROP_VECTOR:
    ierr = DDL_Copy_vector(src, dest);
    break;
  default:
    ierr = DDL_Error(DDL_ERR_DDLTYPE, DDL_ERR_HARD, DDL_COPY);
  }
  return ierr;
}

/* fill object with a constant value
   only fill vectors at present
   Parameters:
     obj: OUT       object to fill
     val: IN        constant value
   Status: H1 */
int DDL_Fill(DDL_Object obj, double val)
{
  int ierr;

  /* check parameters */
  if (obj == NULL) {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_FILL);
    return ierr;
  }
  switch (obj->type) {
  case DDL_PROP_SPARSE:
    ierr = DDL_Error(DDL_ERR_DDLTYPE, DDL_ERR_HARD, DDL_FILL);
    break;
  case DDL_PROP_VECTOR:
    ierr = DDL_Fill_vector(obj, val);
    break;
  default:
    ierr = DDL_Error(DDL_ERR_DDLTYPE, DDL_ERR_HARD, DDL_FILL);
  }
  return ierr;
}

/* print error information to screen; used by default error handler
   Parameters:
     err_no: IN     error number
     err_severity: IN severity of error; not currently used
     ddl_routine: IN DDL routine in which error occurred
   Status: L2 */
int F77(ddl_print_error)(int *err_no, int *err_severity, int *ddl_routine)
{
  if (DDL_Ishost()){
    char *et;
    char *rt;
    /* get error text */
    if     (*err_no == DDL_SUCCESS)             et=DDL_ERRT_SUCCESS;
    else if (*err_no == DDL_ERR_UNALLOCATED)    et=DDL_ERRT_UNALLOCATED;
    else if (*err_no == DDL_ERR_MEMORY)         et=DDL_ERRT_MEMORY;
    else if (*err_no == DDL_ERR_SIZE)           et=DDL_ERRT_SIZE;
    else if (*err_no == DDL_ERR_TYPE)           et=DDL_ERRT_TYPE;
    else if (*err_no == DDL_ERR_FORMAT)         et=DDL_ERRT_FORMAT;
    else if (*err_no == DDL_ERR_BLOCK)          et=DDL_ERRT_BLOCK; 
    else if (*err_no == DDL_ERR_COMMTOP)        et=DDL_ERRT_COMMTOP;
    else if (*err_no == DDL_ERR_COMMDIM)        et=DDL_ERRT_COMMDIM;
    else if (*err_no == DDL_ERR_USED)           et=DDL_ERRT_USED;   
    else if (*err_no == DDL_ERR_FILE)           et=DDL_ERRT_FILE;   
    else if (*err_no == DDL_ERR_NOTIMPLEMENTED) et=DDL_ERRT_NOTIMPLEMENTED;
    else if (*err_no == DDL_ERR_TOOSMALL)       et=DDL_ERRT_TOOSMALL;      
    else if (*err_no == DDL_ERR_FILEFORMAT)     et=DDL_ERRT_FILEFORMAT;    
    else if (*err_no == DDL_ERR_ISIZE)          et=DDL_ERRT_ISIZE;         
    else if (*err_no == DDL_ERR_ITYPE)          et=DDL_ERRT_ITYPE;         
    else if (*err_no == DDL_ERR_IFORMAT)        et=DDL_ERRT_IFORMAT;       
    else if (*err_no == DDL_ERR_ICOMM)          et=DDL_ERRT_ICOMM;         
    else if (*err_no == DDL_ERR_IPROPTAG)       et=DDL_ERRT_IPROPTAG;      
    else if (*err_no == DDL_ERR_DDLTYPE)        et=DDL_ERRT_DDLTYPE;       
    else if (*err_no == DDL_ERR_DIST)           et=DDL_ERRT_DIST;          
    else et="Unknown error condition";

    /* get routine name */
    /* generic routines */
    if (*ddl_routine == DDL_FREE)                    rt=DDL_T_FREE;
    else if (*ddl_routine == DDL_OBJECTTYPE)         rt=DDL_T_OBJECTTYPE;
    else if (*ddl_routine == DDL_TAG)                rt=DDL_T_TAG;
    else if (*ddl_routine == DDL_READ)               rt=DDL_T_READ;
    else if (*ddl_routine == DDL_WRITE)              rt=DDL_T_WRITE;         
    else if (*ddl_routine == DDL_SIZE)               rt=DDL_T_SIZE;          
    else if (*ddl_routine == DDL_GSIZE)              rt=DDL_T_GSIZE;         
    else if (*ddl_routine == DDL_OFFSET)             rt=DDL_T_OFFSET;        
    else if (*ddl_routine == DDL_TYPE)               rt=DDL_T_TYPE;          
    else if (*ddl_routine == DDL_FORMAT)             rt=DDL_T_FORMAT;        
    else if (*ddl_routine == DDL_COMM)               rt=DDL_T_COMM;          
    else if (*ddl_routine == DDL_INIT)               rt=DDL_T_INIT;         
    else if (*ddl_routine == DDL_FINALIZE)           rt=DDL_T_FINALIZE;     
    /* dense routines */
    else if (*ddl_routine == DDL_CREATE_VECTOR)      rt=DDL_T_CREATE_VECTOR; 
    else if (*ddl_routine == DDL_FREE_VECTOR)        rt=DDL_T_FREE_VECTOR;   
    else if (*ddl_routine == DDL_READ_VECTOR)        rt=DDL_T_READ_VECTOR;   
    else if (*ddl_routine == DDL_WRITE_VECTOR)       rt=DDL_T_WRITE_VECTOR;  
    else if (*ddl_routine == DDL_TYPE_VECTOR)        rt=DDL_T_TYPE_VECTOR;   
    else if (*ddl_routine == DDL_SIZE_VECTOR)        rt=DDL_T_SIZE_VECTOR;   
    else if (*ddl_routine == DDL_FORMAT_VECTOR)      rt=DDL_T_FORMAT_VECTOR; 
    else if (*ddl_routine == DDL_OFFSET_VECTOR)      rt=DDL_T_OFFSET_VECTOR; 
    else if (*ddl_routine == DDL_COMM_VECTOR)        rt=DDL_T_COMM_VECTOR;   
    else if (*ddl_routine == DDL_GSIZE_VECTOR)       rt=DDL_T_GSIZE_VECTOR;  
    else if (*ddl_routine == DDL_IN_VECTOR)          rt=DDL_T_IN_VECTOR;     
    else if (*ddl_routine == DDL_OUT_VECTOR)         rt=DDL_T_OUT_VECTOR;    
     /* BLAS routines */
    else if (*ddl_routine == DDL_DOT)                rt=DDL_T_DOT;           
    /* sparse routines */
    else if (*ddl_routine == DDL_CREATE_SPARSE)      rt=DDL_T_CREATE_SPARSE; 
    else if (*ddl_routine == DDL_FREE_GLOBAL)        rt=DDL_T_FREE_GLOBAL;   
    else if (*ddl_routine == DDL_FREE_SPARSE)        rt=DDL_T_FREE_SPARSE;   
    else if (*ddl_routine == DDL_TYPE_SPARSE)        rt=DDL_T_TYPE_SPARSE;   
    else if (*ddl_routine == DDL_SIZE_SPARSE)        rt=DDL_T_SIZE_SPARSE;   
    else if (*ddl_routine == DDL_FORMAT_SPARSE)      rt=DDL_T_FORMAT_SPARSE; 
    else if (*ddl_routine == DDL_INIT_SPARSE)        rt=DDL_T_INIT_SPARSE;   
    else if (*ddl_routine == DDL_FINALIZE_SPARSE)    rt=DDL_T_FINALIZE_SPARSE;
    else if (*ddl_routine == DDL_INFO_SPARSE)        rt=DDL_T_INFO_SPARSE;   
    else if (*ddl_routine == DDL_INHB)               rt=DDL_T_INHB;          
    else if (*ddl_routine == DDL_IN_SPARSE)          rt=DDL_T_IN_SPARSE;     
    else if (*ddl_routine == DDL_OUT_SPARSE)         rt=DDL_T_OUT_SPARSE;    
    else if (*ddl_routine == DDL_COLTOTRIAD)         rt=DDL_T_COLTOTRIAD;    
    else if (*ddl_routine == DDL_OFFSET_SPARSE)      rt=DDL_T_OFFSET_SPARSE; 
    else if (*ddl_routine == DDL_COMM_SPARSE)        rt=DDL_T_COMM_SPARSE;   
    else if (*ddl_routine == DDL_GSIZE_SPARSE)       rt=DDL_T_GSIZE_SPARSE;  
    else if (*ddl_routine == DDL_READ_SPARSE)        rt=DDL_T_READ_SPARSE;   
    else if (*ddl_routine == DDL_USED_SPARSE)        rt=DDL_T_USED_SPARSE;   
    else if (*ddl_routine == DDL_SETUSED_SPARSE)     rt=DDL_T_SETUSED_SPARSE;
    else if (*ddl_routine == DDL_SETFORMAT_SPARSE)   rt=DDL_T_SETFORMAT_SPARSE;
    else if (*ddl_routine == DDL_WRITE_SPARSE)       rt=DDL_T_WRITE_SPARSE;   
    else if (*ddl_routine == DDL_GUSED_SPARSE)       rt=DDL_T_GUSED_SPARSE;   
    else if (*ddl_routine == DDL_GET_IA)             rt=DDL_T_GET_IA;         
    else if (*ddl_routine == DDL_GET_JA)             rt=DDL_T_GET_JA;         
    else if (*ddl_routine == DDL_GET_A)              rt=DDL_T_GET_A;          
    /* support routines */
    else if (*ddl_routine == DDL_OPEN_HOST)          rt=DDL_T_OPEN_HOST;      
    else if (*ddl_routine == DDL_CLOSE_HOST)         rt=DDL_T_CLOSE_HOST;     
    else if (*ddl_routine == DDL_FILEFORMAT)         rt=DDL_T_FILEFORMAT;     
    else if (*ddl_routine == DDL_FILEFORMAT_HOST)    rt=DDL_T_FILEFORMAT_HOST;

    /* ...others */
    else if (*ddl_routine == DDL_SMV_COMM)           rt=DDL_T_SMV_COMM;       
    else if (*ddl_routine == DDL_SMV)                rt=DDL_T_SMV;            
    else if (*ddl_routine == DDL_COPY_SPARSE)        rt=DDL_T_COPY_SPARSE;    
    else if (*ddl_routine == DDL_TRISOLVE)           rt=DDL_T_TRISOLVE;       
    else if (*ddl_routine == DDL_TRISOLVE_COMM)      rt=DDL_T_TRISOLVE_COMM;  
    else if (*ddl_routine == DDL_BLOCK_SPARSE)       rt=DDL_T_BLOCK_SPARSE;   
    else if (*ddl_routine == DDL_CREATE_PROP)        rt=DDL_T_CREATE_PROP;    
    else if (*ddl_routine == DDL_FREE_PROP)          rt=DDL_T_FREE_PROP;      
    else if (*ddl_routine == DDL_FIND_PROP)          rt=DDL_T_FIND_PROP;      
    else if (*ddl_routine == DDL_FREE_ALLPROP)       rt=DDL_T_FREE_ALLPROP;   
    else if (*ddl_routine == DDL_SETPERM_SPARSE)     rt=DDL_T_SETPERM_SPARSE; 

    else  rt="Unknown routine";

    printf("ERROR: (%d) %s in %s\n", *err_no, et, rt);
  }
  return *err_no;
}

/* initialise base level DDL
   basically sets up MPI communicator covering all processes 
   to be used by DDL
   Parameters:
     oldcomm: IN    old MPI communicator; usually MPI_COMM_WORLD
     newcomm: OUT   communicator to be used by DDL
   Status: H3 */
int DDL_Init(MPI_Comm oldcomm, MPI_Comm *newcomm)
{
  int nprocs;                    /* number of processes */
  int dims[2]={0,0};             /* for creating cartesian communicator */
  int periods[2]={0,0}; 
  FILE *trace; /* for nice windowed output */
#ifdef _DEBUG_MALLOC_INC
  union dbmalloptarg  debug;

  debug.i = 1;
  dbmallopt(MALLOC_CKCHAIN, &debug);
#endif

#ifdef WIN
  /* debugging DDL routines=1 */
  DDL_Debug_set(1);
#endif

#ifdef WIN
  /* use traced output */
  trace = DDL_Debug_pipe("filemon2.tcl", "main.%d");
#endif

  /* set up cartesian communicator for demos */
  /* should be Nx1 */
  dims[1] = 1;
  MPI_Comm_size(oldcomm, &nprocs);
  MPI_Dims_create(nprocs, 2, dims);
/*  DDL_DEBUG_CHECK(dims, 2, oldcomm);*/
  MPI_Cart_create(oldcomm, 2, dims, periods, 0, newcomm);
  return DDL_SUCCESS;
}

/* tidy up and exit DDL
   doesn't do anything much yet
   Parameters:
     comm: IN       communicator used by DDL
   Status: H3 */
int DDL_Finalize(MPI_Comm comm)
{
  /* close down base level DDL */
  MPI_Comm_free(&comm);
  return DDL_SUCCESS;
}

