/* $Id: ddl_sparse.c,v 1.4 1995/08/22 09:18:25 tim Exp $ */

/* Copyright 1994 IASC */

/* basic sparse DDL routines
   This file contains routines for the sparse distributed matrix DDL_Spa */

/* Procedures and parameter lists:
   The DDL is continually evolving and this often results in new procedures
   or changes to parameter lists, and sometimes the removal of procedures.
   Each procedure is marked to indicate its longevity and use:
     1-4: indicates longevity of procedure interface 
        1: won't change (ever, honest!)
	2: unlikely to change, but possible
	3: might well be changed
	4: very likely to be changed or removed 
     HL: indicates expected use of procedure
        H: high level procedure, probably documented in written docs,
	   for user programs and library procedures
	L: low level procedure, not documented in written docs,
           for use in Library procedures mainly
   eg. DDL_Create_sparse: H1.
   High level procedures are usually unlikely to change. 
   Parameters are documented:
     IN:      input only parameter; is not changed by procedure
     OUT:     output parameter; any value on entry; changed on exit
     INOUT:   both input and output; read on input; may be changed on exit */

#include <stdio.h>
#include <mpi.h>
#include <stdlib.h>
#include <malloc.h>

#include "ddl.h"

/* low level routine to create DDLR_Sparse structure
   can be called from DDL_Create_sparse for complete object 
   including DDL_Prop or directly after DDL_Prop is created
   Parameters:
     size: IN       global size of matrix
                      0: number of rows
                      1: number of columns
		      2: maximum number of nonzeros to allocate
     type: IN       type of individual elements using MPI datatypes
     block: IN      smallest block size for distribution; see docs
     comm: IN       MPI communicator for object gives processes to use
     sparse: OUT    pointer to DDL_Sparse structure
   Status: L3 */
int DDLL_Create_sparse(int size[3], MPI_Datatype type, int format,
		       int block, MPI_Comm comm, DDL_Sparse *sparse)
{
  int ierr = DDL_SUCCESS;
  DDL_Sparse sp;
  int dim;
  int i, n;
  int proc, nproc;

  /* check parameters */
  if (size[0]<=0 || size[1]<=0 || size[2]<=0) {
    ierr = DDL_Error(DDL_ERR_SIZE, DDL_ERR_HARD, DDL_CREATE_SPARSE);
    if (ierr > 0) return ierr;} /* cleanup and exit with error */
  if (type != MPI_FLOAT && type != MPI_DOUBLE) {
    ierr = DDL_Error(DDL_ERR_TYPE, DDL_ERR_HARD, DDL_CREATE_SPARSE);
    if (ierr > 0) return ierr;} /* cleanup and exit with error */
  if (format != DDL_PCK_ROW && format != DDL_PCK_COL &&
      format != DDL_PAD_ROW && format != DDL_PAD_COL &&
      format != DDL_TRIAD_ROW && format != DDL_TRIAD_COL &&
      format != DDL_TRIAD) {
    ierr = DDL_Error(DDL_ERR_FORMAT, DDL_ERR_HARD, DDL_CREATE_SPARSE);
    if (ierr > 0) return ierr;} /* cleanup and exit with error */
  if (block <= 0) {
    ierr = DDL_Error(DDL_ERR_BLOCK, DDL_ERR_HARD, DDL_CREATE_SPARSE);
    if (ierr > 0) return ierr;} /* cleanup and exit with error */
  if ((ierr = DDL_Check_comm(comm)) != DDL_SUCCESS) {
    ierr = DDL_Error(ierr, DDL_ERR_HARD, DDL_CREATE_SPARSE);
    if (ierr > 0) return ierr;} /* cleanup and exit with error */
  if (format == DDL_PCK_ROW || format == DDL_PAD_ROW ||
      format == DDL_TRIAD_ROW) 
    n=size[0]; /* blocked by rows */
  else
    n=size[1]; /* blocked by columns */
  if (n % block != 0) {
    ierr = DDL_Error(DDL_ERR_BLOCK,DDL_ERR_HARD,DDL_CREATE_SPARSE);
    if (ierr > 0) return ierr; /* cleanup and exit with error */
  }

  /* create sparse storage structure */
  sp = malloc(sizeof(struct DDLR_Sparse));
  if (sp == NULL) {
    ierr = DDL_Error(DDL_ERR_MEMORY, DDL_ERR_HARD, DDL_CREATE_SPARSE);
    if (ierr > 0) return ierr; /* cleanup and exit with error */
  }

  /* set up database entries */
  sp->n = size[0];
  sp->mxnelt = size[2];
  sp->type = type;
  sp->format = format;
  sp->comm = comm;
  sp->nelt = 0;
  sp->block = block;
  /* find block size/offset */
  MPI_Comm_size(comm, &nproc);
  MPI_Comm_rank(comm, &proc);
  sp->offset = DDL_Find_offset(n, block, nproc, proc);
  sp->size[0] = DDL_Find_size(n, block, nproc, proc);
  sp->size[1] = DDL_Find_size(size[2], 1, nproc, proc);
  if (DDL_Find_proc(n, block, nproc) != nproc) {
    ierr = DDL_Error(DDL_ERR_DIST, DDL_ERR_SOFT, DDL_CREATE_SPARSE);
    if (ierr > 0) {
      free(sp);
      return ierr;
    }
  }

  /* size of vectors */
  dim = size[2];
  /* calls to vector routines to allocate space for three vectors */
  ierr = DDL_Create_vector(dim, MPI_INT, 1, comm, &(sp->ia));
  if (ierr != MPI_SUCCESS) {
    ierr = DDL_Error(DDL_ERR_MEMORY, DDL_ERR_HARD, DDL_CREATE_SPARSE);
    if (ierr > 0) {
      free(sp);
      return ierr;}
  }
  ierr = DDL_Create_vector(dim, MPI_INT, 1, comm, &(sp->ja));
  if (ierr != MPI_SUCCESS) {
    ierr = DDL_Error(DDL_ERR_MEMORY, DDL_ERR_HARD, DDL_CREATE_SPARSE);
    if (ierr > 0) {
      DDL_Free_vector(sp->ia);
      free(sp);
      return ierr;}
  }
  ierr = DDL_Create_vector(dim, type, 1, comm, &(sp->a));
  if (ierr != MPI_SUCCESS) {
    ierr = DDL_Error(DDL_ERR_MEMORY, DDL_ERR_HARD, DDL_CREATE_SPARSE);
    if (ierr > 0) {
      DDL_Free_vector(sp->ia);
      DDL_Free_vector(sp->ja);
      free(sp);
      return ierr;}
  }
  
  *sparse = sp;
  return ierr;
}

/* high level routine to create sparse distributed matrix
   Parameters:
     size: IN       global size of matrix
                      0: number of rows
                      1: number of columns
		      2: maximum number of nonzeros to allocate
     type: IN       type of individual elements using MPI datatypes
     block: IN      smallest block size for distribution; see docs
     comm: IN       MPI communicator for object gives processes to use
     obj: OUT       handle on created matrix
   Status: H1 */
int DDL_Create_sparse(int size[3], MPI_Datatype type, int format,
		      int block, MPI_Comm comm, DDL_Object *obj)
{
  int ierr = DDL_SUCCESS;
  DDL_Vector vec;
  DDL_Sparse sp;
  DDL_Prop pr;
  int dim;
  int i, n;
  int proc, nproc;

  /* check parameters */
  if (size[0]<=0 || size[1]<=0 || size[2]<=0) {
    ierr = DDL_Error(DDL_ERR_SIZE, DDL_ERR_HARD, DDL_CREATE_SPARSE);
    if (ierr > 0) return ierr;} /* cleanup and exit with error */
  if (type != MPI_FLOAT && type != MPI_DOUBLE) {
    ierr = DDL_Error(DDL_ERR_TYPE, DDL_ERR_HARD, DDL_CREATE_SPARSE);
    if (ierr > 0) return ierr;} /* cleanup and exit with error */
  if (format != DDL_PCK_ROW && format != DDL_PCK_COL &&
      format != DDL_PAD_ROW && format != DDL_PAD_COL &&
      format != DDL_TRIAD_ROW && format != DDL_TRIAD_COL &&
      format != DDL_TRIAD) {
    ierr = DDL_Error(DDL_ERR_FORMAT, DDL_ERR_HARD, DDL_CREATE_SPARSE);
    if (ierr > 0) return ierr;} /* cleanup and exit with error */
  if (block <= 0) {
    ierr = DDL_Error(DDL_ERR_BLOCK, DDL_ERR_HARD, DDL_CREATE_SPARSE);
    if (ierr > 0) return ierr;} /* cleanup and exit with error */
  if ((ierr = DDL_Check_comm(comm)) != DDL_SUCCESS) {
    ierr = DDL_Error(ierr, DDL_ERR_HARD, DDL_CREATE_SPARSE);
    if (ierr > 0) return ierr;} /* cleanup and exit with error */
  if (format == DDL_PCK_ROW || format == DDL_PAD_ROW ||
      format == DDL_TRIAD_ROW) 
    n=size[0]; /* blocked by rows */
  else
    n=size[1]; /* blocked by columns */
  if (n % block != 0) {
    ierr = DDL_Error(DDL_ERR_BLOCK,DDL_ERR_HARD,DDL_CREATE_SPARSE);
    if (ierr > 0) return ierr; /* cleanup and exit with error */
  }

  /* allocate space for DDL_Prop */
  pr = malloc(sizeof(struct DDLR_Prop));
  if (pr == NULL) {
    ierr = DDL_Error(DDL_ERR_MEMORY, DDL_ERR_HARD, DDL_CREATE_SPARSE);
    if (ierr > 0) return ierr;} /* cleanup and exit with error */
  pr->type = DDL_PROP_SPARSE;
  pr->tag = DDL_SPARSE; /* high level sparse object */
  pr->prop = NULL;
  
  /* create sparse storage structure */
  ierr = DDLL_Create_sparse(size, type, format, block, comm, &sp);
  if (ierr != DDL_SUCCESS) {
    free(pr);
    return ierr; /* cleanup and exit with error */
  }
  pr->data = sp;

  *obj = pr;
  
  return ierr;
}

/* free object memory
   Parameters:
     obj: INOUT     sparse matrix object to delete; NULL on exit
   Status: H1 */
int DDL_Free_sparse(DDL_Object obj)
{
  int ierr = DDL_SUCCESS;
  DDL_Sparse sp;
  if (obj != NULL) {
    sp = (DDL_Sparse)(obj->data);
    ierr = DDL_Free_allprop(obj);
    DDL_Free_vector(sp->a);
    DDL_Free_vector(sp->ia);
    DDL_Free_vector(sp->ja);
    free(sp);
    free(obj);
    return ierr;}
  else {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_FREE_SPARSE);
    return ierr;}
}

/* get pointer to raw data values for IA vector
   so user can update elements of sparse matrix
   Parameters:
     obj: IN        object to manipulate
     ia: OUT        pointer to raw data
   Status: H1 */
int DDL_Get_ia(DDL_Object obj, int **ia)
{
  int ierr = DDL_SUCCESS;
  DDL_Sparse sp;
  if (obj != NULL) {
    sp = (DDL_Sparse)(obj->data);
    ierr = DDL_Get_vector(sp->ia, ia);
    return ierr;}
  else {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_GET_IA);
    return ierr;}
}

/* get pointer to raw data values for JA vector
   so user can update elements of sparse matrix
   Parameters:
     obj: IN        object to manipulate
     ja: OUT        pointer to raw data
   Status: H1 */
int DDL_Get_ja(DDL_Object obj, int **ja)
{
  int ierr = DDL_SUCCESS;
  DDL_Sparse sp;
  if (obj != NULL) {
    sp = (DDL_Sparse)(obj->data);
    ierr = DDL_Get_vector(sp->ja, ja);
    return ierr;}
  else {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_GET_JA);
    return ierr;}
}

/* get pointer to raw data values for A vector
   so user can update elements of sparse matrix
   Parameters:
     obj: IN        object to manipulate
     a: OUT         pointer to raw data
   Status: H1 */
int DDL_Get_a(DDL_Object obj, void **a)
{
  int ierr = DDL_SUCCESS;
  DDL_Sparse sp;
  if (obj != NULL) {
    sp = (DDL_Sparse)(obj->data);
    ierr = DDL_Get_vector(sp->a, a);
    return ierr;}
  else {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_GET_A);
    return ierr;}
}

/* get data type of individual elements of object
   Parameters:
     obj: IN        object to manipulate
     type: OUT      data type
   Status: H1 */
int DDL_Type_sparse(DDL_Object obj, MPI_Datatype *type)
{
  int ierr = DDL_SUCCESS;
  DDL_Sparse sp;
  if (obj != NULL) {
    sp = (DDL_Sparse)(obj->data);
    *type = sp->type;
    return ierr;}
  else {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_TYPE_SPARSE);
    return ierr;}
}

/* get local size of sparse matrix
   Parameters:
     obj: IN        object to manipulate
     size: OUT      local size
		      0: number of rows (columns) held locally
		      1: total number of elements that can be stored
   Status: H1 */
int DDL_Size_sparse(DDL_Object obj, int size[])
{
  int ierr = DDL_SUCCESS;
  DDL_Sparse sp;
  if (obj != NULL) {
    sp = (DDL_Sparse)(obj->data);
    size[0] = sp->size[0];
    size[1] = sp->size[1];
    return ierr;}
  else {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_SIZE_SPARSE);
    return ierr;}
}

/* find global offset of local block of sparse matrix
   Parameters:
     obj: IN        object to access
     offset: OUT    vector giving global offset in each dimension of object
                      0: offset of first row (column) held locally
   Status: H1 */
int DDL_Offset_sparse(DDL_Object obj, int *offset)
{
  int ierr = DDL_SUCCESS;
  DDL_Sparse sp;
  if (obj != NULL) {
    sp = (DDL_Sparse)(obj->data);
    *offset = sp->offset;
    return ierr;}
  else {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_OFFSET_SPARSE);
    return ierr;}
}

/* find blocking factor used in distribution
   Parameters:
     obj: IN        object to access
     block: OUT     block factor
   Status: H1 */
int DDL_Block_sparse(DDL_Object obj, int *block)
{
  int ierr = DDL_SUCCESS;
  DDL_Sparse sp;
  if (obj != NULL) {
    sp = (DDL_Sparse)(obj->data);
    *block = sp->block;
    return ierr;}
  else {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_BLOCK_SPARSE);
    return ierr;}
}

/* find format of sparse matrix object
   Parameters:
     obj: IN        object to access
     format: OUT    format
   Status: H1 */
int DDL_Format_sparse(DDL_Object obj, int *format)
{
  int ierr = DDL_SUCCESS;
  DDL_Sparse sp;
  if (obj != NULL) {
    sp = (DDL_Sparse)(obj->data);
    *format = sp->format;
    return ierr;}
  else {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_FORMAT_SPARSE);
    return ierr;}
}

/* set format of sparse matrix object
   Parameters:
     obj: OUT       object to access
     format: IN     format
   Status: H1 */
int DDL_Setformat_sparse(DDL_Object obj, int format)
{
  int ierr = DDL_SUCCESS;
  DDL_Sparse sp;
  if (obj != NULL) {
    sp = (DDL_Sparse)(obj->data);
    sp->format = format;
    return ierr;}
  else {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_SETFORMAT_SPARSE);
    return ierr;}
}

/* find MPI communicator of object
   Parameters:
     obj: IN        object to access
     comm: OUT      format
   Status: H1 */
int DDL_Comm_sparse(DDL_Object obj, MPI_Comm *comm)
{
  int ierr = DDL_SUCCESS;
  DDL_Sparse sp;
  if (obj != NULL) {
    sp = (DDL_Sparse)(obj->data);
    *comm = sp->comm;
    return ierr;}
  else {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_COMM_SPARSE);
    return ierr;}
}

/* find number of nonzero elements actually held locally
   Parameters:
     obj: IN        object to access
     nelt: OUT      number of nonzeros
   Status: H1 */
int DDL_Used_sparse(DDL_Object obj, int *nelt)
{
  int ierr = DDL_SUCCESS;
  DDL_Sparse sp;
  if (obj != NULL) {
    sp = (DDL_Sparse)(obj->data);
    *nelt = sp->nelt;
    return ierr;}
  else {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_USED_SPARSE);
    return ierr;}
}

/* find number of nonzero elements actually held globally
   Parameters:
     obj: IN        object to access
     nelt: OUT      number of nonzeros
   Status: H1 */
int DDL_Gused_sparse(DDL_Object obj, int *nelt)
{
  int ierr = DDL_SUCCESS;
  int local_nelt;
  DDL_Sparse sp;
  if (obj != NULL) {
    sp = (DDL_Sparse)(obj->data);
    local_nelt = sp->nelt;
    ierr = MPI_Allreduce(&local_nelt, nelt, 1, MPI_INT, MPI_SUM, sp->comm); 
    return ierr;}
  else {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_GUSED_SPARSE);
    return ierr;}
}

/* set number of nonzero elements actually held locally
   Parameters:
     obj: OUT       object to access
     nelt: IN       number of nonzeros
   Status: H1 */
int DDL_Setused_sparse(DDL_Object obj, int nelt)
{
  int ierr = DDL_SUCCESS;
  DDL_Sparse sp;
  if (obj != NULL) {
    sp = (DDL_Sparse)(obj->data);
    sp->nelt = nelt;
    return ierr;}
  else {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_SETUSED_SPARSE);
    return ierr;}
}

/* get global size of sparse matrix object
   Parameters:
     obj: IN        object to access
     size: OUT      vector giving global size in each dimension of object
                      0: total number of rows
		      1: total number of columns
		      2: total number of nonzeros that can be stored
   Status: H1 */
int DDL_Gsize_sparse(DDL_Object obj, int size[])
{
  int ierr = DDL_SUCCESS;
  DDL_Sparse sp;
  if (obj != NULL) {
    sp = (DDL_Sparse)(obj->data);
    size[0] = sp->n;
    size[1] = sp->n;
    size[2] = sp->mxnelt;
    return ierr;}
  else {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_GSIZE_SPARSE);
    return ierr;}
}

/* initialise Sparse DDL
   just calls DDL_Init now
   Parameters:
     oldcomm: IN    old MPI communicator; usually MPI_COMM_WORLD
     newcomm: OUT   communicator to be used by DDL
   Status: H3 */
int DDL_Init_sparse(MPI_Comm oldcomm, MPI_Comm *newcomm)
{
  /* initialise base level DDL */
  DDL_Init(oldcomm, newcomm);

  /* Sparse DDL specific code... */

  return DDL_SUCCESS;
}

/* tidy up and exit Sparse-DDL
   doesn't do anything much yet
   Parameters:
     comm: IN       communicator used by DDL
   Status: H3 */
int DDL_Finalize_sparse(MPI_Comm comm)
{
  /* close down sparse-specific stuff */

  /* close down base level DDL */
  DDL_Finalize(comm);
  return DDL_SUCCESS;
}

/* print information about sparse object
   just for testing purposes
   Parameters:
     sp: IN         object to check
   Status: H4 */
int DDL_Info_sparse(DDL_Object sp) 
{
  int nelt, size[3], size1[3], format, block;
  MPI_Datatype type;
  int offset;
  DDL_Sparse spr;
  spr = (DDL_Sparse)(sp->data);
  printf("Sparse details\n");
  DDL_Gsize_sparse(sp, size);
  DDL_Size_sparse(sp, size1);
  DDL_Offset_sparse(sp, &offset);
  DDL_Format_sparse(sp, &format);
  DDL_Type_sparse(sp, &type);
  DDL_Used_sparse(sp, &nelt);
  printf("global matrix is %d by %d with space for %d elements\n", 
	 size[0], size[0], size[2]);
  printf("offset is %d\n",offset);
  printf("size is %d by %d\n",size1[0], size1[1]);
  printf("format is %d\n",format);
  printf("type is %d\n",type);
  printf("number of elements used is %d\n",nelt);
  printf("local IA vector details\n");
  DDL_Offset_vector(spr->ia, &offset);
  DDL_Size_vector(spr->ia, size);
  printf("I have %d to %d of IA\n", offset, offset+size[0]-1);
  printf("local JA vector details\n");
  DDL_Offset_vector(spr->ja, &offset);
  DDL_Size_vector(spr->ja, size);
  printf("I have %d to %d of JA\n", offset, offset+size[0]-1);
  printf("local A vector details\n");
  DDL_Offset_vector(spr-> a, &offset);
  DDL_Size_vector(spr-> a, size);
  printf("I have %d to %d of A\n", offset, offset+size[0]-1);
}

/* check communicator is compatible with DDL
   Parameters:
     comm: IN       MPI communicator; usually MPI_COMM_WORLD
   Status: L3 */
int DDL_Check_comm(MPI_Comm comm)
{
  int ierr = DDL_SUCCESS;
  int topology, ndim, dims[2], periods[2], coords[2];
  MPI_Topo_test(comm, &topology);
  if (topology != MPI_CART) {
    ierr = DDL_ERR_COMMTOP;
    return ierr;}
  MPI_Cartdim_get(comm, &ndim);
  if (ndim != 2) {
    ierr = DDL_ERR_COMMDIM;
    return ierr;}
  MPI_Cart_get(comm, 2, dims, periods, coords);
  if (dims[1] != 1) {
    ierr = DDL_ERR_COMMDIM;
    return ierr;}
  return ierr;
}

/* sparse matrix I/O for DDL

   data file format uses ascii text, so is not efficient

   format supports vectors and sparse matrices currently

   each object in file has a header which identifies data to follow.
   for sparse matrix:
     format: INTEGER  specifies whether object is vector or sparse matrix:
                        DDL_VECTOR
                        DDL_TRIAD_ROW
                        DDL_TRIAD_COL
     type: INTEGER    data type of individual elements of sparse matrix
                      in memory; supports REAL, DOUBLE types
     n: INTEGER       number of rows (columns) of square matrix
     nelt: INTEGER    total number of nonzeros in matrix
     isym: INTEGER    whether matrix is symmetric (not currently used)
                        0: not symmetric
                        1: symmetric

   After the header the data values follow 1 element per line.
   Each matrix element is stored as a triad: row, column, coefficent
   Elements are sorted by row then column for file format DDL_TRIAD_ROW
    or by column then row for file format DDL_TRIAD_COL
   (see include file ddl_sparse.h for constant values defining types and
   formats) */

/* read distributed sparse matrix object from file 
   Currently, if the object format is stored by rows, then the file
   format must be DDL_TRIAD_ROW; and if the object format is stored by 
   columns, then the file format must be DDL_TRIAD_COL.
   Parameters:
     sp: OUT        object to fill with values from file
     file: INOUT    file handle
   Status: H1 */
int DDL_Read_sparse(DDL_Object sp, FILE *file)
{
  int ierr;
  int n, nelt, isym, ftype;
  MPI_Aint typesize;
  int size[3];
  int sp_n, sp_mxnelt, sp_format, sp_nelt;
  int a_size;
  int want[3], *wanted;
  MPI_Comm comm;
  MPI_Status status;
  MPI_Datatype type;
  int nproc, proc, dest;
  int i, j, total, count, format;
  int *ia, *ja;
  void *a;
  int *iat, *jat;
  void *at;
  int ia_last, ja_last;
  double a_last;
  int idx, line, *ordered;
  int zero = 0;
  void *tmp;
  DDL_Sparse spr;

  isym = 0; /* not symmetrical */

  if (sp != NULL)
    spr = (DDL_Sparse)(sp->data);
  else {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_READ_SPARSE);
    return ierr;}

  ierr = DDL_Format_sparse(sp, &sp_format);

  /* check file parameters */
  ierr = DDL_SUCCESS;
  if (DDL_Ishost()) {
    if (file == NULL) {
      ierr = DDL_Error(DDL_ERR_FILE, DDL_ERR_HARD, DDL_READ_SPARSE);
    }
    /* check file contains sparse matrix in correct format */
    else {
      fscanf(file, "%10d", &format);
      if (format != DDL_TRIAD_ROW && format != DDL_TRIAD_COL) {
	ierr = DDL_Error(DDL_ERR_FILEFORMAT, DDL_ERR_HARD, DDL_READ_SPARSE);
      }
      else {
	/* check format of file matrix is compatible with object */
	if ((format == DDL_TRIAD_ROW && 
	     (sp_format == DDL_TRIAD_COL || sp_format == DDL_PCK_COL)) ||
	    (format == DDL_TRIAD_COL && 
	     (sp_format == DDL_TRIAD_ROW || sp_format == DDL_PCK_ROW))) {
	  ierr = DDL_Error(DDL_ERR_FILEFORMAT, DDL_ERR_HARD, DDL_READ_SPARSE);
	}
	else {
	  /* get sparse matrix details from file */
	  fscanf(file, "%10d%10d%10d%10d\n", &ftype, &n, &nelt, &isym);
	  /* compare with object */
	  ierr = DDL_Gsize_sparse(sp, size);
	  sp_n = size[0];
	  sp_mxnelt = size[2];
	  if (sp_n < n || sp_mxnelt < nelt) {
	    ierr = DDL_Error(DDL_ERR_TOOSMALL, DDL_ERR_HARD, DDL_READ_SPARSE);
	  }
	}
      }
    }
  }
 
  /* get communication info */
  DDL_Comm_sparse(sp, &comm);
  MPI_Comm_size(comm, &nproc);
  MPI_Comm_rank(comm, &proc);

  /* tell all nodes whether to continue */
  MPI_Bcast(&ierr, 1, MPI_INT, 0, comm);

  /* exit if error */
  if (ierr != DDL_SUCCESS) return ierr;

  /* OK, we can continue */

  /* decide which rows/columns this node wants */
  DDL_Offset_sparse(sp, want);
  DDL_Size_sparse(sp, &want[1]);

  /* find out element type */
  ierr = DDL_Type_sparse(sp, &type);
  ierr = DDL_DDLtype(type, &ftype);
  ierr = MPI_Type_size(type, &typesize);

  /* get the information about which bit of the matrix each node wants */
  if (DDL_Ishost())
    wanted = malloc(sizeof(int)*2*nproc);
  MPI_Gather(want, 2, MPI_INT, wanted, 2, MPI_INT, 0, comm);

  /* allocate temporary space for vectors */
  if (DDL_Ishost()) {
    iat = malloc(sizeof(int)*(n+1));
    jat = malloc(sizeof(int)*(n+1));
    at = malloc(typesize*(n+1));
  }

  total = 0;
  dest = 0;
  idx = 0;
  sp_nelt = 0; /* delete any data in sparse object */
  DDL_Size_vector(spr->a, &a_size);

  /* get pointers to raw vector data */
  ierr = DDL_Get_vector(spr->ia, &ia);
  ierr = DDL_Get_vector(spr->ja, &ja);
  ierr = DDL_Get_vector(spr->a, &a);

  /* hosted code to distribute data */
  /* send as ordered triad. Convert on node to packed row/column */
  if (DDL_Ishost()) {
    i = 0;
    while (i<n) {
      /* read in line from file */
      ierr = DDL_Read_sparseline(file, nelt, iat, jat, at, type,
				 format, &count, &total, 
				 &ia_last, &ja_last, &a_last);

      /* get line number */
      if (format == DDL_TRIAD_ROW) line = iat[0]-1;
      else if (format == DDL_TRIAD_COL) line = jat[0]-1;
      /* send empty messages if empty lines skipped*/
      /* in practice there will not be any empty lines
	 since this makes the matrix indefinite---
	 but code structure useful later for arbitrary comms patterns */
      if (line != i) {
	for (j=i;j<line;j++) {
          /* find destination node for empty message j*/	  
	  if (wanted[2*dest]+wanted[2*dest+1] == j) dest = dest+1;
	  if (proc != dest) {
	    /* send empty data */
	    ierr = MPI_Send(&zero, 1, MPI_INT, dest, 0, comm);
	  }
	}
      }
      /* send message for line */
      /* find destination node for line */	  
      if (wanted[2*dest]+wanted[2*dest+1] == line) dest = dest+1;
      if (proc != dest) {
	/* send data */
	ierr = MPI_Send(&count, 1, MPI_INT, dest, 0, comm);
	if (count != 0) {
	  ierr = MPI_Send(iat, count, MPI_INT, dest, 0, comm);
	  ierr = MPI_Send(jat, count, MPI_INT, dest, 0, comm);
	  ierr = MPI_Send(at, count, type, dest, 0, comm);
	}
      }
      else {
        /* copy data to object */
        for (j=0;j<count;j++) ia[idx+j] = iat[j];
	for (j=0;j<count;j++) ja[idx+j] = jat[j];
	if (type == MPI_FLOAT)
	  for (j=0;j<count;j++) ((float *)a)[idx+j] = ((float *)at)[j];
	else if (type == MPI_DOUBLE)
	  for (j=0;j<count;j++) ((double *)a)[idx+j] = ((double *)at)[j];
	idx = idx+count;
      }
      /* update i and loop */
      i = line+1;
    }
  }
  else {
    /* each node receives its own lines only */
    for (i=0;i<want[1];i++) {
      /* receive data */
      ierr = MPI_Recv(&count, 1, MPI_INT, 0, 0, comm, &status);
      /* check for empty message */
      if (count != 0) {
        /* have I got room for data? */
	if (idx+count > a_size) {
          /* not enough memory---abort */
	}
	else {
	  ierr = MPI_Recv(&ia[idx], count, MPI_INT, 0, 0, comm, &status);
	  ierr = MPI_Recv(&ja[idx], count, MPI_INT, 0, 0, comm, &status);
	  ierr = MPI_Recv((char *)a+typesize*idx, count, 
			  type, 0, 0, comm, &status);
	  idx = idx+count;
	}
      }
    }
  }

  /* update local count */
  spr->nelt = idx;

  /* renumber triad format from 1 to give local sparse matrix */
  if (sp_format == DDL_TRIAD_ROW || sp_format == DDL_PCK_ROW) ordered = ia;
  else if (sp_format == DDL_TRIAD_COL || sp_format == DDL_PCK_COL) 
    ordered = ja;
  for (i=0;i<spr->nelt;i++) ordered[i] = ordered[i]-want[0];

  /* currently in TRIAD format---convert?*/
  if (sp_format == DDL_PCK_ROW) 
    ierr = DDL_Triadtorow(want[1], spr->nelt, ia, ja, a, type, isym);
  else if (sp_format == DDL_PCK_COL) 
    ierr = DDL_Triadtocol(want[1], spr->nelt, ia, ja, a, type, isym);

  if (DDL_Ishost()) {
    free(wanted);
    free(iat);
    free(jat);
    free(at);
  }

  return ierr;
}

/* read in a row/column of triad ordered file
   called from DDL_Read_sparse; not useful elsewhere
   Parameters:
     file: INOUT    file handle
     nelt: IN       number of nonzero elements in matrix
     ia, ja, a: OUT triad storage for elements
     type: IN       data type to store in a: REAL or DOUBLE
     format: IN     row or column ordered: DDL_TRIAD_ROW, DDL_TRIAD_COL
     count: OUT     number of elements read on this call
     total: INOUT   number of elements read in total, including previous calls
     ia_last, ja_last, a_last: INOUT on entry:first element of this row,
                    on exit:first element of next row
   Status: L3 */
int DDL_Read_sparseline(FILE *file, int nelt, int ia[], int ja[], double a[],
			MPI_Datatype type, int format, int *count, int *total, 
			int *ia_last, int *ja_last, double *a_last)
{
  int ierr;
  int i;
  int line;
  int *ordered;

  /* set up first triple */
  if (*total == 0) {
    /* new file---read entries */
    fscanf(file, "%d%d%le\n", &ia[0], &ja[0], &a[0]);
  }
  else {
    /* old file---copy last entries */
    ia[0] = *ia_last;
    ja[0] = *ja_last;
    a[0] = *a_last;
  }
  *count = 1;
  /* marker for end of line */
  if (format == DDL_TRIAD_ROW) ordered = ia;
  else if (format == DDL_TRIAD_COL) ordered = ja;
  line = ordered[0];
    
  /* read in the rest of the line */
  for (i=1;i<nelt-*total;i++) {
    fscanf(file, "%d%d%le", &ia[i], &ja[i], &a[i]);
    if (ordered[i] != line) {
      *ia_last = ia[i];
      *ja_last = ja[i];
      *a_last = a[i];
      break;
    }
  }
  *count = *count + i-1; 
  *total = *total + *count;

  return DDL_SUCCESS;
}

/* write distributed sparse matrix object to file 
   Currently, if the object format is stored by rows, then the file
   format must be DDL_TRIAD_ROW; and if the object format is stored by 
   columns, then the file format must be DDL_TRIAD_COL.
   Parameters:
     sp: OUT        object to write to file
     file: INOUT    file handle
   Status: H1 */
int DDL_Write_sparse(DDL_Object sp, FILE *file)
{
  int ierr;
  int n, nelt, isym, ftype;
  MPI_Aint typesize;
  int size[3];
  int sp_n, sp_mxnelt, sp_format, sp_nelt;
  int a_size;
  int own[3], *owned;
  MPI_Comm comm;
  MPI_Status status;
  MPI_Datatype type;
  int nproc, proc, src;
  int i, j, total, count, icount, jcount, format;
  int *ia, *ja;
  double *a;
  int *iat, *jat;
  double *at;
  int ia_last, ja_last;
  double a_last;
  int idx, line , *ordered;
  int zero = 0;
  DDL_Sparse spr;

  if (sp != NULL)
    spr = (DDL_Sparse)(sp->data);
  else {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_WRITE_SPARSE);
    return ierr;}

  isym = 0; /* not symmetrical */
  ftype = DDL_DOUBLE;

  /* get sparse object details */
  ierr = DDL_SUCCESS;
  ierr = DDL_Format_sparse(sp, &sp_format);
  ierr = DDL_Gsize_sparse(sp, size);
  sp_n = size[0];
  sp_mxnelt = size[2];
  ierr = DDL_Gused_sparse(sp, &sp_nelt);
  ierr = DDL_Used_sparse(sp, &nelt);

  if (sp_format == DDL_TRIAD_COL || sp_format == DDL_PCK_COL)
    format = DDL_TRIAD_COL;
  else if (sp_format == DDL_TRIAD_ROW || sp_format == DDL_PCK_ROW)
    format = DDL_TRIAD_ROW;

  /* get communication info */
  DDL_Comm_sparse(sp, &comm);
  MPI_Comm_size(comm, &nproc);
  MPI_Comm_rank(comm, &proc);

  if (DDL_Ishost())
    if (file == NULL) {
      ierr = DDL_Error(DDL_ERR_FILE, DDL_ERR_HARD, DDL_WRITE_SPARSE);
    }

  /* tell all nodes whether to continue */
  MPI_Bcast(&ierr, 1, MPI_INT, 0, comm);

  /* exit if error */
  if (ierr != DDL_SUCCESS) return ierr;

  /* OK, we can continue */

  /* decide which rows/columns this node owns */
  DDL_Offset_sparse(sp, own);
  DDL_Size_sparse(sp, &own[1]);

  /* get the information about which bit of the matrix each node owns */
  if (DDL_Ishost())
    owned = malloc(sizeof(int)*2*nproc);
  MPI_Gather(own, 2, MPI_INT, owned, 2, MPI_INT, 0, comm);

  total = 0;
  src = 0;
  idx = 0;

  /* find out element type */
  ierr = DDL_Type_sparse(sp, &type);
  ierr = DDL_DDLtype(type, &ftype);
  ierr = MPI_Type_size(type, &typesize);

  /* get pointers to raw vector data */
  ierr = DDL_Get_vector(spr->ia, &ia);
  ierr = DDL_Get_vector(spr->ja, &ja);
  ierr = DDL_Get_vector(spr->a, &a);

  /* hosted code to gather data */
  /* gathered as packed row/column or ordered triad. 
     Converted on host to ordered triad */
  if (DDL_Ishost()) {
    /* allocate temporary space for vectors */
    iat = malloc(sizeof(int)*sp_n);
    jat = malloc(sizeof(int)*sp_n);
    at = malloc(typesize*sp_n);

    /* write out sparse matrix file format */
    fprintf(file, "%10d\n", format);
    
    /* write sparse matrix details to file */
    fprintf(file, "%10d%10d%10d%10d\n", ftype, sp_n, sp_nelt, isym);
    
    i = 0;
    while (i<sp_n) {
      /* get line */
      /* find source node for line */	  
      if (owned[2*src]+owned[2*src+1] == i) src = src+1;
      if (proc != src) {
	/* get data */
	ierr = MPI_Recv(&count, 1, MPI_INT, src, 0, comm, &status);
	/* check for empty message */
	if (count != 0) {
	  ierr = MPI_Recv(iat, count, MPI_INT, src, 0, comm, &status);
	  ierr = MPI_Recv(jat, count, MPI_INT, src, 0, comm, &status);
	  ierr = MPI_Recv(at, count, type, src, 0, comm, &status);
	}
      }
      else {
        /* copy data to temp arrays */
	if (sp_format == DDL_TRIAD_ROW || sp_format == DDL_TRIAD_COL) { 
	  /* ordered triad format object */
	  /* find which vector is ordered */
	  if (format == DDL_TRIAD_ROW) ordered = ia;
	  else if (format == DDL_TRIAD_COL) ordered = ja;
	  /* get line number */
	  line = ordered[idx]-1;
	  /* check we have something in this line */
	  if (line == i) {
	    for (j=0;j<nelt-idx;j++) {
	      if (ordered[idx+j]-1 != line) break;
	      iat[j] = ia[idx+j];
	      jat[j] = ja[idx+j];
	    }
	    count = j;
	    /* copy a vector */
	    if (type == MPI_FLOAT)
	      for (j=0;j<count;j++) ((float *)at)[j] = ((float *)a)[idx+j];
	    else if (type == MPI_DOUBLE)
	      for (j=0;j<count;j++) ((double *)at)[j] = ((double *)a)[idx+j];
	    idx = idx+count;
	  }
	  else {
	    count = 0;
	  }
	}
	else if (sp_format == DDL_PCK_ROW || sp_format == DDL_PCK_COL) {
	  /* packed row or column format object */
	  /* find which vector is ordered */
	  if (format == DDL_TRIAD_ROW) ordered = ia;
	  else if (format == DDL_TRIAD_COL) ordered = ja;
	  count = ordered[idx+1]-ordered[idx];
	  /* check we have something in this line, and copy to temp */
	  if (count != 0) {
	    if (sp_format == DDL_PCK_ROW) {
	      for (j=0;j<count;j++) {
		jat[j] = ja[ia[idx]-1+j];
	      }
	      if (type == MPI_FLOAT)
		for (j=0;j<count;j++) 
		  ((float *)at)[j] = ((float *)a)[ia[idx]-1+j];
	      else if (type == MPI_DOUBLE)
		for (j=0;j<count;j++) 
		  ((double *)at)[j] = ((double *)a)[ia[idx]-1+j];
	      iat[0] = 1;
	      iat[1] = count+1;
	    }
	    else if (sp_format == DDL_PCK_COL) {
	      for (j=0;j<count;j++) {
		iat[j] = ia[ja[idx]-1+j];
	      }
	      if (type == MPI_FLOAT)
		for (j=0;j<count;j++) 
		  ((float *)at)[j] = ((float *)a)[ja[idx]-1+j];
	      else if (type == MPI_DOUBLE)
		for (j=0;j<count;j++) 
		  ((double *)at)[j] = ((double *)a)[ja[idx]-1+j];
	      jat[0] = 1;
	      jat[1] = count+1;
	    }
	  }
	  idx = idx+1;
	}
      }
      /* if not empty line, convert and write */
      if (count != 0) {
	/* convert to global triad */
	/* only one line i to convert */
	/* for triad---renumber ordered vector using global index
	   for packed---insert global index into ordered vector */
	/* find which vector is ordered */
	if (format == DDL_TRIAD_ROW) ordered = iat;
	else if (format == DDL_TRIAD_COL) ordered = jat;
	for (j=0;j<count;j++) 
	  ordered[j] = i+1;
	/* write line to file */
	ierr = DDL_Write_sparseline(file, iat, jat, at, type, &count);
      }
      i = i+1;
    }
  }
  else {
    /* node sends its own lines */
    for (i=0;i<own[1];i++) {
      if (sp_format == DDL_TRIAD_ROW || sp_format == DDL_TRIAD_COL) { 
	/* ordered triad format object */
	/* find which vector is ordered */
	if (format == DDL_TRIAD_ROW) ordered = ia;
	else if (format == DDL_TRIAD_COL) ordered = ja;
	/* get line number */
	line = ordered[idx]-1;
	/* check we have something in this line */
	if (line == i) {
	  for (j=1;j<nelt-idx;j++)
	    if (ordered[idx+j]-1 != line) break;
	  count = j;
	}
	else {
	  count = 0;
	}
	/* send data */
	ierr = MPI_Send(&count, 1, MPI_INT, 0, 0, comm);
	if (count != 0) {
	  ierr = MPI_Send(&ia[idx], count, MPI_INT, 0, 0, comm);
	  ierr = MPI_Send(&ja[idx], count, MPI_INT, 0, 0, comm);
	  ierr = MPI_Send((char *)a+typesize*idx, count, type, 0, 0, comm);
	}
	idx = idx+count;
      }
      else if (sp_format == DDL_PCK_ROW || sp_format == DDL_PCK_COL) {
	/* packed row or column format object */
	/* find which vector is ordered */
	if (format == DDL_TRIAD_ROW) ordered = ia;
	else if (format == DDL_TRIAD_COL) ordered = ja;
	count = ordered[i+1]-ordered[i];
	/* send data */
	ierr = MPI_Send(&count, 1, MPI_INT, 0, 0, comm);
	if (count != 0) {
	  if (format == DDL_TRIAD_ROW) {
	    icount = 0;
	    jcount = count;}
	  else {
	    icount = count;
	    jcount = 0; }
	  ierr = MPI_Send(&ia[idx], icount, MPI_INT, 0, 0, comm);
	  ierr = MPI_Send(&ja[idx], jcount, MPI_INT, 0, 0, comm);
	  ierr = MPI_Send((char *)a+typesize*idx, count, type, 0, 0, comm);
	}
	idx = idx+count;
      }
    }
  }

  if (DDL_Ishost()) {
    free(owned);
    free(iat);
    free(jat);
    free(at);
  }

  return ierr;
}

/* write out a row/column of triad ordered file
   called from DDL_Write_sparse; not useful elsewhere
   Parameters:
     file: INOUT    file handle
     ia, ja, a: OUT triad stored elements 
     type: IN       data type held in a: REAL or DOUBLE
     count: IN      number of elements to write
   Status: L3 */
int DDL_Write_sparseline(FILE *file, int ia[], int ja[], double a[],
			 MPI_Datatype type, int *count)
{
  int ierr;
  int i;
  
  /* write out the line */
  for (i=0;i<*count;i++) {
    fprintf(file, " %5d %5d %16.7le\n", ia[i], ja[i], a[i]);
  }
  return DDL_SUCCESS;
}

/* copy one sparse matrix to another
   matrices must currently be identical in size, format and distribution
   supports REAL and DOUBLE
   Parameters:
     src: IN        source sparse matrix object
     dest: OUT      destination sparse matrix object
   Status: H1 */
int DDL_Copy_sparse(DDL_Object src, DDL_Object dest)
{
  int ierr, i, ret;
  int s_size[3], s_format, s_nelt;
  MPI_Comm s_comm;
  MPI_Datatype s_type;
  void *s_data;
  int d_size[3], d_format, d_nelt;
  MPI_Comm d_comm;
  MPI_Datatype d_type;
  void *d_data;

  ierr = DDL_SUCCESS;

  /* check input parameters */
  ierr = DDL_Gsize_sparse(src, s_size);
  ierr = DDL_Gsize_sparse(dest, d_size);
  if ((s_size[0] != d_size[0]) || (s_size[1] != d_size[1])) {
    ierr = DDL_Error(DDL_ERR_ISIZE, DDL_ERR_HARD, DDL_COPY_SPARSE);
    return ierr;}
  ierr = DDL_Type_sparse(src, &s_type);
  ierr = DDL_Type_sparse(dest, &d_type);
  if (s_type != d_type) {
    ierr = DDL_Error(DDL_ERR_ITYPE, DDL_ERR_HARD, DDL_COPY_SPARSE);
    return ierr;}
  ierr = DDL_Format_sparse(src, &s_format);
  ierr = DDL_Format_sparse(dest, &d_format);
  if (s_format != d_format) {
    ierr = DDL_Error(DDL_ERR_IFORMAT, DDL_ERR_HARD, DDL_COPY_SPARSE);
    return ierr;}
  ierr = DDL_Comm_sparse(src, &s_comm);
  ierr = DDL_Comm_sparse(dest, &d_comm);
  /* check communicators are **identical**
     same number of processes and ranking of processes */
  ierr = MPI_Comm_compare(s_comm, d_comm, &ret);
  if (ret != MPI_IDENT) {
    ierr = DDL_Error(DDL_ERR_ICOMM, DDL_ERR_HARD, DDL_COPY_SPARSE);
    return ierr;}
  ierr = DDL_Gused_sparse(src, &s_nelt);
  if (s_nelt > d_size[2]) {
    ierr = DDL_Error(DDL_ERR_ISIZE, DDL_ERR_HARD, DDL_COPY_SPARSE);
    return ierr;}

  /* find local sizes */
  ierr = DDL_Size_sparse(src, s_size);
  ierr = DDL_Size_sparse(dest, d_size);
  ierr = DDL_Used_sparse(src, &s_nelt);
  ierr = DDL_Used_sparse(dest, &d_nelt);

  if ((s_format == DDL_TRIAD_ROW) || (s_format == DDL_TRIAD_COL) ||
      (s_format == DDL_TRIAD)) {
    /* just copy vectors up to nelt */
    ierr = DDL_Get_ia(src, (int**)&s_data);
    ierr = DDL_Get_ia(dest, (int**)&d_data);
    for (i=0;i<s_nelt;i++) 
      ((int*)d_data)[i] = ((int*)s_data)[i];
    ierr = DDL_Get_ja(src, (int**)&s_data);
    ierr = DDL_Get_ja(dest, (int**)&d_data);
    for (i=0;i<s_nelt;i++) 
      ((int*)d_data)[i] = ((int*)s_data)[i];
    ierr = DDL_Get_a(src, &s_data);
    ierr = DDL_Get_a(dest, &d_data);
    if (s_type == MPI_DOUBLE)
      for (i=0;i<s_nelt;i++) 
        ((double*)d_data)[i] = ((double*)s_data)[i];
    else if (s_type == MPI_FLOAT)
      for (i=0;i<s_nelt;i++) 
        ((float*)d_data)[i] = ((float*)s_data)[i];
  }
  else if ((s_format == DDL_PCK_ROW) || (s_format == DDL_PCK_COL)) {
    if (s_format == DDL_PCK_ROW) {
      ierr = DDL_Get_ia(src, (int**)&s_data);
      ierr = DDL_Get_ia(dest, (int**)&d_data);
      for (i=0;i<s_size[0];i++) 
	  ((int*)d_data)[i] = ((int*)s_data)[i];
      ierr = DDL_Get_ja(src, (int**)&s_data);
      ierr = DDL_Get_ja(dest, (int**)&d_data);
      for (i=0;i<s_nelt;i++) 
	  ((int*)d_data)[i] = ((int*)s_data)[i];
    }
    else {
      ierr = DDL_Get_ja(src, (int**)&s_data);
      ierr = DDL_Get_ja(dest, (int**)&d_data);
      for (i=0;i<s_size[0];i++) 
	  ((int*)d_data)[i] = ((int*)s_data)[i];
      ierr = DDL_Get_ia(src, (int**)&s_data);
      ierr = DDL_Get_ia(dest, (int**)&d_data);
      for (i=0;i<s_nelt;i++) 
	  ((int*)d_data)[i] = ((int*)s_data)[i];
    }
    ierr = DDL_Get_a(src, &s_data);
    ierr = DDL_Get_a(dest, &d_data);
    if (s_type == MPI_DOUBLE)
      for (i=0;i<s_nelt;i++) 
        ((double*)d_data)[i] = ((double*)s_data)[i];
    else if (s_type == MPI_FLOAT)
      for (i=0;i<s_nelt;i++) 
        ((float*)d_data)[i] = ((float*)s_data)[i];
  } 
  else if ((s_format == DDL_PAD_ROW) || (s_format == DDL_PAD_COL)) {
  }
  DDL_Setused_sparse(dest, s_nelt);
  return ierr;
}

/* Some communication routines useful for sparse operations */

/* All to all communication similar to MPI Alltoall 
   except only a subset of processes in comm are involved.

   All processes in communicator should call this routine even though
   some may do nothing. 

   The processes listed in snd (count in snd_tot_proc) will be sent a 
   message by the calling process. For process j that message is 
   snd_buf[snd_ptr[j]] for snd_cnt[j] elements.

   The calling process will receive a message from each of the processes 
   in rcv (count in rcv_tot_proc). For process j that message is received
   into rcv_buf[rcv_ptr[j]] for rcv_cnt[j] elements.

   The communication data structure is set up routines such as
   DDL_Smv_comm and DDL_Trisolve_comm.

   Parameters:
     snd_buf IN      buffer for sent data
     snd_tot_proc IN number of procs to send to
     snd IN          vector of process numbers (in comm) to send data to
     snd_cnt[j] IN   number of elements to send to process snd[j]
     snd_ptr[j] IN   index of first element in snd_buf to send to snd[j]
     snd_type IN     MPI data type of element
     rcv_buf OUT     buffer for received data
     rcv_tot_proc IN number of procs to receive from
     rcv IN          vector of process numbers (in comm) to get data from
     rcv_cnt[j] IN   number of elements to get from process rcv[j]
     rcv_ptr[j] IN   index of first element in rcv_buf to get from rcv[j]
     rcv_type IN     MPI data type of element
   Status: L2 */
int DDL_Alltoallvs(void *snd_buf, int snd_tot_proc, int snd[], 
		   int snd_cnt[],
		   int snd_ptr[], MPI_Datatype snd_type,
		   void *rcv_buf, int rcv_tot_proc, int rcv[], 
		   int rcv_cnt[],
		   int rcv_ptr[], MPI_Datatype rcv_type,
		   MPI_Comm comm)
{
  int i, max_comm, ierr;
  MPI_Aint sz;
  MPI_Status *status;
  MPI_Request *rreq, *sreq;

  ierr = DDL_SUCCESS;

  /* exchange vectors */
  max_comm = rcv_tot_proc>snd_tot_proc?rcv_tot_proc:snd_tot_proc;

  rreq = malloc(rcv_tot_proc*sizeof(MPI_Request));
  sreq = malloc(snd_tot_proc*sizeof(MPI_Request));
  status = malloc(max_comm*sizeof(MPI_Status));

  ierr = MPI_Type_size(snd_type, &sz);
  
  /* post expected receives (nonblocking) */
  for (i=0;i<rcv_tot_proc;i++) {
    MPI_Irecv(&(((char*)rcv_buf)[rcv_ptr[i]*sz]), rcv_cnt[i],
	      rcv_type, rcv[i], 0, comm, &rreq[i]);
  }

  /* now post nonblocking sends */
  for (i=0;i<snd_tot_proc;i++) {
    MPI_Isend(&(((char*)snd_buf)[snd_ptr[i]*sz]), snd_cnt[i],
	      snd_type, snd[i], 0, comm, &sreq[i]);
  }
  
  /* now wait for all communications to terminate */
  MPI_Waitall(snd_tot_proc, sreq, status);
  MPI_Waitall(rcv_tot_proc, rreq, status);

  free(status);
  free(rreq);
  free(sreq);

  return ierr;
}

/* send identical message to multiple processes without sending 
   message to myself.
   receiving processes use MPI_Recv to receive message.
   Process list comes from comms data structures such as DDL_Ind.

   Parameters:
     buf: IN         buffer containing output data
     cnt: IN         number of elements in buffer
     type: IN        MPI_Datatype of elements
     nproc: IN       number of procs to send to
     proc: IN        vector of process numbers to send data to
     tag: IN         MPI comms tag
     comm: IN        MPI Communicator
   Status: L2 */
int DDL_Msend(void *buf, int cnt, MPI_Datatype type, 
	      int nproc, int *proc, int tag, MPI_Comm comm)
{
  int i, j, max_comm, ierr, rank;
  MPI_Aint sz;
  MPI_Status *status, stat;
  MPI_Request *req;
  ierr = DDL_SUCCESS;
  req = malloc(nproc*sizeof(MPI_Request));
  status = malloc(nproc*sizeof(MPI_Status));
  ierr = MPI_Comm_rank(comm, &rank);
  j = 0;
  /* now post nonblocking sends */
  for (i=0;i<nproc;i++) 
    if (proc[i] != rank) {
      MPI_Isend(buf, cnt, type, proc[i], tag, comm, &(req[j]));
      j = j+1;
    }
  /* now wait for all communications to terminate */
  MPI_Waitall(j, req, status);
  free(status);
  free(req);
  return ierr;
}

/* Broadcast communication similar to MPI Bcast
   except only a subset of processes are involved

   Only the root process and those processes expecting to receive 
   the message should call this routine.

   Parameters:
     buf: INOUT      buffer for data; OUT on root IN on others
     cnt: IN         number of elements in buffer
     type: IN        MPI_Datatype of elements
     root: IN        source process for data
     nproc: IN       number of procs to send to
     proc: IN        vector of process numbers to send data to
     comm: IN        MPI Communicator
   Status: L2 */
int DDL_Bcasts(void *buf, int cnt, MPI_Datatype type, 
	       int root, int nproc, int *proc, MPI_Comm comm)
{
  int i, max_comm, ierr, rank;
  MPI_Aint sz;
  MPI_Status *status, stat;
  MPI_Request *req;
  ierr = DDL_SUCCESS;
  ierr = MPI_Comm_rank(comm, &rank); 
  if (rank == root) {
    req = malloc(nproc*sizeof(MPI_Request));
    status = malloc(nproc*sizeof(MPI_Status));
    /* now post nonblocking sends */
    for (i=0;i<nproc;i++) 
      MPI_Isend(&buf, cnt, type, proc[i], 0, comm, &(req[i]));
  }
  /* now receive message */
  MPI_Recv(&buf, cnt, type, root, 0, comm, &stat);  
  if (rank == root) {
    /* now wait for all communications to terminate */
    MPI_Waitall(nproc, req, status);
    free(status);
    free(req);
  }
  return ierr;
}

/* sequential sparse matrix, dense vector multiply:
         y=alpha*Ax+beta*y
   Currently, only supports DDL_PCK_ROW sparse matrix format
   Parameters:
     n: IN           number of rows in matrix
     m: IN           number of columns in matrix
     ia, ja, a: IN   matrix elements in DDL_PCK_ROW format
     type: IN        MPI_Datatype of coefficient elements
     nelt: IN        number of nonzeros in matrix
     format: IN      DDL format of matrix; must be DDL_PCK_ROW currently
     x: IN           input vector
     y: INOUT        result vector
     alpha: IN       scale factor
     beta: IN        scale factor
   Status: L1 */
int DDL_Smv_seq(int n, int m, int ia[],int ja[], 
		void *a, MPI_Datatype type,
		int nelt, int format, void *x, void *y,
		double alpha, double beta)
{
  int i,j;
  double tmp;
  if (format == DDL_PCK_ROW) {
    if (type == MPI_DOUBLE) {
      for (i=0;i<n;i++) {
	tmp = 0.0;
	for (j=ia[i]-1;j<ia[i+1]-1;j++) 
	  tmp = tmp + ((double*)a)[j]*((double*)x)[ja[j]-1];
	((double*)y)[i] = beta*((double*)y)[i]+alpha*tmp;
      }
    }
    else if (type == MPI_FLOAT) {
      for (i=0;i<n;i++) {
	tmp = 0.0;
	for (j=ia[i]-1;j<ia[i+1]-1;j++) 
	  tmp = tmp + ((float*)a)[j]*((float*)x)[ja[j]-1];
	((float*)y)[i] = beta*((float*)y)[i]+alpha*tmp;
      }
    }
  }
  return DDL_SUCCESS;
}

/* find indices of coeff vector for a row of sparse matrix
   in DDL_PCK_ROW format.
   Parameters:
     row: IN         number of row in matrix to find
     ia: IN          vector of row pointers
     low: OUT        index into coefficient vector, a, of first element of row
     high: OUT       index of first element of next row
   Status: L3 */
int DDL_Coeff_ind(int row, int ia[], int *low, int *high)
{
  *low = ia[row];
  *high = ia[row+1];
  return DDL_SUCCESS;
}
