/* $Id: ddl_sphigh.c,v 1.5 1995/08/22 14:59:50 tim Exp $ */

/* Copyright 1995 IASC */

/* DDL routines supporting property lists, permutations
   and sparse operations such as mv and triangular solve */

/* Procedures and parameter lists:
   The DDL is continually evolving and this often results in new procedures
   or changes to parameter lists, and sometimes the removal of procedures.
   Each procedure is marked to indicate its longevity and use:
     1-4: indicates longevity of procedure interface 
        1: won't change (ever, honest!)
	2: unlikely to change, but possible
	3: might well be changed
	4: very likely to be changed or removed 
     HL: indicates expected use of procedure
        H: high level procedure, probably documented in written docs,
	   for user programs and library procedures
	L: low level procedure, not documented in written docs,
           for use in Library procedures mainly
   eg. DDL_Create_sparse: H1.
   High level procedures are usually unlikely to change. 
   Parameters are documented:
     IN:      input only parameter; is not changed by procedure
     OUT:     output parameter; any value on entry; changed on exit
     INOUT:   both input and output; read on input; may be changed on exit */

#include <stdio.h>
#include <mpi.h>
#include <stdlib.h>
#include <malloc.h>

#include "ddl.h"

/* create dummy object 
   creates an object with no primary data record
   useful in Library routines to create an object to which 
   temporary properties can be added
   Parameters:
     d: OUT         pointer to dummy object
   Status: L1 */
int DDL_Create_dummy(DDL_Object *d)
{
  int ierr = DDL_SUCCESS;
  DDL_Prop pr;

  /* allocate space for DDL_Prop */
  pr = malloc(sizeof(struct DDLR_Prop));
  pr->type = DDL_PROP_DUMMY;
  pr->tag = DDL_DUMMY; /* high level sparse object */
  pr->prop = NULL;
  pr->data = NULL;

  *d = pr;
  return ierr;
}

/* free dummy object
   Parameters:
     d: IN          dummy object
   Status: L1 */
int DDL_Free_dummy(DDL_Object d)
{
  int ierr = DDL_SUCCESS;
  ierr = DDL_Free_allprop(d);
  free(d);
  return ierr;
}

/* compare 2 vectors; compares size, type, and comm
   Parameters:
     x: IN          vector 1
     y: IN          vector 2
     ret: OUT       result:
                      MPI_IDENT: objects are compatible
		      others an error value
   Status: L2 */
int DDL_Compare_vecvec(DDL_Object x, DDL_Object y, int *ret)
{
  int ierr = DDL_SUCCESS;
  int x_size, y_size; 
  MPI_Datatype x_type, y_type;
  MPI_Comm x_comm, y_comm;
  if (x == NULL || y == NULL) {
    *ret = DDL_ERR_UNALLOCATED;
    return ierr;}
  ierr = DDL_Gsize(x, &x_size);
  ierr = DDL_Gsize(y, &y_size);
  ierr = DDL_Type(x, &x_type);
  ierr = DDL_Type(y, &y_type);
  ierr = DDL_Comm(x, &x_comm);
  ierr = DDL_Comm(y, &y_comm);
  if (x_type != y_type) {
    *ret = DDL_ERR_ITYPE;
    return ierr;}
  if (x_size != y_size) {
    *ret = DDL_ERR_ISIZE;
    return ierr;}
  /* check communicators are **identical**
     same number of processes and ranking of processes */
  ierr = MPI_Comm_compare(x_comm, y_comm, ret);
  if (*ret != MPI_IDENT) {
    *ret = DDL_ERR_ICOMM;
    return ierr;}
  *ret = MPI_IDENT;
  return ierr;
}

/* compare sparse matrix and vector
   check #cols of matrix == size of vector
   for mv multiply or solve vec must be LHS vec
   Parameters:
     x: IN          matrix
     y: IN          vector
     ret: OUT       result:
                      MPI_IDENT: objects are compatible
		      others an error value
   Status: L2 */
int DDL_Compare_matvec(DDL_Object x, DDL_Object y, int *ret)
{
  int ierr = DDL_SUCCESS;
  int x_size, y_size;
  int size[3];
  MPI_Datatype x_type, y_type;
  MPI_Comm x_comm, y_comm;
  if (x == NULL || y == NULL) {
    *ret = DDL_ERR_UNALLOCATED;
    return ierr;}
  ierr = DDL_Gsize(x, size);
  ierr = DDL_Gsize(y, &y_size);
  ierr = DDL_Type(x, &x_type);
  ierr = DDL_Type(y, &y_type);
  ierr = DDL_Comm(x, &x_comm);
  ierr = DDL_Comm(y, &y_comm);
  if (x_type != y_type) {
    *ret = DDL_ERR_ITYPE;
    return ierr;}
  if (size[1] != y_size) {
    *ret = DDL_ERR_ISIZE;
    return ierr;}
  /* check communicators are **identical**
     same number of processes and ranking of processes */
  ierr = MPI_Comm_compare(x_comm, y_comm, ret);
  if (*ret != MPI_IDENT) {
    *ret = DDL_ERR_ICOMM;
    return ierr;}
  *ret = MPI_IDENT;
  return ierr;
}

/* compare vector with permutation vector
   check distributions are compatible, 
   ie, comm is same and size is same 
   Parameters:
     x: IN          vector
     p: IN          permutation vector
     ret: OUT       result:
                      MPI_IDENT: objects are compatible
		      others an error value
   Status: L2 */
int DDL_Compare_vecperm(DDL_Object x, DDL_Object p, int *ret)
{
  int ierr = DDL_SUCCESS;
  int x_size, p_size;
  int size[3];
  MPI_Datatype x_type, p_type;
  MPI_Comm x_comm, p_comm;
  if (x == NULL || p == NULL) {
    *ret = DDL_ERR_UNALLOCATED;
    return ierr;}
  ierr = DDL_Gsize(x, &x_size);
  ierr = DDL_Gsize(p, &p_size);
  ierr = DDL_Comm(x, &x_comm);
  ierr = DDL_Comm(p, &p_comm);
  if (x_size != p_size) {
    *ret = DDL_ERR_ISIZE;
    return ierr;}
  /* check communicators are **identical**
     same number of processes and ranking of processes */
  ierr = MPI_Comm_compare(x_comm, p_comm, ret);
  if (*ret != MPI_IDENT) {
    *ret = DDL_ERR_ICOMM;
    return ierr;}
  *ret = MPI_IDENT;
  return ierr;
}

/* compare distributed matrix with row permutation vector
   check distributions are compatible, 
   ie, comm is same and size is same 
   Parameters:
     x: IN          matrix
     p: IN          permutation vector
     ret: OUT       result:
                      MPI_IDENT: objects are compatible
		      others an error value
   Status: L2 */
int DDL_Compare_matpermx(DDL_Object x, DDL_Object p, int *ret)
{
  int ierr = DDL_SUCCESS;
  int x_size, p_size;
  int size[3];
  MPI_Datatype x_type, p_type;
  MPI_Comm x_comm, p_comm;
  if (x == NULL || p == NULL) {
    *ret = DDL_ERR_UNALLOCATED;
    return ierr;}
  ierr = DDL_Gsize(x, size);
  ierr = DDL_Gsize(p, &p_size);
  ierr = DDL_Comm(x, &x_comm);
  ierr = DDL_Comm(p, &p_comm);
  if (size[0] != p_size) {
    *ret = DDL_ERR_ISIZE;
    return ierr;}
  /* check communicators are **identical**
     same number of processes and ranking of processes */
  ierr = MPI_Comm_compare(x_comm, p_comm, ret);
  if (*ret != MPI_IDENT) {
    *ret = DDL_ERR_ICOMM;
    return ierr;}
  *ret = MPI_IDENT;
  return ierr;
}

/* compare distributed matrix with column permutation vector
   check distributions are compatible, 
   ie, comm is same and size is same 
   Parameters:
     x: IN          matrix
     p: IN          permutation vector
     ret: OUT       result:
                      MPI_IDENT: objects are compatible
		      others an error value
   Status: L2 */
int DDL_Compare_matpermy(DDL_Object x, DDL_Object p, int *ret)
{
  int ierr = DDL_SUCCESS;
  int x_size, p_size;
  int size[3];
  MPI_Datatype x_type, p_type;
  MPI_Comm x_comm, p_comm;
  if (x == NULL || p == NULL) {
    *ret = DDL_ERR_UNALLOCATED;
    return ierr;}
  ierr = DDL_Gsize(x, size);
  ierr = DDL_Gsize(p, &p_size);
  ierr = DDL_Comm(x, &x_comm);
  ierr = DDL_Comm(p, &p_comm);
  if (size[1] != p_size) {
    *ret = DDL_ERR_ISIZE;
    return ierr;}
  /* check communicators are **identical**
     same number of processes and ranking of processes */
  ierr = MPI_Comm_compare(x_comm, p_comm, ret);
  if (*ret != MPI_IDENT) {
    *ret = DDL_ERR_ICOMM;
    return ierr;}
  *ret = MPI_IDENT;
  return ierr;
}

/* set subtype for sparse matrix object
   Parameters:
     obj: INOUT     matrix
     subtype: IN    sub type; values are:
                      DDL_UPPER: upper triangular matrix
                      DDL_LOWER: lower triangular matrix
   Status: H1 */
int DDL_Setsubtype_sparse(DDL_Object obj, int subtype)
{
  int ierr = DDL_SUCCESS;
  DDL_Sparse sp;
  if (obj != NULL) {
    sp = (DDL_Sparse)(obj->data);
    sp->subtype = subtype;
    return ierr;}
  else {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_SETSUBTYPE_SPARSE);
    return ierr;}
} 

/* get subtype of sparse matrix object
   Parameters:
     obj: IN        matrix
     subtype: OUT   sub type; values are:
                      DDL_UPPER: upper triangular matrix
                      DDL_LOWER: lower triangular matrix
   Status: H1 */
int DDL_Subtype_sparse(DDL_Object obj, int *subtype)
{
  int ierr = DDL_SUCCESS;
  DDL_Sparse sp;
  if (obj != NULL) {
    sp = (DDL_Sparse)(obj->data);
    *subtype = sp->subtype;
    return ierr;}
  else {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_SUBTYPE_SPARSE);
    return ierr;}
} 

/* check a distributed vector is a valid permutation 
   well, only checks type is integer at present!
   Parameters:
     p: IN          vector
     ret: OUT       result:
                      DDL_SUCCESS: vector is a valid permutation
		      others an error value
   Status: L2 */
int DDL_Check_perm(DDL_Object p, int *ret)
{
  int ierr = DDL_SUCCESS;
  MPI_Datatype type;
  int ddltype;
  if (p == NULL) {
    *ret = DDL_ERR_UNALLOCATED;
    return ierr;
  }
  ddltype = p->type;
  if (ddltype != DDL_PROP_VECTOR) {
    *ret = DDL_ERR_DDLTYPE;
    return ierr;
  }
  ierr = DDL_Type(p, &type);
  if (type != MPI_INT) {
    *ret = DDL_ERR_ITYPE;
    return ierr;
  }
  *ret = DDL_SUCCESS;
  return ierr;
}

/* object property list operations:
     create entry in list
     remove entry from list
     find entry with particular tag value */

/* create property
   Parameters:
     obj: INOUT     distributed object to add property structure to
     tag: IN        name tag for structure:
                      DDL_PX:          row permutation for matrix
                      DDL_PY:          column permutation for matrix
                      DDL_IPX:         inverse row permutation for matrix
                      DDL_IPY:         inverse column permutation for matrix
                      DDL_COMM_PX:     row communication pattern for matrix-
		                       vector multiply
                      DDL_COMM_PY:     column communication pattern for matrix-
		                       vector multiply and trisolve
                      DDL_COMM_IND:    result element communication pattern 
		                       for trisolve
     prop: OUT      pointer to new property structure
   Status: L1 */
int DDL_Create_prop(DDL_Object obj, int tag, DDL_Prop *prop)
{
  int ierr = DDL_SUCCESS;
  int i, type, ddltype;
  DDL_Prop *p;
  DDL_Glob g;
  DDL_Ind ind;

  /* check parameters */
  if (obj == NULL) {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_CREATE_PROP);
    if (ierr > 0) return ierr; /* cleanup and exit with error */
  }

  switch (obj->type) {
  case DDL_PROP_DUMMY:
    /* dummy distributed object for temporary property structures etc. */
    break;
  case DDL_PROP_SPARSE:
    /* check tag value is valid */
    if (tag != DDL_PX && tag != DDL_PY && 
	tag != DDL_IPX && tag != DDL_IPY && 
	tag != DDL_COMM_PX && tag != DDL_COMM_PY &&
	tag != DDL_COMM_IND) {
      ierr = DDL_Error(DDL_ERR_IPROPTAG, DDL_ERR_HARD, DDL_CREATE_PROP);
      if (ierr > 0) return ierr; /* cleanup and exit with error */
    }
    break;
  case DDL_PROP_VECTOR:
    /* check tag value is valid */
    if (tag != DDL_PV && tag != DDL_IPV && 
	tag != DDL_COMM_PV && tag != DDL_COMM_IPV) {
      ierr = DDL_Error(DDL_ERR_IPROPTAG, DDL_ERR_HARD, DDL_CREATE_PROP);
      if (ierr > 0) return ierr; /* cleanup and exit with error */
    }
    break;
  default:
    ierr = DDL_Error(DDL_ERR_IPROPTAG, DDL_ERR_HARD, DDL_CREATE_PROP);
    if (ierr > 0) return ierr; /* cleanup and exit with error */
  }
  /* find end of structure list */
  p = &(obj->prop);
  while (*p != NULL) {
    /* check tag doesn't already exist */
    if ((*p)->tag == tag) {
      ierr = DDL_Error(DDL_ERR_IPROPTAG, DDL_ERR_HARD, DDL_CREATE_PROP);
      if (ierr > 0) return ierr; /* cleanup and exit with error */
    }
    p = &((*p)->prop);
  }
  /* allocate and initialise storage for new property structure */
  *p = malloc(sizeof(struct DDLR_Prop));

  switch(tag) {
  case DDL_PV: type = DDL_PROP_VECTOR; break;
  case DDL_IPV: type = DDL_PROP_VECTOR; break;
  case DDL_PX: type = DDL_PROP_VECTOR; break;
  case DDL_PY: type = DDL_PROP_VECTOR; break;
  case DDL_IPX: type = DDL_PROP_VECTOR; break;
  case DDL_IPY: type = DDL_PROP_VECTOR; break;
  case DDL_COMM_PX: type = DDL_PROP_GLOB; break;
  case DDL_COMM_PY: type = DDL_PROP_GLOB; break;
  case DDL_COMM_PV: type = DDL_PROP_GLOB; break;
  case DDL_COMM_IND: type = DDL_PROP_IND; break;
  }

  (*p)->prop = NULL;
  (*p)->tag = tag;
  (*p)->type = type;
  *prop = *p;

  /* initialise type-specific structure elements */
  switch (type) {
  case DDL_PROP_GLOB:
    g = malloc(sizeof(struct DDLR_Glob));
    (*p)->data = g;
    g->rcv.nproc = 0;
    g->rcv.proc = NULL;
    g->rcv.ind = NULL;
    g->rcv.pind = NULL;
    g->rcv.cnt = NULL;
    g->rcv.ptr = NULL;
    g->snd.nproc = 0;
    g->snd.proc = NULL;
    g->snd.ind = NULL;
/*    g->snd.pind = NULL;*/
    g->snd.cnt = NULL;
    g->snd.ptr = NULL;
    break;
  case DDL_PROP_IND:
    ind = malloc(sizeof(struct DDLR_Ind));
    (*p)->data = ind;
    ind->rcv.nind = 0;
    ind->rcv.proc = NULL;
    ind->rcv.ind = NULL;
    ind->rcv.pind = NULL;
    ind->snd.nind = 0;
    ind->snd.proc = NULL;
    ind->snd.ind = NULL;
/*    ind->snd.pind = NULL;*/
    ind->snd.cnt = NULL;
    ind->snd.ptr = NULL;
    break;
  case DDL_PROP_VECTOR:
    /* don't know sizes here so just return DDLR_Prop; 
       call DDLL_Create_vector in calling routine */
    (*p)->data = NULL;
    break;
  case DDL_PROP_SPARSE:
    /* don't know sizes here so just return DDLR_Prop; 
       call DDLL_Create_sparse in calling routine */
    (*p)->data = NULL;
    break;
  default:
    ierr = DDL_Error(DDL_ERR_ITYPE, DDL_ERR_HARD, DDL_CREATE_PROP);
    if (ierr > 0) return ierr; /* cleanup and exit with error */
  }
  return ierr;
}

/* free object property; just frees up memory; doesn't relink list
   low level; no checking
   Parameters:
     prop: INOUT    pointer to property structure to free
   Status: L2 */
int DDLL_Free_prop(DDL_Prop prop)
{
  int ierr = DDL_SUCCESS;
  DDL_Glob g;
  DDL_Ind ind;
  switch (prop->type) {
  case DDL_PROP_GLOB:
    g = (DDL_Glob)(prop->data);
    free(g->snd.proc);
    free(g->snd.ind);
/*    free(g->snd.pind);*/
    free(g->snd.cnt);
    free(g->snd.ptr);
    free(g->rcv.proc);
    free(g->rcv.ind);
    free(g->rcv.pind);
    free(g->rcv.cnt);
    free(g->rcv.ptr);
    free(g);
    free(prop);
    break;
  case DDL_PROP_IND:
    ind = (DDL_Ind)(prop->data);
    free(ind->rcv.proc);
    free(ind->rcv.ind);
    free(ind->rcv.pind);
    free(ind->snd.proc);
    free(ind->snd.ind);
/*    free(ind->snd.pind);*/
    free(ind->snd.cnt);
    free(ind->snd.ptr);
    free(ind);
    free(prop);
    break;
  case DDL_PROP_VECTOR:
    ierr = DDL_Free_vector(prop);
    break;
  case DDL_PROP_SPARSE:
    ierr = DDL_Free_sparse(prop);
    break;
  }

  return ierr;
}

/* remove object property
   relinking property list
   Parameters:
     obj: INOUT     object with property
     tag: IN        name tag for property (see create for values)
   Status: L1 */
int DDL_Free_prop(DDL_Object obj, int tag)
{
  int ierr = DDL_SUCCESS;
  DDL_Prop *p, np;

  /* check parameters */
  if (obj == NULL) {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_FREE_PROP);
    if (ierr > 0) return ierr; /* cleanup and exit with error */
  }
  /* find matching tag */
  p = &(obj->prop);
  while (*p != NULL) {
    /* check tag */
    if ((*p)->tag == tag) {
      /* remember links */
      np = (*p)->prop;
      /* remove structure */
      ierr = DDLL_Free_prop(*p);
      /* relink structure list */
      *p = np;
      return ierr;
    }
    p = &((*p)->prop);
  }
  /* didn't find property */
  ierr = DDL_Error(DDL_ERR_IPROPTAG, DDL_ERR_HARD, DDL_FREE_PROP);
  return ierr; /* cleanup and exit with error */
}
 
/* remove entire property list
   Parameters:
     obj: INOUT     object
   Status: L2 */
int DDL_Free_allprop(DDL_Object obj)
{
  int ierr = DDL_SUCCESS;
  DDL_Prop p, np;

  /* check parameters */
  if (obj == NULL) {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_FREE_ALLPROP);
    if (ierr > 0) return ierr; /* cleanup and exit with error */
  }
  p = obj->prop;
  while (p != NULL) {
    /* remember links */
    np = p->prop;
    /* remove structure */
    ierr = DDLL_Free_prop(p);
    p = np;
  }
}
 
/* find property with given tag 
   Parameters:
     obj: IN        object with property
     tag: IN        name tag for property (see create for values)
     p: OUT         pointer to property if found; else NULL
   Status: L1 */
int DDL_Find_prop(DDL_Object obj, int tag, DDL_Prop *p)
{
  int ierr = DDL_SUCCESS;

  /* check parameters */
  if (obj == NULL) {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_FIND_PROP);
    if (ierr > 0) return ierr; /* cleanup and exit with error */
  }
  /* find matching tag */
  *p = obj->prop;
  while (*p != NULL) {
    /* check tag */
    if ((*p)->tag == tag) return ierr;
    *p = (*p)->prop;
  }
  *p = NULL; /* tag not found */
  return ierr;
}
 
/* specify permutation vectors for distributed sparse matrix
   these are stored with matrix object

   permutation vectors use C indices from 0

   Parameters:
     a: INOUT       sparse triangular matrix
     px_ddl: IN     possible row permutation vector; or NULL
     py_ddl: IN     possible column permutation vector; or NULL
   Status: H1 */
int DDL_Setperm_sparse(DDL_Object a, DDL_Object px_ddl,
		       DDL_Object py_ddl)
{
  int ierr = DDL_SUCCESS;
  int nproc, proc, dummy, offset, ret;
  int rcv_tot_data, snd_tot_data;
  int *ia, *ja, *lpy, *lpx, *px, *py, *tmp;
  int nr, nc, n, size[3], i, j, low, high, subtype, block;
  int format;
  MPI_Comm comm;
  DDL_Prop p;
  DDL_Vector v;

  /* check parameters */
  if (a == NULL) {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_SETPERM_SPARSE);
    if (ierr > 0) return ierr; /* cleanup and exit with error */
  }
  /* check for permutation vectors */
  if (px_ddl != NULL) {
    ierr = DDL_Get_vector(px_ddl, &lpx);
    ierr = DDL_Check_perm(px_ddl, &ret);
    if (ret != DDL_SUCCESS) {
      ierr = DDL_Error(ret, DDL_ERR_HARD, DDL_SETPERM_SPARSE);
      if (ierr > 0) return ierr; /* cleanup and exit with error */
    }
    ierr = DDL_Compare_matpermx(a, px_ddl, &ret);
    if (ret != DDL_SUCCESS) {
      ierr = DDL_Error(ret, DDL_ERR_HARD, DDL_SETPERM_SPARSE);
      if (ierr > 0) return ierr; /* cleanup and exit with error */
    }
  }
  if (py_ddl != NULL) {
    ierr = DDL_Get_vector(py_ddl, &lpy);
    ierr = DDL_Check_perm(py_ddl, &ret);
    if (ret != DDL_SUCCESS) {
      ierr = DDL_Error(ret, DDL_ERR_HARD, DDL_SETPERM_SPARSE);
      if (ierr > 0) return ierr; /* cleanup and exit with error */
    }
    ierr = DDL_Compare_matpermy(a, py_ddl, &ret);
    if (ret != DDL_SUCCESS) {
      ierr = DDL_Error(ret, DDL_ERR_HARD, DDL_SETPERM_SPARSE);
      if (ierr > 0) return ierr; /* cleanup and exit with error */
    }
  }

  /* only packed rows supported at the moment */
  ierr = DDL_Format_sparse(a, &format);
  if (format != DDL_PCK_ROW) {
    ierr = DDL_Error(DDL_ERR_IFORMAT, DDL_ERR_HARD, DDL_SETPERM_SPARSE);
    return ierr;}

  ierr = DDL_Comm_sparse(a, &comm);
  ierr = DDL_Gsize_sparse(a, size);
  nr = size[0]; /* total number of rows */
  nc = size[1]; /* total number of columns */

  /* copy x permutation vector to perm structures */
  if (px_ddl != NULL) {
    ierr = DDL_Find_prop(a, DDL_PX, &p);
    if (ierr != DDL_SUCCESS) {
      ierr = DDL_Error(ierr, DDL_ERR_HARD, DDL_SETPERM_SPARSE);
      if (ierr > 0) return ierr; /* cleanup and exit with error */
    }
    if (p == NULL) {
      /* create DDL_PX property */
      ierr = DDL_Block(px_ddl, &block);
      ierr = DDL_Create_prop(a, DDL_PX, &p);
      ierr = DDLL_Create_vector(nr, MPI_INT, block, comm, &v);
      p->data = v; 
    }
    /* copy data */
    ierr = DDL_Get_vector(p, &px);
    ierr = DDL_Get_vector(px_ddl, &lpx);
    ierr = DDL_Size_vector(px_ddl, &n);
    for (i=0;i<n;i++)
      px[i] = lpx[i]; /* global C indices */
  }

  /* copy y permutation vector to perm structures */
  if (py_ddl != NULL) {
    ierr = DDL_Find_prop(a, DDL_PY, &p);
    if (ierr != DDL_SUCCESS) {
      ierr = DDL_Error(ierr, DDL_ERR_HARD, DDL_SETPERM_SPARSE);
      if (ierr > 0) return ierr; /* cleanup and exit with error */
    }
    if (p == NULL) {
      /* create DDL_PX property */
      ierr = DDL_Block(py_ddl, &block);
      ierr = DDL_Create_prop(a, DDL_PY, &p);
      ierr = DDLL_Create_vector(nc, MPI_INT, block, comm, &v);
      p->data = v;
    }
    /* copy data */
    ierr = DDL_Get_vector(p, &py);
    ierr = DDL_Get_vector(py_ddl, (void**)&lpy);
    ierr = DDL_Size_vector(py_ddl, &n);
    for (i=0;i<n;i++)
      py[i] = lpy[i]; /* global C indices */
  }

  /* calc inverse permutation vectors???? */

  return ierr;
}

/* calculate inverse permutation vector
   Parameters:
     pv_ddl: IN     original permutation vector
     ipv_ddl: OUT   inverse permutation vector
   Status: H1 */
int DDL_Invperm(DDL_Object pv_ddl, DDL_Object ipv_ddl)
{
  int ierr = DDL_SUCCESS;
  int *pv, *ipv, ret, *rcv_buf, *snd_buf;
  int *lpv;
  int i, n, nr, size, block, proc, nproc, offset;
  MPI_Comm comm;
  DDL_Prop p;
  DDL_Glob g;
  DDL_Object d;
  DDL_Vector p_ddl;
  int rcv_tot_data, snd_tot_data;
  int rcv_tot_proc, snd_tot_proc;

  /* check parameters */
  if (pv_ddl == NULL || ipv_ddl == NULL) {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_INVPERM);
    if (ierr > 0) return ierr; /* cleanup and exit with error */
  }
  /* check permutation vector */
  ierr = DDL_Check_perm(pv_ddl, &ret);
  if (ret != DDL_SUCCESS) {
    ierr = DDL_Error(ret, DDL_ERR_HARD, DDL_INVPERM);
    if (ierr > 0) return ierr; /* cleanup and exit with error */
  }
  /* check permutation vector */
  ierr = DDL_Check_perm(ipv_ddl, &ret);
  if (ret != DDL_SUCCESS) {
    ierr = DDL_Error(ret, DDL_ERR_HARD, DDL_INVPERM);
    if (ierr > 0) return ierr; /* cleanup and exit with error */
  }
  ierr = DDL_Compare_vecvec(pv_ddl, ipv_ddl, &ret);
  if (ret != DDL_SUCCESS) {
    ierr = DDL_Error(ret, DDL_ERR_HARD, DDL_INVPERM);
    if (ierr > 0) return ierr; /* cleanup and exit with error */
  }

  ierr = DDL_Get_vector(pv_ddl, &pv);
  ierr = DDL_Get_vector(ipv_ddl, &ipv);
  ierr = DDL_Size_vector(pv_ddl, &n);
  ierr = DDL_Gsize_vector(pv_ddl, &nr);
  ierr = DDL_Offset_vector(pv_ddl, &offset);
  ierr = DDL_Block_vector(pv_ddl, &block);
  ierr = DDL_Comm_vector(pv_ddl, &comm);
  ierr = MPI_Comm_size(comm, &nproc);
  ierr = MPI_Comm_rank(comm, &proc);

  /* create dummy object and communications structure */
  ierr = DDL_Create_dummy(&d);
  /* create COMM_PV global comms data structure for vector */
  /* create communication structure */
  ierr = DDL_Create_prop(d, DDL_COMM_PV, &p);
  g = (p->data);
  g->rcv.proc = malloc(nproc*sizeof(int));
  g->rcv.ind = malloc(n*sizeof(int));
  g->rcv.pind = malloc(n*sizeof(int));
  g->rcv.cnt = malloc(nproc*sizeof(int));
  g->rcv.ptr = malloc((nproc+1)*sizeof(int));
  g->snd.proc = malloc(nproc*sizeof(int));
  g->snd.ind = malloc(n*sizeof(int));
  g->snd.cnt = malloc(nproc*sizeof(int));
  g->snd.ptr = malloc((nproc+1)*sizeof(int));

  /* calculate row communications */

  /* find global C indices of elements I have */
  for (i=0;i<n;i++) g->rcv.ind[i] = i+offset;
  /* find permuted indices */
  for (i=0;i<n;i++) g->rcv.pind[i] = pv[i];

  rcv_tot_data = n;

  /* sort permuted index vector, but remeber original index in rcv_ind */
  DDL_qs2i(g->rcv.pind, g->rcv.ind, rcv_tot_data, 1);

  /* set up receive structure */
  ierr = DDL_Packcomm(nr, block, proc, nproc, &g->rcv.nproc, &rcv_tot_data, 
		      g->rcv.proc, g->rcv.ptr, g->rcv.pind);

  /* exchange elements */
  ierr = DDL_Excomm(nproc, comm, 
		    &g->rcv.nproc, &g->snd.nproc, 
		    &rcv_tot_data, &snd_tot_data, 
		    g->rcv.proc, g->snd.proc, 
		    g->rcv.ptr, g->snd.ptr, 
		    g->rcv.cnt, g->snd.cnt, 
		    g->rcv.pind, g->snd.ind);

  /* snd data structure now says which process holds which
     rows I should have if there was no permutation */
  /* rcv data structure says which logical rows I have */
  /* now find out the physical row numbers for the logical rows */
  /* create permuted indices output vector */

  /* allocate temporary buffers */
  rcv_buf = malloc(n*sizeof(int));
  snd_buf = malloc(n*sizeof(int));
  
  /* pack rcv_buf */
  for (i=0;i<n;i++)
    rcv_buf[i] = g->rcv.ind[i];
  /* do alltoall sending rcv structure */
  ierr = DDL_Alltoallvs(rcv_buf, g->rcv.nproc, 
			g->rcv.proc, 
			g->rcv.cnt,
			g->rcv.ptr, MPI_INT,
			snd_buf, g->snd.nproc, 
			g->snd.proc, 
			g->snd.cnt,
			g->snd.ptr, MPI_INT,
			comm);
  
  /* now snd_buf contains physical global indices of logical rows I should
     have */
 
  /* unpack snd_buf to ipv */
  for (i=0;i<n;i++)
    ipv[g->snd.ind[i]-offset] = snd_buf[i];

  DDL_Free_dummy(d);
  free(rcv_buf);
  free(snd_buf);

  return ierr;
}

/* sparse matrix vector multiply routines
   supports permutations for matrix
   matrix must currently be format DDL_PCK_ROW */

/* calculate communication pattern for sparse matrix-vector multiply
   Parameters:
     a: INOUT       sparse matrix
   Status: H1 */
int DDL_Smv_comm(DDL_Object a)
{
  int ierr = DDL_SUCCESS;
  int nproc, proc, dummy, offset, ret;
  int rcv_tot_data, snd_tot_data;
  int *ia, *ja, *py, *px, *tmp;
  int nr, nc, n, size[3], i, j, low, high, subtype;
  int format, block;
  MPI_Comm comm;
  DDL_Prop p;
  DDL_Glob g;

  /* check parameters */
  if (a == NULL) {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_SMV_COMM);
    if (ierr > 0) return ierr; /* cleanup and exit with error */
  }
  /* only packed rows supported at the moment */
  ierr = DDL_Format_sparse(a, &format);
  if (format != DDL_PCK_ROW) {
    ierr = DDL_Error(DDL_ERR_IFORMAT, DDL_ERR_HARD, DDL_SMV_COMM);
    return ierr;}

  ierr = DDL_Get_ia(a, &ia);
  ierr = DDL_Get_ja(a, &ja);
  ierr = DDL_Gsize_sparse(a, size);
  nr = size[0]; /* total number of rows */
  nc = size[1]; /* total number of columns */
  ierr = DDL_Size_sparse(a, size);
  n = size[0]; /* number of rows I have */
  ierr = DDL_Block_sparse(a, &block);
  ierr = DDL_Comm(a, &comm);
  ierr = MPI_Comm_size(comm, &nproc);
  ierr = MPI_Comm_rank(comm, &proc);

  /* create COMM_PY global comms data structure for matrix */
  ierr = DDL_Find_prop(a, DDL_COMM_PY, &p);
  if (ierr != DDL_SUCCESS) {
    ierr = DDL_Error(ierr, DDL_ERR_HARD, DDL_SMV_COMM);
    if (ierr > 0) return ierr; /* cleanup and exit with error */
  }
  if (p == NULL) {
    /* create communication structure */
    ierr = DDL_Create_prop(a, DDL_COMM_PY, &p);
    g = (p->data);
    g->rcv.proc = malloc(nproc*sizeof(int));
    g->rcv.ind = malloc(nc*sizeof(int));
    g->rcv.cnt = malloc(nproc*sizeof(int));
    g->rcv.ptr = malloc((nproc+1)*sizeof(int));
    g->snd.proc = malloc(nproc*sizeof(int));
    g->snd.ind = malloc(n*nproc*sizeof(int));
    g->snd.cnt = malloc(nproc*sizeof(int));
    g->snd.ptr = malloc((nproc+1)*sizeof(int));
  }
  else {
    /* structure already exists-just make ind vectors full size */
    g = (p->data);
    free(g->rcv.ind);
    free(g->snd.ind);
    g->rcv.ind = malloc(nc*sizeof(int));
    g->snd.ind = malloc(n*nproc*sizeof(int));
  }

  /* calculate column communications 
     ie which elements of result vector x I need */
  ierr = DDL_Find_prop(a, DDL_PY, &p);
  py = NULL;
  if (p != NULL) {
    ierr = DDL_Get_vector(p, &py);
    /* allocate space for permuted receive index vector */
    free(g->rcv.pind); /* in case there was an old vector */
    g->rcv.pind = malloc(nc*sizeof(int));
  }
  ierr = DDL_Col(ia, ja, n, nc, block, py,
		 proc, nproc, comm, 
		 &(g->rcv.nproc), &(g->snd.nproc), 
		 &rcv_tot_data, &snd_tot_data, 
		 g->rcv.proc, g->snd.proc, 
		 g->rcv.ptr, g->snd.ptr, 
		 g->rcv.cnt, g->snd.cnt, 
		 g->rcv.ind, g->snd.ind,
		 g->rcv.pind);

  /* now truncate rcv.ind, and snd.ind since these are *BIG* */
  ierr = DDL_Truncate(&(g->rcv.ind), rcv_tot_data);
  ierr = DDL_Truncate(&(g->snd.ind), snd_tot_data);
  if (py != NULL) ierr = DDL_Truncate(&(g->rcv.pind), rcv_tot_data);

  /* look at row swapping; 
     if no row permutation vector px then skip this */
  ierr = DDL_Find_prop(a, DDL_PX, &p);
  if (p != NULL) {
    ierr = DDL_Get_vector(p, &px);
    /* need COMM_PX row data structure for elements of RHS vector b */
    /* create COMM_PX comms data structure for matrix */
    ierr = DDL_Find_prop(a, DDL_COMM_PX, &p);
    if (ierr != DDL_SUCCESS) {
      ierr = DDL_Error(ierr, DDL_ERR_HARD, DDL_SMV_COMM);
      if (ierr > 0) return ierr; /* cleanup and exit with error */
    }
    if (p == NULL) {
      /* create communication structure */
      ierr = DDL_Create_prop(a, DDL_COMM_PX, &p);
      g = (p->data);
      g->rcv.proc = malloc(nproc*sizeof(int));
      g->rcv.ind = malloc(n*sizeof(int));
      g->rcv.pind = malloc(n*sizeof(int));
      g->rcv.cnt = malloc(nproc*sizeof(int));
      g->rcv.ptr = malloc((nproc+1)*sizeof(int));
      g->snd.proc = malloc(nproc*sizeof(int));
      g->snd.ind = malloc(n*nproc*sizeof(int));
      g->snd.cnt = malloc(nproc*sizeof(int));
      g->snd.ptr = malloc((nproc+1)*sizeof(int));
    }
    else {
      /* structure already exists-just make ind vectors full size */
      g = (p->data);
      free(g->rcv.ind);
      free(g->rcv.pind);
      free(g->snd.ind);
      g->rcv.ind = malloc(n*sizeof(int));
      g->rcv.pind = malloc(n*sizeof(int));
      g->snd.ind = malloc(n*nproc*sizeof(int));
    }

    /* calculate row communications 
       ie which elements of RHS vector b I need */
    ierr = DDL_Row(ia, ja, n, nr, block, px,
		   proc, nproc, comm, 
		   &(g->rcv.nproc), &(g->snd.nproc), 
		   &rcv_tot_data, &snd_tot_data, 
		   g->rcv.proc, g->snd.proc, 
		   g->rcv.ptr, g->snd.ptr, 
		   g->rcv.cnt, g->snd.cnt, 
		   g->rcv.ind, g->snd.ind,
		   g->rcv.pind);

    /* now truncate rcv.ind and snd.ind since these are *BIG* */
    ierr = DDL_Truncate(&(g->rcv.ind), rcv_tot_data);
    ierr = DDL_Truncate(&(g->snd.ind), snd_tot_data);
    ierr = DDL_Truncate(&(g->rcv.pind), rcv_tot_data);
  }
  return ierr;
}

/* sparse matrix, dense vector multiply
       y=alpha*Ax+beta*y
   Parameters:
     sp: IN         sparse matrix
     x: IN          input vector
     y: IN          result vector
     alpha: IN      scaling factor
     beta: IN       scaling factor
   Status: H1 */
int DDL_Smv(DDL_Object sp, DDL_Object x, DDL_Object y,
	    double alpha, double beta)
{
  int ierr, i, j, low, high, next, ret;
  int x_size, x_itype, x_offset, y_offset;
  MPI_Comm x_comm;
  MPI_Datatype x_type;
  void *x_data;
  int y_size, y_itype;
  MPI_Comm y_comm;
  MPI_Datatype y_type;
  void *y_data;
  int sp_size[3], sp_itype;
  MPI_Comm sp_comm;
  MPI_Datatype sp_type;
  int *ia, *ja;
  void *a, *tmp, *res, *ysnd_buf, *yrcv_buf, *snd_buf, *rcv_buf;
  int *snd_cnt, *rcv_cnt;
  int format, size;
  MPI_Aint typesize;
  int n, m, nelt;
  DDL_Prop p, p1;
  DDL_Glob gx, gy;

  ierr = DDL_SUCCESS;

  /* check input parameters */
  ierr = DDL_Gsize_vector(x, &x_size);
  ierr = DDL_Gsize_vector(y, &y_size);
  ierr = DDL_Gsize_sparse(sp, sp_size);
  m = sp_size[1];
  ierr = DDL_Type_vector(x, &x_type);
  ierr = DDL_Type_vector(y, &y_type);
  ierr = DDL_Type_sparse(sp, &sp_type);
  ierr = DDL_Comm_vector(x, &x_comm);
  ierr = DDL_Comm_vector(y, &y_comm);
  ierr = DDL_Comm_sparse(sp, &sp_comm);

  if ((x_type != y_type) || (x_type != sp_type)) {
    ierr = DDL_Error(DDL_ERR_ITYPE, DDL_ERR_HARD, DDL_SMV);
    return ierr;}
  if ((x_size != sp_size[1]) || (y_size != sp_size[0])) {
    ierr = DDL_Error(DDL_ERR_ISIZE, DDL_ERR_HARD, DDL_SMV);
    return ierr;}
  /* check communicators are **identical**
     same number of processes and ranking of processes */
  ierr = MPI_Comm_compare(x_comm, y_comm, &ret);
  if (ret != MPI_IDENT) {
    ierr = DDL_Error(DDL_ERR_ICOMM, DDL_ERR_HARD, DDL_SMV);
    return ierr;}
  ierr = MPI_Comm_compare(x_comm, sp_comm, &ret);
  if (ret != MPI_IDENT) {
    ierr = DDL_Error(DDL_ERR_ICOMM, DDL_ERR_HARD, DDL_SMV);
    return ierr;}

  ierr = DDL_Format_sparse(sp, &format);
  if (format != DDL_PCK_ROW) {
    ierr = DDL_Error(DDL_ERR_IFORMAT, DDL_ERR_HARD, DDL_SMV);
    return ierr;}

  /* get pointers to raw data values */
  ierr = DDL_Get_vector(x, &x_data);
  ierr = DDL_Get_vector(y, &y_data);
  ierr = DDL_Get_ia(sp, &ia);
  ierr = DDL_Get_ja(sp, &ja);
  ierr = DDL_Get_a(sp, &a);
  ierr = DDL_Size_vector(x, &x_size);
  ierr = DDL_Offset_vector(x, &x_offset);
  ierr = DDL_Offset_vector(y, &y_offset);
  ierr = DDL_Size_sparse(sp, sp_size);
  n = sp_size[0];
  ierr = MPI_Type_size(x_type, &typesize);

  /* check global comms. data exists, else create */
  ierr = DDL_Find_prop(sp, DDL_COMM_PY, &p);
  ierr = DDL_Find_prop(sp, DDL_COMM_PX, &p1);
  if (p == NULL && p1 == NULL)
    ierr = DDL_Smv_comm(sp);

  /* get global comms strcutures */
  ierr = DDL_Find_prop(sp, DDL_COMM_PY, &p);
  if (ierr != DDL_SUCCESS) {
    ierr = DDL_Error(ierr, DDL_ERR_HARD, DDL_SMV);
    if (ierr > 0) return ierr; /* cleanup and exit with error */
  }
  gy = (DDL_Glob)(p->data);
  ierr = DDL_Find_prop(sp, DDL_COMM_PX, &p);
  if (ierr != DDL_SUCCESS) {
    ierr = DDL_Error(ierr, DDL_ERR_HARD, DDL_SMV);
    if (ierr > 0) return ierr; /* cleanup and exit with error */
  }
  if (p != NULL)
    gx = (DDL_Glob)(p->data);
  else
    gx = NULL;

  /* temporary buffer for result vector */
  res = malloc(n*typesize);

  /* get elements of result y I need for permutation */
  if (gx != NULL) {
    /* allocate temporary buffers */
    size = gx->rcv.ptr[gx->rcv.nproc];
    yrcv_buf = malloc(size*typesize);
    size = gx->snd.ptr[gx->snd.nproc];
    ysnd_buf = malloc(size*typesize);

    /* pack ysnd_buf */
    for (i=0;i<gx->snd.ptr[gx->snd.nproc];i++)
      ((double*)ysnd_buf)[i] = ((double*)y_data)[gx->snd.ind[i]]; 
    /* do alltoall */
    ierr = DDL_Alltoallvs(ysnd_buf, gx->snd.nproc, 
			  gx->snd.proc, 
			  gx->snd.cnt,
			  gx->snd.ptr, x_type,
			  yrcv_buf, gx->rcv.nproc, 
			  gx->rcv.proc, 
			  gx->rcv.cnt,
			  gx->rcv.ptr, x_type,
			  sp_comm);
    
    /* unpack yrcv_buf to res */
    for (i=0;i<gx->rcv.ptr[gx->rcv.nproc];i++)
      ((double*)res)[gx->rcv.ind[i]-y_offset] = ((double*)yrcv_buf)[i]; 
  }
  else {
    /* gx == NULL just copy my elements of y */
    for (i=0;i<n;i++)
      ((double*)res)[i] = ((double*)y_data)[i];
  }

  /* now get elements of x I need */
  /* allocate temporary buffers */
  size = gy->rcv.ptr[gy->rcv.nproc];
  rcv_buf = malloc(size*typesize);
  size = gy->snd.ptr[gy->snd.nproc];
  snd_buf = malloc(size*typesize);
  tmp = malloc(m*typesize);

  /* pack snd_buf */
  for (i=0;i<gy->snd.ptr[gy->snd.nproc];i++)
    ((double*)snd_buf)[i] = ((double*)x_data)[gy->snd.ind[i]]; 
  
  /* do alltoall */
  ierr = DDL_Alltoallvs(snd_buf, gy->snd.nproc, 
			gy->snd.proc, 
			gy->snd.cnt,
			gy->snd.ptr, x_type,
			rcv_buf, gy->rcv.nproc, 
			gy->rcv.proc, 
			gy->rcv.cnt,
			gy->rcv.ptr, x_type,
			sp_comm);
  
  /* unpack rcv_buf to tmp */
  for (i=0;i<gy->rcv.ptr[gy->rcv.nproc];i++)
    ((double*)tmp)[gy->rcv.ind[i]] = ((double*)rcv_buf)[i]; 

  /* do matrix-vector multiply */
  ierr = DDL_Smv_seq(n, m, ia, ja, 
		     a, sp_type,
		     nelt, format, tmp, res,
		     alpha, beta);
  
  /* send back elements of result vector if permutation */
  /* opposite to fetching y values at start: 
     use rcv.ind to pack sending vector etc */
  if (gx != NULL) {
    /* pack yrcv_buf */
    for (i=0;i<gx->rcv.ptr[gx->rcv.nproc];i++)
      ((double*)yrcv_buf)[i] = ((double*)res)[gx->rcv.ind[i]-y_offset]; 

    /* do alltoall sending rcv structure */
    ierr = DDL_Alltoallvs(yrcv_buf, gx->rcv.nproc, 
			  gx->rcv.proc, 
			  gx->rcv.cnt,
			  gx->rcv.ptr, x_type,
			  ysnd_buf, gx->snd.nproc, 
			  gx->snd.proc, 
			  gx->snd.cnt,
			  gx->snd.ptr, x_type,
			  sp_comm);
    
    /* unpack ysnd_buf to y_data */
    for (i=0;i<gx->snd.ptr[gx->snd.nproc];i++)
      ((double*)y_data)[gx->snd.ind[i]] = ((double*)ysnd_buf)[i]; 
  }
  else {
    /* gx == NULL just copy my result elements to y */
    for (i=0;i<n;i++)
      ((double*)y_data)[i] = ((double*)res)[i];
  }

  /* free temporary arrays */
  if (gx != NULL) {
    free(ysnd_buf);
    free(yrcv_buf);
  }
  free(snd_buf);
  free(rcv_buf);
  free(res);
  free(tmp);

  return ierr;
}  

/* sparse matrix triangular solve routines
   supports permutations for matrix
   matrix must currently be format DDL_PCK_ROW
   matrix must also be triangular with subtype set to
   DDL_UPPER or DDL_LOWER */

/* calculate communication pattern for sparse triangular matrix solve
   Parameters:
     a: INOUT       sparse triangular matrix
   Status: H1 */
int DDL_Trisolve_comm(DDL_Object a)
{
  int ierr = DDL_SUCCESS;
  int nproc, proc, dummy, offset, ret;
  int rcv_tot_data, snd_tot_data;
  int *ia, *ja, *py, *px, *ipx, *perm, *tpind;
  int nr, nc, n, size[3], i, j, low, high, subtype;
  int format, block;
  MPI_Comm comm;
  DDL_Prop p;
  DDL_Glob g;
  DDL_Ind ind;
  DDL_Object d;
  DDL_Vector v;
  DDL_Vec pv, ipv;
  char *tmp;        

  /* check parameters */
  if (a == NULL) {
    ierr = DDL_Error(DDL_ERR_UNALLOCATED, DDL_ERR_HARD, DDL_TRISOLVE_COMM);
    if (ierr > 0) return ierr; /* cleanup and exit with error */
  }
  /* only packed rows supported at the moment */
  ierr = DDL_Format_sparse(a, &format);
  if (format != DDL_PCK_ROW) {
    ierr = DDL_Error(DDL_ERR_IFORMAT, DDL_ERR_HARD, DDL_TRISOLVE_COMM);
    return ierr;}

  ierr = DDL_Get_ia(a, &ia);
  ierr = DDL_Get_ja(a, &ja);
  ierr = DDL_Gsize_sparse(a, size);
  nr = size[0]; /* total number of rows */
  nc = size[1]; /* total number of columns */
  ierr = DDL_Size_sparse(a, size);
  n = size[0]; /* number of rows I have */
  ierr = DDL_Offset_sparse(a, &offset);
  ierr = DDL_Block_sparse(a, &block);
  ierr = DDL_Comm(a, &comm);
  ierr = MPI_Comm_size(comm, &nproc);
  ierr = MPI_Comm_rank(comm, &proc);

  /* look at row swapping; 
     if no row permutation vector px then skip this */
  ierr = DDL_Find_prop(a, DDL_PX, &p);
  px = NULL;
  if (p != NULL) {
    pv = p;
    ierr = DDL_Get_vector(p, &px);
    /* need COMM_PX row data structure for elements of RHS vector b */
    /* create COMM_PX comms data structure for matrix */
    ierr = DDL_Find_prop(a, DDL_COMM_PX, &p);
    if (ierr != DDL_SUCCESS) {
      ierr = DDL_Error(ierr, DDL_ERR_HARD, DDL_TRISOLVE_COMM);
      if (ierr > 0) return ierr; /* cleanup and exit with error */
    }
    if (p == NULL) {
      /* create communication structure */
      ierr = DDL_Create_prop(a, DDL_COMM_PX, &p);
      g = (p->data);
      g->rcv.proc = malloc(nproc*sizeof(int));
      g->rcv.ind = malloc(n*sizeof(int));
      g->rcv.pind = malloc(n*sizeof(int));
      g->rcv.cnt = malloc(nproc*sizeof(int));
      g->rcv.ptr = malloc((nproc+1)*sizeof(int));
      g->snd.proc = malloc(nproc*sizeof(int));
      g->snd.ind = malloc(n*nproc*sizeof(int));
      g->snd.cnt = malloc(nproc*sizeof(int));
      g->snd.ptr = malloc((nproc+1)*sizeof(int));
    }
    else {
      /* structure already exists-just make ind vectors full size */
      g = (p->data);
      free(g->rcv.ind);
      free(g->rcv.pind);
      free(g->snd.ind);
      g->rcv.ind = malloc(n*sizeof(int));
      g->rcv.pind = malloc(n*sizeof(int));
      g->snd.ind = malloc(n*nproc*sizeof(int));
    }

    /* calculate row communications 
       ie which elements of RHS vector b I need */
    ierr = DDL_Row(ia, ja, n, nr, block, px,
		   proc, nproc, comm, 
		   &(g->rcv.nproc), &(g->snd.nproc), 
		   &rcv_tot_data, &snd_tot_data, 
		   g->rcv.proc, g->snd.proc, 
		   g->rcv.ptr, g->snd.ptr, 
		   g->rcv.cnt, g->snd.cnt, 
		   g->rcv.ind, g->snd.ind,
		   g->rcv.pind);

    /* now truncate rcv.ind and snd.ind since these are *BIG* */
    ierr = DDL_Truncate(&(g->rcv.ind), rcv_tot_data);
    ierr = DDL_Truncate(&(g->snd.ind), snd_tot_data);
    ierr = DDL_Truncate(&(g->rcv.pind), rcv_tot_data);
  }

  /* column swapping */
  /* create COMM_IND global comms data structure for matrix */
  ierr = DDL_Find_prop(a, DDL_COMM_IND, &p);
  if (ierr != DDL_SUCCESS) {
    ierr = DDL_Error(ierr, DDL_ERR_HARD, DDL_TRISOLVE_COMM);
    if (ierr > 0) return ierr; /* cleanup and exit with error */
  }
  if (p == NULL) {
    /* create communication structure */
    ierr = DDL_Create_prop(a, DDL_COMM_IND, &p);
    ind = (p->data);
    ind->rcv.proc = malloc(nc*sizeof(int));
    ind->rcv.ind = malloc(nc*sizeof(int));
    ind->rcv.pind = NULL;
    ind->snd.proc = malloc(nproc*n*sizeof(int));
    ind->snd.ind = malloc(n*sizeof(int));
    ind->snd.cnt = malloc(n*sizeof(int));
    ind->snd.ptr = malloc((n+1)*sizeof(int));
  }
  else {
    /* structure already exists-just make ind vectors full size */
    ind = (p->data);
  }

  /* for now first calculate dummy DDL_COMM_PY structure then convert.. */
  /* use temporary dummy object for comms properties */
  /* or just malloc DDL_GLOB since new object doesn't hide much work? */
  /* or keep DDL_COMM_PY for matrix as well? */
  ierr = DDL_Create_dummy(&d);
  /* create COMM_PY global comms data structure for matrix */
  /* create communication structure */
  ierr = DDL_Create_prop(d, DDL_COMM_PY, &p);
  g = (p->data);
  g->rcv.proc = malloc(nproc*sizeof(int));
  g->rcv.ind = malloc(nc*sizeof(int));
  g->rcv.pind = NULL;
  g->rcv.cnt = malloc(nproc*sizeof(int));
  g->rcv.ptr = malloc((nproc+1)*sizeof(int));
  g->snd.proc = malloc(nproc*sizeof(int));
  g->snd.ind = malloc(n*nproc*sizeof(int));
  g->snd.cnt = malloc(nproc*sizeof(int));
  g->snd.ptr = malloc((nproc+1)*sizeof(int));

  /* calculate column communications 
     ie which elements of result vector x I need */
  ierr = DDL_Find_prop(a, DDL_PY, &p);
  py = NULL;
  if (p != NULL) {
    ierr = DDL_Get_vector(p, &py);
    /* allocate space for permuted receive index vector */
    free(g->rcv.pind); /* in case there was an old vector */
    g->rcv.pind = malloc(nc*sizeof(int));
    free(ind->rcv.pind);
    ind->rcv.pind = malloc(nc*sizeof(int));
  }

  /* calc which elements of x I will need */
  /* find non-zero columns */
  tmp = malloc(nc*sizeof(char));
  ierr = DDL_Findnz(ia, ja, n, nc, tmp);
  /* add elements of x I own and want at end of solve */
  /* or do this by just permuting solution vector after solve is complete */
  /* now pack as index vector in rcv.ind */
  ierr = DDL_Pack(tmp, g->rcv.ind, nc, &rcv_tot_data);
  free(tmp);
  
  ierr = DDL_Calccomm(ia, ja, n, nc, block, py,
		      proc, nproc, comm, 
		      &(g->rcv.nproc), &(g->snd.nproc), 
		      &rcv_tot_data, &snd_tot_data, 
		      g->rcv.proc, g->snd.proc, 
		      g->rcv.ptr, g->snd.ptr, 
		      g->rcv.cnt, g->snd.cnt, 
		      g->rcv.ind, g->snd.ind,
		      g->rcv.pind);

  /* this structure is now correct if the elements of the solution
     vector are calculated on their owners. If the rows are permuted then
     we must permute this structure again using the row permutation! */

  if (px != NULL) {
    tpind = malloc(rcv_tot_data*sizeof(int));
    perm = malloc(rcv_tot_data*sizeof(int));
    /* create inverse row permutation vector for matrix */
    ierr = DDL_Find_prop(a, DDL_IPX, &p);
    if (ierr != DDL_SUCCESS) {
      ierr = DDL_Error(ierr, DDL_ERR_HARD, DDL_TRISOLVE_COMM);
      if (ierr > 0) return ierr; /* cleanup and exit with error */
    }
    if (p == NULL) {
      /* create DDL_IPX property */
      ierr = DDL_Create_prop(a, DDL_IPX, &p);
      ierr = DDLL_Create_vector(nr, MPI_INT, block, comm, &v);
      p->data = v;
      ipv = p;
    }
    else 
      ipv = p;
    /* calculate inverse */
    ierr = DDL_Invperm(pv, ipv);
    ierr = DDL_Get_vector(ipv, &ipx);
    /* set up physical indices */
    for (i=0;i<snd_tot_data;i++) 
      g->snd.ind[i] = ipx[g->snd.ind[i]];

    /* send and receive blocks of permuted indices */
    ierr = DDL_Alltoallvs(g->snd.ind, g->snd.nproc, 
			  g->snd.proc, g->snd.cnt, g->snd.ptr, MPI_INT,
			  tpind, g->rcv.nproc,
			  g->rcv.proc, g->rcv.cnt, g->rcv.ptr, MPI_INT,
			  comm);
    /* tpind now contains permuted indices I want */
    /* sort permuted index vector, but keep mapping between
       original indices wanted and final indices wanted */
    for (i=0;i<rcv_tot_data;i++) perm[i] = i;
    DDL_qs2i(tpind, perm, rcv_tot_data, 1);
    ierr = DDL_Lpermute_vector(g->rcv.ind, MPI_INT, perm, rcv_tot_data);
    /* sort permuted indices if they exist */
    if (py != NULL) 
      ierr = DDL_Lpermute_vector(g->rcv.pind, MPI_INT, perm, rcv_tot_data);
    /* set up receive structure again */
    ierr = DDL_Packcomm(nc, block, proc, nproc, &g->rcv.nproc, &rcv_tot_data, 
			g->rcv.proc, g->rcv.ptr, tpind);
    /* exchange elements again */
    ierr = DDL_Excomm(nproc, comm, 
		      &g->rcv.nproc, &g->snd.nproc, 
		      &rcv_tot_data, &snd_tot_data, 
		      g->rcv.proc, g->snd.proc, 
		      g->rcv.ptr, g->snd.ptr, 
		      g->rcv.cnt, g->snd.cnt, 
		      tpind, g->snd.ind);
    /* structure now correct taking into account the row permutation */

    /* convert global indexes in snd_ind to local indexes */
    for (i=0;i<snd_tot_data;i++)
      g->snd.ind[i] = g->snd.ind[i]-offset;

    free(tpind);
    free(perm);
  }

  /* now reorder communication structure for sending individual indices */
  ierr = DDL_ReorderPI(&(g->snd.nproc), 
		       &snd_tot_data, 
		       g->snd.proc, 
		       g->snd.ptr, 
		       g->snd.cnt, 
		       g->snd.ind,
		       &(g->rcv.nproc), 
		       &rcv_tot_data,
		       g->rcv.proc,
		       g->rcv.ptr,
		       g->rcv.cnt,
		       g->rcv.ind,
		       g->rcv.pind,
		       &(ind->snd.nind),
		       ind->snd.ind, ind->snd.ptr,
		       ind->snd.cnt, ind->snd.proc,
		       &(ind->rcv.nind), 
		       ind->rcv.ind, ind->rcv.pind, ind->rcv.proc);

  /* release dummy object and associated data structure */
  ierr = DDL_Free(d);

  /* now truncate rcv.ind and snd.ind since these are *BIG* */

  return ierr;
}

/* perform sparse triangular solve
       Ax=b
   Parameters:
     sp: IN         sparse triangular matrix
     x: OUT         solution vector
     b: IN          RHS vector
   Status: H1 */
int DDL_Trisolve(DDL_Object sp, DDL_Object x, DDL_Object b)
{
  int ierr, i, j, k, low, high, next, ret;
  int x_size, x_itype, x_offset, b_offset;
  MPI_Comm x_comm;
  MPI_Datatype x_type;
  void *x_data;
  int b_size, b_itype;
  MPI_Comm b_comm;
  MPI_Datatype b_type;
  void *b_data;
  int sp_size[3], sp_itype;
  MPI_Comm sp_comm;
  MPI_Datatype sp_type, type;
  MPI_Status status;
  int *ia, *ja;
  void *a, *rhs, *sum, *sol, *bsnd_buf, *brcv_buf, *snd_buf, *rcv_buf;
  int *snd_cnt, *rcv_cnt, *px,  *py, *tmp;
  int *pind, *sxpi, *sypi, *sspi, *xind, *xpind, *ypind;
  int format, sformat, size, proc;
  MPI_Aint typesize;
  int n, m, nr, ni, nelt;
  DDL_Prop p, p1;
  DDL_Glob gx;
  DDL_Ind iy;
  double val, coeff;
  int li, lj, nextr, nextx, nexts;
        
  ierr = DDL_SUCCESS;

  /* check input parameters */
  ierr = DDL_Gsize_vector(x, &x_size);
  ierr = DDL_Gsize_vector(b, &b_size);
  ierr = DDL_Gsize_sparse(sp, sp_size);
  n = sp_size[0];
  m = sp_size[1];
  ierr = DDL_Type_vector(x, &x_type);
  ierr = DDL_Type_vector(b, &b_type);
  ierr = DDL_Type_sparse(sp, &sp_type);
  ierr = DDL_Comm_vector(x, &x_comm);
  ierr = DDL_Comm_vector(b, &b_comm);
  ierr = DDL_Comm_sparse(sp, &sp_comm);

  if ((x_type != b_type) || (x_type != sp_type)) {
    ierr = DDL_Error(DDL_ERR_ITYPE, DDL_ERR_HARD, DDL_TRISOLVE);
    return ierr;}
  if ((x_size != sp_size[1]) || (b_size != sp_size[0])) {
    ierr = DDL_Error(DDL_ERR_ISIZE, DDL_ERR_HARD, DDL_TRISOLVE);
    return ierr;}
  /* check communicators are **identical**
     same number of processes and ranking of processes */
  ierr = MPI_Comm_compare(x_comm, b_comm, &ret);
  if (ret != MPI_IDENT) {
    ierr = DDL_Error(DDL_ERR_ICOMM, DDL_ERR_HARD, DDL_TRISOLVE);
    return ierr;}
  ierr = MPI_Comm_compare(x_comm, sp_comm, &ret);
  if (ret != MPI_IDENT) {
    ierr = DDL_Error(DDL_ERR_ICOMM, DDL_ERR_HARD, DDL_TRISOLVE);
    return ierr;}

  ierr = DDL_Format_sparse(sp, &format);
  if (format != DDL_PCK_ROW) {
    ierr = DDL_Error(DDL_ERR_IFORMAT, DDL_ERR_HARD, DDL_TRISOLVE);
    return ierr;}

  ierr = DDL_Subtype_sparse(sp, &sformat);
  if (sformat != DDL_UPPER && sformat != DDL_LOWER) {
    ierr = DDL_Error(DDL_ERR_IFORMAT, DDL_ERR_HARD, DDL_TRISOLVE);
    return ierr;}

  /* get pointers to raw data values */
  ierr = DDL_Get_vector(x, &x_data);
  ierr = DDL_Get_vector(b, &b_data);
  ierr = DDL_Get_ia(sp, &ia);
  ierr = DDL_Get_ja(sp, &ja);
  ierr = DDL_Get_a(sp, &a);
  ierr = DDL_Size_vector(x, &x_size);
  ierr = DDL_Offset_vector(x, &x_offset);
  ierr = DDL_Offset_vector(b, &b_offset);
  ierr = DDL_Size_sparse(sp, sp_size);
  nr = sp_size[0];
  ierr = MPI_Type_size(x_type, &typesize);

  /* check global comms. data exists, else create */
  ierr = DDL_Find_prop(sp, DDL_COMM_PX, &p);
  ierr = DDL_Find_prop(sp, DDL_COMM_IND, &p1);
  if (p == NULL && p1 == NULL)
    ierr = DDL_Trisolve_comm(sp);

  /* get global comms structures */
  ierr = DDL_Find_prop(sp, DDL_COMM_PX, &p);
  if (ierr != DDL_SUCCESS) {
    ierr = DDL_Error(ierr, DDL_ERR_HARD, DDL_TRISOLVE);
    if (ierr > 0) return ierr; /* cleanup and exit with error */
  }
  if (p != NULL)
    gx = (DDL_Glob)(p->data);
  else
    gx = NULL;
  ierr = DDL_Find_prop(sp, DDL_COMM_IND, &p);
  if (ierr != DDL_SUCCESS) {
    ierr = DDL_Error(ierr, DDL_ERR_HARD, DDL_TRISOLVE);
    if (ierr > 0) return ierr; /* cleanup and exit with error */
  }
  iy = (DDL_Ind)(p->data);
  
  /* temporary buffer for result vector b */
  rhs = malloc(nr*typesize);
  /* temporary buffer for partial sums */
  sum = malloc(nr*typesize);
  if (type = MPI_DOUBLE)
    for (i=0;i<nr;i++)
      ((double*)sum)[i] = 0;
  else if (type = MPI_FLOAT)
    for (i=0;i<nr;i++)
      ((float*)sum)[i] = 0;

  /* get elements of result b I need for permutation */
  if (gx != NULL) {
    /* allocate temporary buffers */
    size = gx->rcv.ptr[gx->rcv.nproc];
    brcv_buf = malloc(size*typesize);
    size = gx->snd.ptr[gx->snd.nproc];
    bsnd_buf = malloc(size*typesize);

    /* pack bsnd_buf */
    for (i=0;i<gx->snd.ptr[gx->snd.nproc];i++)
      ((double*)bsnd_buf)[i] = ((double*)b_data)[gx->snd.ind[i]]; 
    /* do alltoall */
    ierr = DDL_Alltoallvs(bsnd_buf, gx->snd.nproc, 
			  gx->snd.proc, 
			  gx->snd.cnt,
			  gx->snd.ptr, x_type,
			  brcv_buf, gx->rcv.nproc, 
			  gx->rcv.proc, 
			  gx->rcv.cnt,
			  gx->rcv.ptr, x_type,
			  sp_comm);
    
    /* unpack brcv_buf to res */
    for (i=0;i<gx->rcv.ptr[gx->rcv.nproc];i++)
      ((double*)rhs)[gx->rcv.ind[i]-b_offset] = ((double*)brcv_buf)[i]; 
  }
  else {
    /* gx == NULL just copy my elements of b */
    for (i=0;i<nr;i++)
      ((double*)rhs)[i] = ((double*)b_data)[i];
  }

  type = x_type;

  /* we support either both x & y permutations or no permutations
     directly. Support for one permutation is implemented by
     creating temporary identity permutation for other dimension */
  if (gx != NULL || iy->rcv.pind != NULL) {
    if (gx == NULL && iy->rcv.pind != NULL) {
      /* create temporary gx->rcv.ind and gx->rcv.ind vectors */
      xpind = malloc(nr*sizeof(int));
      xind = xpind;
      for (i=0;i<nr;i++) xpind[i] = i+b_offset;
    }
    else {
      xpind = gx->rcv.pind;
      xind = gx->rcv.ind;
    }
    if (gx != NULL && iy->rcv.pind == NULL) {
      /* create temporary iy->rcv.pind vector */
      ypind = iy->rcv.ind;
    }
    else {
      ypind = iy->rcv.pind;
    }
    
    /* create sorted index lists if there is a permutation */

    if (sformat == DDL_LOWER) {
      /* ascending order lists */
      /* need mapping to sorted row permutation vector */
      sxpi = malloc(nr*sizeof(int));
      for (i=0;i<nr;i++) sxpi[i] = i;
      tmp = malloc(nr*sizeof(int));
      memcpy(tmp, xpind, nr*sizeof(int));
      DDL_qs2i(tmp, sxpi, nr, 1);
      free(tmp);
      /* need mapping to sorted column permutation vector */
      sypi = malloc(iy->rcv.nind*sizeof(int));
      for (i=0;i<iy->rcv.nind;i++) sypi[i] = i;
      tmp = malloc(iy->rcv.nind*sizeof(int));
      memcpy(tmp, ypind, iy->rcv.nind*sizeof(int));
      DDL_qs2i(tmp, sypi, iy->rcv.nind, 1);
      free(tmp);
      /* need mapping to sorted permuted solution snd.ind vector */
      sspi = malloc(iy->snd.nind*sizeof(int));
      for (i=0;i<iy->snd.nind;i++) sspi[i] = i;
      /* if there is a row permutation comms structure gx then permute */
      if (gx != NULL) {
	ierr = DDL_Find_prop(sp, DDL_PX, &p);
	ierr = DDL_Get_vector(p, &px);
	tmp = malloc(iy->snd.nind*sizeof(int));
	for (i=0; i<iy->snd.nind;i++) tmp[i] = px[iy->snd.ind[i]];
	DDL_qs2i(tmp, sspi, iy->snd.nind, 1);
	free(tmp);
      }
    }
    else { 
      /* descending order lists */
      /* need mapping to sorted row permutation vector */
      sxpi = malloc(nr*sizeof(int));
      for (i=0;i<nr;i++) sxpi[i] = i;
      tmp = malloc(nr*sizeof(int));
      memcpy(tmp, xpind, nr*sizeof(int));
      DDL_qs2i(tmp, sxpi, nr, -1);
      free(tmp);
      /* need mapping to sorted column permutation vector */
      sypi = malloc(iy->rcv.nind*sizeof(int));
      for (i=0;i<iy->rcv.nind;i++) sypi[i] = i;
      tmp = malloc(iy->rcv.nind*sizeof(int));
      memcpy(tmp, ypind, iy->rcv.nind*sizeof(int));
      DDL_qs2i(tmp, sypi, iy->rcv.nind, -1);
      free(tmp);
      /* need mapping to sorted permuted solution snd.ind vector */
      sspi = malloc(iy->snd.nind*sizeof(int));
      if (gx == NULL) {
	for (i=0;i<iy->snd.nind;i++) sspi[i] = iy->snd.nind-1-i;
      }
      else {
	/* if there is a row permutation comms structure gx then permute */
	for (i=0;i<iy->snd.nind;i++) sspi[i] = i;
	ierr = DDL_Find_prop(sp, DDL_PX, &p);
	ierr = DDL_Get_vector(p, &px);
	tmp = malloc(iy->snd.nind*sizeof(int));
	for (i=0; i<iy->snd.nind;i++) tmp[i] = px[iy->snd.ind[i]];
	DDL_qs2i(tmp, sspi, iy->snd.nind, -1);
	free(tmp);
      }
    }
  }

  /* main loop to solve equations */
  /* assume permuted matrix is triangular as given by subformat */
  if (sformat == DDL_LOWER) {
    /* lower triangular solve */
    if (gx != NULL || iy->rcv.pind != NULL) {
      /* with permutations */
      nextx = 0; /* index into sxpi of my next row */
      nextr = 0; /* index into sypi of next element to receive */ 
      nexts = 0; /* index of next element to send */ 
      for (k=0;k<n;k++) {
	if (k == xpind[sxpi[nextx]]) {
	  /* I own next element to calculate */
	  li = xind[sxpi[nextx]]-b_offset; /* local C row in stored matrix */
	  /* find local column index to matrix for diagonal element */
	  lj = iy->rcv.ind[sypi[nextr]];
	  ierr = DDL_Get_coeff(ia,ja,a,type,format, li,lj, &coeff); 
	  /* calculate new result value */
	  if (type = MPI_DOUBLE)
	    val = 
	      (((double*)rhs)[li]-((double*)sum)[li])/(double)coeff; 
	  else if (type = MPI_FLOAT)
	    val = 
	      (((float*)rhs)[li]-((float*)sum)[li])/(float)coeff; 
	  /* send solution element to other processes */
	  /* better in blocks of blocksize elements? */
	  ierr = DDL_Msend(&val, 1, MPI_DOUBLE,
			   iy->snd.cnt[sspi[nexts]],
			   &iy->snd.proc[iy->snd.ptr[sspi[nexts]]], 0,
			   x_comm);
	  /* increment nextx and check for no more rows */
	  nextx = nextx+1;
	  if (nextx == nr) nextx = 0;
	  /* increment nexts */
	  nexts = nexts+1;
	  /* keep elements I calculate */
	  if (type = MPI_DOUBLE)
	    ((double*)x_data)[li] = val;
	  else if (type = MPI_FLOAT)
	    ((float*)x_data)[li] = val;
	}
	/* do I want diagonal element and haven't got it already? */
	else if (k == ypind[sypi[nextr]]) {
	  /* find who owns this row */
	  proc = iy->rcv.proc[sypi[nextr]];
	  /* do broadcast/receive of element */
	  /* better in blocks of blocksize elements? */
	  /* receive if not mine already */
	  ierr = MPI_Recv(&val, 1, MPI_DOUBLE, proc, 0,
			  x_comm, &status);
	}
	/* if I have new solution element update partial sums */
	if (k == ypind[sypi[nextr]]) {
	  /* find local column index to matrix for this element */
	  lj = iy->rcv.ind[sypi[nextr]];
	  /* now update partial sums for column permuted matrix */
	  for (i=nr-1;i>=0;i--) {
	    /* only try to update if row is greater than k */
	    if (xpind[sxpi[i]] > k) {
	      li = xind[sxpi[i]]-b_offset; /* local C row in stored matrix */
	      ierr = DDL_Get_coeff(ia,ja,a,type,format, li,lj, &coeff); 
	      if (ierr == DDL_SUCCESS) {
		/* update this row partial sum */
		if (type = MPI_DOUBLE)
		  ((double*)sum)[li] = ((double*)sum)[li]+
		    (double)coeff*(double)val;
		else if (type = MPI_FLOAT)
		  ((float*)sum)[li] = ((float*)sum)[li]+
		    (float)coeff*(float)val;
	      }
	    }
	    else break;
	  }
	  /* increment nextr and check for no more columns */
	  nextr = nextr+1;
	  if (nextr == iy->rcv.nind) nextr = 0;
	}
      }
    }
    else if (gx == NULL && iy->rcv.pind == NULL) {
      /* no permutation */
      nextx = 0; /* my next row */
      nextr = 0; /* next element to receive */ 
      nexts = 0; /* next element to send */ 
      for (k=0;k<n;k++) {
	if (k == nextx+b_offset) {
	  /* I own next element to calculate */
	  li = nextx; /* local C row in stored matrix */
	  /* find local column index to matrix for diagonal element */
	  lj = k;
	  ierr = DDL_Get_coeff(ia,ja,a,type,format, li,lj, &coeff); 
	  /* calculate new result value */
	  if (type = MPI_DOUBLE)
	    val = 
	      (((double*)rhs)[li]-((double*)sum)[li])/(double)coeff; 
	  else if (type = MPI_FLOAT)
	    val = 
	      (((float*)rhs)[li]-((float*)sum)[li])/(float)coeff; 
	  /* send solution element to other processes */
	  /* better in blocks of blocksize elements? */
	  ierr = DDL_Msend(&val, 1, MPI_DOUBLE,
			   iy->snd.cnt[nexts],
			   &iy->snd.proc[iy->snd.ptr[nexts]], 0,
			   x_comm);
	  /* increment nextx and check for no more rows */
	  nextx = nextx+1;
	  if (nextx == nr) nextx = 0;
	  /* increment nexts */
	  nexts = nexts+1;
	  /* keep elements I calculate */
	  if (type = MPI_DOUBLE)
	    ((double*)x_data)[li] = val;
	  else if (type = MPI_FLOAT)
	    ((float*)x_data)[li] = val;
	}
	/* do I want diagonal element and haven't got it already? */
	else if (k == iy->rcv.ind[nextr]) {
	  /* find who owns this row */
	  proc = iy->rcv.proc[nextr];
	  /* do broadcast/receive of element */
	  /* better in blocks of blocksize elements? */
	  /* receive if not mine already */
	  ierr = MPI_Recv(&val, 1, MPI_DOUBLE, proc, 0,
			  x_comm, &status);
	}
	/* if I have new solution element update partial sums */
	if (k == iy->rcv.ind[nextr]) {
	  /* find local column index to matrix for this element */
	  lj = k;
	  /* now update partial sums for column permuted matrix */
	  for (i=nr-1;i>=0;i--) {
	    /* only try to update if row is greater than k */
	    if (i+b_offset > k) {
	      li = i; /* local C row in stored matrix */
	      ierr = DDL_Get_coeff(ia,ja,a,type,format, li,lj, &coeff); 
	      if (ierr == DDL_SUCCESS) {
		/* update this row partial sum */
		if (type = MPI_DOUBLE)
		  ((double*)sum)[li] = ((double*)sum)[li]+
		    (double)coeff*(double)val;
		else if (type = MPI_FLOAT)
		  ((float*)sum)[li] = ((float*)sum)[li]+
		    (float)coeff*(float)val;
	      }
	    }
	    else break;
	  }
	  /* increment nextr and check for no more columns */
	  nextr = nextr+1;
	  if (nextr == iy->rcv.nind) nextr = 0;
	}
      }
    }
  }
  else if (sformat == DDL_UPPER) {
    /* upper triangular solve */
    /* same as lower triangular except k counts backwards
       and sorted index lists are reversed */
    if (gx != NULL || iy->rcv.pind != NULL) {
      /* with permutations */
      nextx = 0; /* index into sxpi of my next row */
      nextr = 0; /* index into sypi of next element to receive */ 
      nexts = 0; /* index of next element to send */ 
      for (k=n-1;k>=0;k--) {
	if (k == xpind[sxpi[nextx]]) {
	  /* I own next element to calculate */
	  li = xind[sxpi[nextx]]-b_offset; /* local C row in stored matrix */
	  /* find local column index to matrix for diagonal element */
	  lj = iy->rcv.ind[sypi[nextr]];
	  ierr = DDL_Get_coeff(ia,ja,a,type,format, li,lj, &coeff); 
	  /* calculate new result value */
	  if (type = MPI_DOUBLE)
	    val = 
	      (((double*)rhs)[li]-((double*)sum)[li])/(double)coeff; 
	  else if (type = MPI_FLOAT)
	    val = 
	      (((float*)rhs)[li]-((float*)sum)[li])/(float)coeff; 
	  /* send solution element to other processes */
	  /* better in blocks of blocksize elements? */
	  ierr = DDL_Msend(&val, 1, MPI_DOUBLE,
			   iy->snd.cnt[sspi[nexts]],
			   &iy->snd.proc[iy->snd.ptr[sspi[nexts]]], 0,
			   x_comm);
	  /* increment nextx and check for no more rows */
	  nextx = nextx+1;
	  if (nextx == nr) nextx = 0;
	  /* increment nexts */
	  nexts = nexts+1;
	  /* keep elements I calculate */
	  if (type = MPI_DOUBLE)
	    ((double*)x_data)[li] = val;
	  else if (type = MPI_FLOAT)
	    ((float*)x_data)[li] = val;
	}
	/* do I want diagonal element and haven't got it already? */
	else if (k == ypind[sypi[nextr]]) {
	  /* find who owns this row */
	  proc = iy->rcv.proc[sypi[nextr]];
	  /* do broadcast/receive of element */
	  /* better in blocks of blocksize elements? */
	  /* receive if not mine already */
	  ierr = MPI_Recv(&val, 1, MPI_DOUBLE, proc, 0,
			  x_comm, &status);
	}
	/* if I have new solution element update partial sums */
	if (k == ypind[sypi[nextr]]) {
	  /* find local column index to matrix for this element */
	  lj = iy->rcv.ind[sypi[nextr]];
	  /* now update partial sums for column permuted matrix */
	  for (i=nr-1;i>=0;i--) {
	    /* only try to update if row is less than k */
	    if (xpind[sxpi[i]] < k) {
	      li = xind[sxpi[i]]-b_offset; /* local C row in stored matrix */
	      ierr = DDL_Get_coeff(ia,ja,a,type,format, li,lj, &coeff); 
	      if (ierr == DDL_SUCCESS) {
		/* update this row partial sum */
		if (type = MPI_DOUBLE)
		  ((double*)sum)[li] = ((double*)sum)[li]+
		    (double)coeff*(double)val;
		else if (type = MPI_FLOAT)
		  ((float*)sum)[li] = ((float*)sum)[li]+
		    (float)coeff*(float)val;
	      }
	    }
	    else break;
	  }
	  /* increment nextr and check for no more columns */
	  nextr = nextr+1;
	  if (nextr == iy->rcv.nind) nextr = 0;
	}
      }
    }
    else if (gx == NULL && iy->rcv.pind == NULL) {
      /* no permutation */
      nextx = nr-1; /* my next row */
      nextr = iy->rcv.nind-1; /* next element to receive */ 
      nexts = nr-1; /* next element to send */ 
      for (k=n-1;k>=0;k--) {
	if (k == nextx+b_offset) {
	  /* I own next element to calculate */
	  li = nextx; /* local C row in stored matrix */
	  /* find local column index to matrix for diagonal element */
	  lj = k;
	  ierr = DDL_Get_coeff(ia,ja,a,type,format, li,lj, &coeff); 
	  /* calculate new result value */
	  if (type = MPI_DOUBLE)
	    val = 
	      (((double*)rhs)[li]-((double*)sum)[li])/(double)coeff; 
	  else if (type = MPI_FLOAT)
	    val = 
	      (((float*)rhs)[li]-((float*)sum)[li])/(float)coeff; 
	  /* send solution element to other processes */
	  /* better in blocks of blocksize elements? */
	  ierr = DDL_Msend(&val, 1, MPI_DOUBLE,
			   iy->snd.cnt[nexts],
			   &iy->snd.proc[iy->snd.ptr[nexts]], 0,
			   x_comm);
	  /* increment nextx and check for no more rows */
	  nextx = nextx-1;
	  if (nextx < 0) nextx = nr;
	  /* increment nexts */
	  nexts = nexts-1;
	  /* keep elements I calculate */
	  if (type = MPI_DOUBLE)
	    ((double*)x_data)[li] = val;
	  else if (type = MPI_FLOAT)
	    ((float*)x_data)[li] = val;
	}
	/* do I want diagonal element and haven't got it already? */
	else if (k == iy->rcv.ind[nextr]) {
	  /* find who owns this row */
	  proc = iy->rcv.proc[nextr];
	  /* do broadcast/receive of element */
	  /* better in blocks of blocksize elements? */
	  /* receive if not mine already */
	  ierr = MPI_Recv(&val, 1, MPI_DOUBLE, proc, 0,
			  x_comm, &status);
	}
	/* if I have new solution element update partial sums */
	if (k == iy->rcv.ind[nextr]) {
	  /* find local column index to matrix for this element */
	  lj = k;
	  /* now update partial sums for column permuted matrix */
	  for (i=0;i<nr;i++) {
	    /* only try to update if row is less than k */
	    if (i+b_offset < k) {
	      li = i; /* local C row in stored matrix */
	      ierr = DDL_Get_coeff(ia,ja,a,type,format, li,lj, &coeff); 
	      if (ierr == DDL_SUCCESS) {
		/* update this row partial sum */
		if (type = MPI_DOUBLE)
		  ((double*)sum)[li] = ((double*)sum)[li]+
		    (double)coeff*(double)val;
		else if (type = MPI_FLOAT)
		  ((float*)sum)[li] = ((float*)sum)[li]+
		    (float)coeff*(float)val;
	      }
	    }
	    else break;
	  }
	  /* increment nextr and check for no more columns */
	  nextr = nextr-1;
	  if (nextr < 0) nextr = iy->rcv.nind-1;
	}
      }
    }
  } 

  /* exchange solution vector x elements */
  /* opposite to fetching b values at start: 
     use rcv.ind to pack sending vector etc */
  if (gx != NULL) {
    /* pack brcv_buf */
    for (i=0;i<gx->rcv.ptr[gx->rcv.nproc];i++)
      ((double*)brcv_buf)[i] = ((double*)x_data)[gx->rcv.ind[i]-b_offset]; 

    /* do alltoall sending rcv structure */
    ierr = DDL_Alltoallvs(brcv_buf, gx->rcv.nproc, 
			  gx->rcv.proc, 
			  gx->rcv.cnt,
			  gx->rcv.ptr, x_type,
			  bsnd_buf, gx->snd.nproc, 
			  gx->snd.proc, 
			  gx->snd.cnt,
			  gx->snd.ptr, x_type,
			  x_comm);
    
    /* unpack ysnd_buf to y_data */
    for (i=0;i<gx->snd.ptr[gx->snd.nproc];i++)
      ((double*)x_data)[gx->snd.ind[i]] = ((double*)bsnd_buf)[i]; 
  }

  /* free temporary arrays */
  free(rhs);
  free(sum);
  if (gx != NULL) {
    free(bsnd_buf);
    free(brcv_buf);
  }
  if (gx == NULL && iy->rcv.pind != NULL) {
    /* free temporary gx->rcv.ind and gx->rcv.ind vectors */
    free(xpind);
  }
  /* free sorted index lists if there is a permutation */
  if (gx != NULL || iy->rcv.pind != NULL) {
    free(sxpi);
    free(sypi);
    free(sspi);
  }

  return ierr;
}  

/* take communication structure grouped by process numbers (eg. DDL_COMM_PX)
   and reorder into groups by element index number (eg. DDL_COMM_IND)

   useful when communications are of individual elements not of groups of
   elements, eg, for sending result elements in triangular solve

   Parameters:
     psnd_tot_proc: IN number of procs src must send data to
     psnd_tot_data: IN number of elements of x vector that src must send
     psnd: IN          vector of process numbers that src must send data to
     psnd_ptr: IN      index of first element in snd_ind that src must send
                       to each process in snd
     psnd_cnt: IN      number of elements to send to each process    
     psnd_ind: IN      vector of global indices of elements of x that src
                       must send to each process in snd
     prcv_tot_proc: IN number of procs src must get data from
     prcv_tot_data: IN number of elements of x vector that src must get
     prcv: IN          vector of process numbers that src must get data from
     prcv_ptr: IN      index of first element in rcv_ind that src must get
                       from each process in rcv
     prcv_cnt: IN      number of elements to receive from each process    
     prcv_ind: IN      vector of global indices of elements of x that src
                       must get from each process in rcv
     prcv_pind: IN     permuted version of prcv_ind index vector

     isnd_nind: OUT    number of indices process must send        
     isnd_ind: OUT     vector of indices to send
     isnd_ptr: OUT     index of first element in isnd_ind that src must send
                       to each process in isnd_proc
     isnd_cnt: OUT     number of elements to send to each process    
     isnd_proc: OUT    vector of process numbers that src must send data to
     ircv_nind: OUT    number of indices process must receive     
     ircv_ind: OUT     vector of indices to receive
     ircv_pind: OUT    permuted version of ircv_ind index vector
     ircv_proc: OUT    vector of process numbers that src receives data from
   Status: L3 */
int DDL_ReorderPI(int *psnd_tot_proc, 
		  int *psnd_tot_data, 
		  int psnd[], 
		  int psnd_ptr[], 
		  int psnd_cnt[], 
		  int psnd_ind[],
		  int *prcv_tot_proc,
		  int *prcv_tot_data,
		  int prcv[],
		  int prcv_ptr[],
		  int prcv_cnt[],
		  int prcv_ind[], int prcv_pind[],
		  int *isnd_nind,
		  int isnd_ind[], int isnd_ptr[],
		  int isnd_cnt[], int isnd_proc[],
		  int *ircv_nind, int ircv_ind[], 
		  int ircv_pind[], int ircv_proc[])
{
  int ierr = DDL_SUCCESS;
  int *tmp, *t2, *t3, i, j, p, cnt, ind;
  /* reorder send data structure */
  tmp = malloc(*psnd_tot_data*sizeof(int));
  /* expand process list */
  p = 0;
  for (i=0;i<*psnd_tot_data;i++) {
    if (i >= psnd_ptr[p+1]) p = p+1;
    isnd_proc[i] = psnd[p]; 
  }
  memcpy(tmp, psnd_ind, *psnd_tot_data*sizeof(int));
  DDL_qs2i(tmp, isnd_proc, *psnd_tot_data, 1); /* sort indices in t2 */
  /* compact again grouping by index */
  i = 0; /* index vector ptr */
  cnt = 1;
  ind = tmp[0];
  isnd_ptr[i] = 0;
  for (j=1;j<*psnd_tot_data;j++) {
    if (tmp[j] != ind) {
      isnd_cnt[i] = cnt; 
      isnd_ind[i] = ind;
      i = i+1;
      isnd_ptr[i] = j;
      ind = tmp[j];
      cnt = 0;
    }
    cnt = cnt+1;
  }
  /* finish off last entry */
  isnd_cnt[i] = cnt;
  isnd_ind[i] = ind;
  isnd_ptr[i+1] = j;
  *isnd_nind = i+1;
  free(tmp);

  /* now sort process blocks per index as well */
  for (i=0;i<*isnd_nind;i++)
    DDL_qs1i(&isnd_proc[isnd_ptr[i]], isnd_cnt[i], 1);

  /* reorder receive data structure */
  *ircv_nind = *prcv_tot_data;
  memcpy(ircv_ind, prcv_ind, *prcv_tot_data*sizeof(int));
  /* expand process list */
  p = 0;
  for (i=0;i<*prcv_tot_data;i++) {
    if (i >= prcv_ptr[p+1]) p = p+1;
    ircv_proc[i] = prcv[p]; 
  }

  tmp = malloc(*prcv_tot_data*sizeof(int));
  for (i=0;i<*prcv_tot_data;i++) tmp[i]=i;
  DDL_qs2i(ircv_ind, tmp, *prcv_tot_data, 1);
  /* permute ircv_proc */
  ierr = DDL_Lpermute_vector(ircv_proc, MPI_INT, tmp, *prcv_tot_data);
  if (prcv_pind != NULL) {
    /* convert permuted index vector as well */
    memcpy(ircv_pind, prcv_pind, *prcv_tot_data*sizeof(int));
    ierr = DDL_Lpermute_vector(ircv_pind, MPI_INT, tmp, *prcv_tot_data);
  }
  free(tmp);
}

/* find columns with a non-zero coefficient in them 
   from my block of sparse matrix 
   matrix must be DDL_PCK_ROW currently
   Parameters:
     ia,ja,n: IN    sparse matrix; n is number of rows I have locally
     nc : IN        number of columns in global/local matrix
     ind: OUT       vector of non-zero column (global) indices in order
     nz: OUT        number of non-zero columns 
   Status: L2 */
int DDL_Sparse_nzcol(int *ia, int *ja, int n, int nc, 
		     int *ind, int *nz)
{
  int i, j, low, high, ierr = DDL_SUCCESS;
  char *tmp;
  tmp = malloc(nc*sizeof(char));
  ierr = DDL_Findnz(ia, ja, n, nc, tmp);
  ierr = DDL_Pack(tmp, ind, nc, nz);
  free(tmp);
  return ierr;
}

/* find non-zero columns for my block of sparse matrix 
   matrix must be DDL_PCK_ROW currently
   Parameters:
     ia,ja,n: IN    sparse matrix; n is number of rows I have locally
     nc : IN        number of columns in global/local matrix
     sparse: OUT    sparsity vector; each element corresponds to one
                    column of the matrix; 1=has nonzero; 0=empty
   Status: L2 */   
int DDL_Findnz(int *ia, int *ja, int n, int nc, char *sparse)
{
  int i, j, low, high;
  for (i=0;i<nc;i++) sparse[i] = 0;
  /* find non-zero columns */
  for (i=0;i<n;i++) {
    DDL_Coeff_ind(i, ia, &low, &high);
    for (j=low-1;j<high-1;j++) {
      /* column indices in ja are global when stored by rows */
      sparse[ja[j]-1] = 1;
    }
  }
  return DDL_SUCCESS;
}

/* pack sparse vector template 
   Parameters:
     sparse: IN     sparsity vector; each element corresponds to one
                    column of the matrix; 1=has nonzero; 0=empty
     packed: OUT    packed vector of column indices to keep
     n: IN          size of sparse template
     nz: OUT        number of elements to keep
   Status: L2 */
int DDL_Pack(char *sparse, int *packed, int n, int *nz)
{
  int i, j;
  /* pack non-zero columns into ind vector as C indices */
  j = 0;
  for (i=0;i<n;i++) {
    if (sparse[i] != 0) {
      packed[j] = i;
      j=j+1;
    }
  }
  *nz = j; 
  return DDL_SUCCESS;
}

/* pack matrix non-zero columns into individual messages for each process
   ie. tell process owning row i of vector that I need it.
   (can also be used to form matrix row permutation comms)

   Parameters:
     n: IN          size of matrix (column or row size)
     block: IN      block size used for distribution of object
     dest: IN       destination process for data (ie me)
     nproc: IN      total number of processes distributed object uses
     tot_proc: OUT  number of processes desc must get data from
     tot_data: IN   total number of elements of x vector that I must get
     src: OUT       vector of process numbers that I must communicate with
     ptr: OUT       index of first element in ind that dest must get
                    from process src
     ind: IN        contains *global* indices of elements of x 
                    that I must get from each process in src
   Status: L2 */
int DDL_Packcomm(int n, int block,
		 int dest, int nproc, int *tot_proc, int *tot_data, 
		 int *src, int *ptr, int *ind)
{
  int ierr = DDL_SUCCESS;
  int size, cnt, i, j, low, high, np, proc;
  *tot_proc = 0;
  proc = 0;
  /* get range of vector elements held by process proc */ 
  low = DDL_Find_offset(n, block, nproc, proc);
  high = DDL_Find_offset(n, block, nproc, proc+1);
  cnt = 0;
  np = 0;
  ptr[0] = 0;
  for (i=0;i<*tot_data;i++) {
    /* find process holding row ind[i] */ 
    if (ind[i] >= high) {
      /* record count for previous process */
      if (cnt != 0) {
	src[np] = proc; /* source process number */
	ptr[np+1] = ptr[np]+cnt; /* pointer to first element from proc */
	cnt = 0;
	np = np+1;
      }
      while (ind[i] >= high) {
	proc = proc+1;
	low = high;
	high = DDL_Find_offset(n, block, nproc, proc+1);
      }
    }
    cnt = cnt+1;
  }
  /* record count for last process */
  if (cnt != 0) {
    src[np] = proc; /* source process number */
    ptr[np+1] = ptr[np]+cnt; /* pointer to first element from proc */
    np = np+1;
  }
  *tot_proc = np;
}

/* exchange index elements with other processes to complete 
   send communication structures DDL_COMM_PX or DDL_COMM_PY
   for sparse matrix vector multiply and triangular solve 

   Parameters:
     nproc: IN      total number of processes distributed object uses
     comm: IN       communicator used by object
     rcv_tot_proc: IN number of procs src must get data from
     snd_tot_proc: OUT number of procs src must send data to
     rcv_tot_data: IN total number of elements of x vector that src must get
     snd_tot_data: OUT total number of elements of x vector that src must send
     rcv: IN        vector of process numbers that src must get data from
     snd: OUT       vector of process numbers that src must send data to
     rcv_ptr: IN    index of first element in rcv_ind that src must get 
                    from rcv
     snd_ptr: OUT   index of first element in snd_ind that src must send to snd
     rcv_cnt: OUT   number of elements from each process    
     snd_cnt: OUT   number of elements to each process    
     rcv_ind: IN    vector contains global C indices of elements of x that src 
                    must get from each process in rcv
     snd_ind: OUT   vector contains global C indices of elements of x that src 
                    must send to each process in snd
   Status: L2 */
int DDL_Excomm(int nproc, MPI_Comm comm, 
	       int *rcv_tot_proc, int *snd_tot_proc, 
	       int *rcv_tot_data, int *snd_tot_data, 
	       int rcv[], int snd[], 
	       int rcv_ptr[], int snd_ptr[], 
	       int rcv_cnt[], int snd_cnt[], 
	       int rcv_ind[], int snd_ind[])
{
  int cnt, i, next, next_proc, ierr;
  ierr = DDL_SUCCESS;

  /* initialize counts of elements to receive from each process */ 
  for (i=0;i<nproc;i++)
    rcv_cnt[i] = 0;
  for (i=0;i<*rcv_tot_proc;i++)
    rcv_cnt[rcv[i]] = rcv_ptr[i+1]-rcv_ptr[i];
  /* everyone sends messages to everyone else saying how many
     elements they need from each process */
  MPI_Alltoall(rcv_cnt, 1, MPI_INT, snd_cnt, 1, MPI_INT, comm);

  /* set up arrays snd, snd_ptr */
  next_proc = 0;
  next = 0;
  for (i=0;i<nproc;i++) {
    cnt = snd_cnt[i];
    if (cnt != 0) {
      snd[next_proc] = i; /* remember process number */
      snd_ptr[next_proc] = next;
      next_proc = next_proc+1;
      next = next+cnt;
    }
  }
  snd_ptr[next_proc] = next;
  *snd_tot_proc = next_proc;
  *snd_tot_data = next;

  /* set up count vectors for DDL_Alltoallvs */
  for (i=0;i<*rcv_tot_proc;i++)
    rcv_cnt[i] = rcv_ptr[i+1]-rcv_ptr[i];
  for (i=0;i<*snd_tot_proc;i++)
    snd_cnt[i] = snd_ptr[i+1]-snd_ptr[i];

  /* exchange index vectors to complete snd_ind */
  ierr = DDL_Alltoallvs(rcv_ind, *rcv_tot_proc, 
			rcv, rcv_cnt, rcv_ptr, MPI_INT,
			snd_ind, *snd_tot_proc,
			snd, snd_cnt, snd_ptr, MPI_INT,
			comm);

  return ierr;
}

/* find the elements a process must send to other processes
   and receive from other processes
   for sparse matrix vector multiply and triangular solve 
   matrix must be DDL_PCK_ROW currently

   Parameters:
     ia,ja,n: IN    sparse matrix data local indices (n number of local rows)
     nc: IN         number of columns in global matrix
     block: IN      block size used for distribution of object
     pv: IN         permutation vector for columns of matrix; or NULL
     proc: IN       calling process number
     nproc: IN      total number of processes distributed object uses
     comm: IN       communicator used by object
     rcv_tot_proc: OUT number of procs src must get data from
     snd_tot_proc: OUT number of procs src must send data to
     rcv_tot_data: OUT number of elements of x vector that src must get
     snd_tot_data: OUT total number of elements of x vector that src must send
     rcv: OUT       vector of process numbers that src must get data from
     snd: OUT       vector of process numbers that src must send data to
     rcv_ptr: OUT   index of first element in rcv_ind that src must get 
                    from rcv
     snd_ptr: OUT   index of first element in snd_ind that src must send to 
                    snd
     rcv_cnt: OUT   number of elements from each process    
     snd_cnt: OUT   number of elements to each process    
     rcv_ind: OUT   vector contains global indices of elements of x that src 
                    must get from each process in rcv
     snd_ind: OUT   vector contains local indices of elements of x that src 
                    must send to each process in snd
     rcv_pind: OUT  permuted index vector
   Status: L2 */
int DDL_Col(int *ia, int *ja, int n, int nc, int block, int *pv,
	    int proc, int nproc, MPI_Comm comm, 
	    int *rcv_tot_proc, int *snd_tot_proc, 
	    int *rcv_tot_data, int *snd_tot_data, 
	    int rcv[], int snd[], 
	    int rcv_ptr[], int snd_ptr[], 
	    int rcv_cnt[], int snd_cnt[],
	    int rcv_ind[], int snd_ind[],
	    int rcv_pind[])
{
  int cnt, i, next, next_proc, ierr;
  int offset;
  int *tmp;

  ierr = DDL_SUCCESS;
  /* find non-zero columns */
  ierr = DDL_Sparse_nzcol(ia, ja, n, nc, rcv_ind, rcv_tot_data);
  /* set up receive structure */
  ierr = DDL_Calccomm(ia, ja, n, nc, block, pv,
		      proc, nproc, comm, 
		      rcv_tot_proc, snd_tot_proc, 
		      rcv_tot_data, snd_tot_data, 
		      rcv, snd, 
		      rcv_ptr, snd_ptr, 
		      rcv_cnt, snd_cnt, 
		      rcv_ind, snd_ind,
		      rcv_pind);
  return ierr;
}

/* calculate comms structure given rcv_tot_data and rcv_ind sorted 
   for sparse matrix vector multiply and triangular solve 
   matrix must be DDL_PCK_ROW currently

   Parameters:
     ia,ja,n: IN    sparse matrix data local indices (n number of local rows)
     nc: IN         number of columns in global matrix
     block: IN      block size used for distribution of object
     pv: IN         permutation vector for columns of matrix; or NULL
     proc: IN       calling process number
     nproc: IN      total number of processes distributed object uses
     comm: IN       communicator used by object
     rcv_tot_proc: OUT number of procs src must get data from
     snd_tot_proc: OUT number of procs src must send data to
     rcv_tot_data: IN number of elements of x vector that src must get
     snd_tot_data: OUT total number of elements of x vector that src must send
     rcv: OUT       vector of process numbers that src must get data from
     snd: OUT       vector of process numbers that src must send data to
     rcv_ptr: OUT   index of first element in rcv_ind that src must get 
                    from rcv
     snd_ptr: OUT   index of first element in snd_ind that src must send to 
                    snd
     rcv_cnt: OUT   number of elements from each process    
     snd_cnt: OUT   number of elements to each process    
     rcv_ind: IN    vector contains global indices of elements of x that src 
                    must get from each process in rcv
     snd_ind: OUT   vector contains local indices of elements of x that src 
                    must send to each process in snd
     rcv_pind: OUT  permuted index vector
   Status: L2 */
int DDL_Calccomm(int *ia, int *ja, int n, int nc, int block, int *pv,
		 int proc, int nproc, MPI_Comm comm, 
		 int *rcv_tot_proc, int *snd_tot_proc, 
		 int *rcv_tot_data, int *snd_tot_data, 
		 int rcv[], int snd[], 
		 int rcv_ptr[], int snd_ptr[], 
		 int rcv_cnt[], int snd_cnt[],
		 int rcv_ind[], int snd_ind[],
		 int rcv_pind[])
{
  int cnt, i, next, next_proc, ierr;
  int offset;
  int *tmp;

  ierr = DDL_SUCCESS;
  /* set up receive structure */
  ierr = DDL_Packcomm(nc, block, proc, nproc, rcv_tot_proc, rcv_tot_data, 
		      rcv, rcv_ptr, rcv_ind);
  /* exchange elements */
  ierr = DDL_Excomm(nproc, comm, 
		    rcv_tot_proc, snd_tot_proc, 
		    rcv_tot_data, snd_tot_data, 
		    rcv, snd, 
		    rcv_ptr, snd_ptr, 
		    rcv_cnt, snd_cnt, 
		    rcv_ind, snd_ind);

  offset = DDL_Find_offset(nc, block, nproc, proc);      
  
  /* data structure now correct if no permutation is given */
  /* so check whether columns of matrix are permuted */
  if (pv != NULL) {
    /* create permuted indices output vector */
    for (i=0;i<*snd_tot_data;i++)
      snd_ind[i] = pv[snd_ind[i]-offset];

    /* send and receive blocks of permuted indices */
    ierr = DDL_Alltoallvs(snd_ind, *snd_tot_proc, 
			  snd, snd_cnt, snd_ptr, MPI_INT,
			  rcv_pind, *rcv_tot_proc,
			  rcv, rcv_cnt, rcv_ptr, MPI_INT,
			  comm);
    /* receive structure now contains permuted indices I want */
    /* sort permuted index vector, but keep mapping between
       original indices wanted and final indices wanted */
    DDL_qs2i(rcv_pind, rcv_ind, *rcv_tot_data, 1);
    /* set up receive structure again */
    ierr = DDL_Packcomm(nc, block, proc, nproc, rcv_tot_proc, rcv_tot_data, 
			rcv, rcv_ptr, rcv_pind);
    /* exchange elements again */
    ierr = DDL_Excomm(nproc, comm, 
		      rcv_tot_proc, snd_tot_proc, 
		      rcv_tot_data, snd_tot_data, 
		      rcv, snd, 
		      rcv_ptr, snd_ptr, 
		      rcv_cnt, snd_cnt, 
		      rcv_pind, snd_ind);
    /* structure now correct taking into account the permutation */
  }

  /* convert global indexes in snd_ind to local indexes */
  for (i=0;i<*snd_tot_data;i++)
    snd_ind[i] = snd_ind[i]-offset;

  return ierr;
}

/* find the elements a process must send to other processes
   and receive from other processes
   for sparse matrix vector multiply and triangular solve 
   matrix must be DDL_PCK_ROW currently

   Parameters:
     ia,ja,n: IN    sparse matrix data local indices (n number of local rows)
     nr: IN         number of rows in global matrix
     block: IN      block size used for distribution of object
     pv: IN         permutation vector for columns of matrix; or NULL
     proc: IN       calling process number
     nproc: IN      total number of processes distributed object uses
     comm: IN       communicator used by object
     rcv_tot_proc: OUT number of procs src must get data from
     snd_tot_proc: OUT number of procs src must send data to
     rcv_tot_data: OUT number of elements of x vector that src must get
     snd_tot_data: OUT total number of elements of x vector that src must send
     rcv: OUT       vector of process numbers that src must get data from
     snd: OUT       vector of process numbers that src must send data to
     rcv_ptr: OUT   index of first element in rcv_ind that src must get 
                    from rcv
     snd_ptr: OUT   index of first element in snd_ind that src must send to 
                    snd
     rcv_cnt: OUT   number of elements from each process    
     snd_cnt: OUT   number of elements to each process    
     rcv_ind: OUT   vector contains global indices of elements of x that src 
                    must get from each process in rcv
     snd_ind: OUT   vector contains local indices of elements of x that src 
                    must send to each process in snd
     rcv_pind: OUT  permuted index vector
   Status: L2 */
int DDL_Row(int *ia, int *ja, int n, int nr, int block, int *pv,
	    int proc, int nproc, MPI_Comm comm, 
	    int *rcv_tot_proc, int *snd_tot_proc, 
	    int *rcv_tot_data, int *snd_tot_data, 
	    int rcv[], int snd[], 
	    int rcv_ptr[], int snd_ptr[], 
	    int rcv_cnt[], int snd_cnt[],
	    int rcv_ind[], int snd_ind[],
	    int rcv_pind[])
{
  int cnt, i, next, next_proc, ierr, offset;

  ierr = DDL_SUCCESS;

  offset = DDL_Find_offset(nr, block, nproc, proc);      

  /* find global C indices of elements of RHS needed */
  for (i=0;i<n;i++) rcv_ind[i] = i+offset;
  /* find permuted indices */
  for (i=0;i<n;i++) rcv_pind[i] = pv[i];

  *rcv_tot_data = n;

  /* sort permuted index vector, but remeber original index in rcv_ind */
  DDL_qs2i(rcv_pind, rcv_ind, *rcv_tot_data, 1);

  /* set up receive structure */
  ierr = DDL_Packcomm(nr, block, proc, nproc, rcv_tot_proc, rcv_tot_data, 
		      rcv, rcv_ptr, rcv_pind);

  /* exchange elements */
  ierr = DDL_Excomm(nproc, comm, 
		    rcv_tot_proc, snd_tot_proc, 
		    rcv_tot_data, snd_tot_data, 
		    rcv, snd, 
		    rcv_ptr, snd_ptr, 
		    rcv_cnt, snd_cnt, 
		    rcv_pind, snd_ind);

  /* convert global indexes in snd_ind to local indexes */
  for (i=0;i<*snd_tot_data;i++)
    snd_ind[i] = snd_ind[i]-offset;

  /* data structure now correct */
  return ierr;
}

/* calculate inverse permutation vector given original permutation
   Parameters:
     pv: IN         original permutation vector
     ipv: OUT       inverse permutation vector
     n: IN          size of vector
   Status: L1 */
int DDL_Calc_invperm(int *pv, int *ipv, int n)
{
  int i;
  for (i=0;i<n;i++)
    ipv[pv[i]] = i;
}

/* apply permutation to a vector 
       v[i] = v[pv[i]] 
   Parameters:
     v: INOUT       vector to permute
     type: IN       MPI data type of vector
     pv: IN         permutation vector
     n: IN          size of vector
   Status: L1 */
int DDL_Lpermute_vector(void *v, MPI_Datatype type, int *pv, int n)
{
  int i, j, i1, i2, itype, ierr;
  double d1, d2;
  float f1, f2;
  MPI_Aint size;
  void *tv;
  ierr = MPI_Type_size(type, &size);
  tv = malloc(n*size);
  ierr = DDL_DDLtype(type, &itype);
  switch (itype) {
  case DDL_DOUBLE:
    /* copy old vector */
    for (i=0;i<n;i++) ((double*)tv)[i] = ((double*)v)[i];
    /* now permute back to original vector */
    for (i=0;i<n;i++) ((double*)v)[i] = ((double*)tv)[pv[i]];
    break;
  case DDL_FLOAT:
    /* copy old vector */
    for (i=0;i<n;i++) ((float*)tv)[i] = ((float*)v)[i];
    /* now permute back to original vector */
    for (i=0;i<n;i++) ((float*)v)[i] = ((float*)tv)[pv[i]];
    break;
  case DDL_INT:
    /* copy old vector */
    for (i=0;i<n;i++) ((int*)tv)[i] = ((int*)v)[i];
    /* now permute back to original vector */
    for (i=0;i<n;i++) ((int*)v)[i] = ((int*)tv)[pv[i]];
    break;
  }
  free(tv);
  return DDL_SUCCESS;
}

/* routines to calculate red-black ordering permutation vector */

/* top level routine
   Parameters:
     nx,ny,nz: IN   number of components in each dimension
     pv: OUT        permutation vector
   Status: H1 */
int DDL_calc_perm(int nx, int ny, int nz, int *pv)
{
  int i, nxny;
  nxny = nx*ny;
  DDL_cat_plane(&pv, 0, 1, nx, ny);
  for (i=1;i<nz;i++)
    if (i%2 == 1)
      DDL_cat_plane(&pv, i*nxny, 2, nx, ny);
    else
      DDL_cat_plane(&pv, i*nxny, 1, nx, ny);
  DDL_cat_plane(&pv, 0, 2, nx, ny);
  for (i=1;i<nz;i++)
    if (i%2 == 1)
      DDL_cat_plane(&pv, i*nxny, 1, nx, ny);
    else
      DDL_cat_plane(&pv, i*nxny, 2, nx, ny);
}


/* concatenate whole plane to current permutation.
   Parameters:
     pv: INOUT      pointer to end of current permutation fragment
     base: IN       base value to add
     plane: IN      which plane to add; 1 or 2?
     nx,ny: IN      number of components in each dimension
   Status: L1 */
int DDL_cat_plane(int **pv, int base, int plane, int nx, int ny) 
{
  int i;
  if (nx%2 == 1) {
    for (i=0;i<=ny-1;i++)
      if (i%2 == 1)
	if (plane == 1) 
	  DDL_cat_part(pv, base+(i-1)*nx, 1, nx, 2);
	else
	  DDL_cat_part(pv, base+(i-1)*nx, 0, nx, 2);
      else
	if (plane == 1) 
	  DDL_cat_part(pv, base+i*nx, 1, nx, 1);
	else
	  DDL_cat_part(pv, base+i*nx, 0, nx, 1);
  }
  else {
    for (i=0;i<=ny-1;i++)
      if (i%2 == 1)
	if (plane == 1) 
	  DDL_cat_part(pv, base+i*nx, 0, nx, 1);
	else
	  DDL_cat_part(pv, base+i*nx, 1, nx, 1);
      else
	if (plane == 1) 
	  DDL_cat_part(pv, base+i*nx, 1, nx, 1);
	else
	  DDL_cat_part(pv, base+i*nx, 0, nx, 1);
  }
}

/* concatenate odd or even permutation to current permutation.
   Parameters:
     pv: INOUT      pointer to end of current permutation fragment
     base: IN       base value to add to all added perm values
     odd: IN        odd or even permutation to be added? odd = 1 odd perm
     nx: IN         number of components in dimension
     range: IN      range of permutation to add; 
                    1 for 1--nx or 2 for nx--2*nx      
   Status: L1 */
int DDL_cat_part(int **pv, int base, int odd, int nx, int range) 
{
  int odd_nx, even_nx, i;
  even_nx = nx/2;
  if (even_nx*2 == nx) odd_nx = even_nx;
  else odd_nx=even_nx+1;
  if (range == 1) {
    /* nx perm to be added */
    if (odd) {
      /* odd perm;  length=odd_nx */
      for (i=0;i<odd_nx;i++)
	*(*pv)++ = base+2*i;
    }
    else {
      /* even perm;  length=even_nx */
      for (i=0;i<even_nx;i++)
	*(*pv)++ = base+1+2*i;
    }
  }
  else {
    /* 2*nx perm to be added */
    if (odd) {
      /* odd perm;  length=even_nx */
      for (i=0;i<even_nx;i++)
	*(*pv)++ = base+2*odd_nx+2*i;
    }
    else {
      /* even perm;  length=odd_nx */
      for (i=0;i<odd_nx;i++)
	*(*pv)++ = base+2*even_nx+1+2*i;
    }
  }
}

/* assorted quick sort functions
   sorting one vector and permuting a number of other vectors similarly 
   adapted from QS2I1D */

/* ordinary quick sort 
   Parameters:
     ia: INOUT      integer array of values to be sorted.
     n: IN          number of values in integer array ia to be sorted.
     kflag: IN      control parameter
                      1: means sort ia in INCREASING order.
		      -1: means sort ia in DECREASING order.
   Status: L1 */
int DDL_qs1i(int *ia, int n, int kflag)
{
  int il[22], iu[22];
  int l, m, i, j, k, nn, kk, ij, it, iit, jt, jjt;
  double r, ta, tta;
  int ierr= DDL_SUCCESS;

  nn = n-1;
  if (nn < 0) return -1;
  if (n == 1) return ierr;
  kk = abs(kflag);
  if (kk != 1) return -2;
  /* Alter array IA to get decreasing order if needed. */
  if (kflag < 1) 
    for (i=0;i<=nn;i++)
      ia[i] = -ia[i];

  /* Sort IA and carry JA and A along.
     And now...Just a little black magic... */
  m = 0;
  i = 0;
  j = nn;
  r = 0.375;
g210:
  if (r <= 0.5898437) r = r+3.90625E-2;
  else r = r-0.21875;
g225:
  k = i;

  /* Select a central element of the array and save it in location 
     it, jt, at. */
  ij = i + (double)(j-i)*r;
  it = ia[ij];

  /* If first element of array is greater than it, interchange with it. */
  if (ia[i] > it) {
    ia[ij] = ia[i];
    ia[i]  = it;
    it     = ia[ij];
  }
  l=j;

  /* If last element of array is less than it, swap with it. */
  if (ia[j] < it) {
    ia[ij] = ia[j];
    ia[j]  = it;
    it     = ia[ij];
    /* If first element of array is greater than it, swap with it.*/
    if (ia[i] > it) {
      ia[ij] = ia[i];
      ia[i]  = it;
      it     = ia[ij];
    }
  }

  /* Find an element in the second half of the array which is 
     smaller than it. */
g240:
  l=l-1;
  if (ia[l] > it) goto g240;
  /* Find an element in the first half of the array which is 
     greater than it. */
g245:
  k=k+1;
  if (ia[k] < it) goto g245;

  /* Interchange these elements. */
  if (k <= l) {
    iit   = ia[l];
    ia[l] = ia[k];
    ia[k] = iit;
    goto g240;
  }

  /* Save upper and lower subscripts of the array yet to be sorted. */
  if (l-i > j-k) {
    il[m] = i;
    iu[m] = l;
    i = k;
    m = m+1;
  }
  else {
    il[m] = k;
    iu[m] = j;
    j = l;
    m = m+1;
  }
  goto g260;

/* Begin again on another portion of the unsorted array. */
g255:
  m = m-1;
  if (m < 0) goto g300;
  i = il[m];
  j = iu[m];
g260:
  if (j-i >= 1) goto g225;
  if (i==j) goto g255;
  if(i==0) goto g210;
  i = i-1;
g265:
  i = i+1;
  if (i==j) goto g255;
  it = ia[i+1];
  if (ia[i] <= it) goto g265;
  k=i;
g270:
  ia[k+1] = ia[k];
  k = k-1;
  if (it<ia[k]) goto g270;
  ia[k+1] = it;
  goto g265;

/* Clean up, if necessary. */
g300:
  if (kflag < 1) 
    for (i=0;i<=nn;i++)
      ia[i] = -ia[i];
  return ierr;
}

/* quick sort with additional integer and double vectors
   Parameters:
     ia: INOUT      integer array of values to be sorted.
     ja: INOUT      integer array to be carried along.
     a: INOUT       double precision array to be carried along.
     n: IN          number of values in integer array ia to be sorted.
     kflag: IN      control parameter
                      1: means sort ia in INCREASING order.
		      -1: means sort ia in DECREASING order.
   Status: L1 */
int DDL_qs2i1d(int *ia, int *ja, double *a, int n, int kflag)
{
  int il[22], iu[22];
  int l, m, i, j, k, nn, kk, ij, it, iit, jt, jjt;
  double r, ta, tta;
  int ierr= DDL_SUCCESS;

  nn = n-1;
  if (nn < 0) return -1;
  if (n == 1) return ierr;
  kk = abs(kflag);
  if (kk != 1) return -2;
  /* Alter array IA to get decreasing order if needed. */
  if (kflag < 1) 
    for (i=0;i<=nn;i++)
      ia[i] = -ia[i];

  /* Sort IA and carry JA and A along.
     And now...Just a little black magic... */
  m = 0;
  i = 0;
  j = nn;
  r = 0.375;
g210:
  if (r <= 0.5898437) r = r+3.90625E-2;
  else r = r-0.21875;
g225:
  k = i;

  /* Select a central element of the array and save it in location 
     it, jt, at. */
  ij = i + (double)(j-i)*r;
  it = ia[ij];
  jt = ja[ij];
  ta = a[ij];

  /* If first element of array is greater than it, interchange with it. */
  if (ia[i] > it) {
    ia[ij] = ia[i];
    ia[i]  = it;
    it     = ia[ij];
    ja[ij] = ja[i];
    ja[i]  = jt;
    jt     = ja[ij];
    a[ij]  = a[i];
    a[i]   = ta;
    ta     = a[ij];
  }
  l=j;

  /* If last element of array is less than it, swap with it. */
  if (ia[j] < it) {
    ia[ij] = ia[j];
    ia[j]  = it;
    it     = ia[ij];
    ja[ij] = ja[j];
    ja[j]  = jt;
    jt     = ja[ij];
    a[ij]  = a[j];
    a[j]   = ta;
    ta     = a[ij];
    /* If first element of array is greater than it, swap with it.*/
    if (ia[i] > it) {
      ia[ij] = ia[i];
      ia[i]  = it;
      it     = ia[ij];
      ja[ij] = ja[i];
      ja[i]  = jt;
      jt     = ja[ij];
      a[ij]  = a[i];
      a[i]   = ta;
      ta     = a[ij];
    }
  }

  /* Find an element in the second half of the array which is 
     smaller than it. */
g240:
  l=l-1;
  if (ia[l] > it) goto g240;
  /* Find an element in the first half of the array which is 
     greater than it. */
g245:
  k=k+1;
  if (ia[k] < it) goto g245;

  /* Interchange these elements. */
  if (k <= l) {
    iit   = ia[l];
    ia[l] = ia[k];
    ia[k] = iit;
    jjt   = ja[l];
    ja[l] = ja[k];
    ja[k] = jjt;
    tta   = a[l];
    a[l]  = a[k];
    a[k]  = tta;
    goto g240;
  }

  /* Save upper and lower subscripts of the array yet to be sorted. */
  if (l-i > j-k) {
    il[m] = i;
    iu[m] = l;
    i = k;
    m = m+1;
  }
  else {
    il[m] = k;
    iu[m] = j;
    j = l;
    m = m+1;
  }
  goto g260;

/* Begin again on another portion of the unsorted array. */
g255:
  m = m-1;
  if (m < 0) goto g300;
  i = il[m];
  j = iu[m];
g260:
  if (j-i >= 1) goto g225;
  if (i==j) goto g255;
  if(i==0) goto g210;
  i = i-1;
g265:
  i = i+1;
  if (i==j) goto g255;
  it = ia[i+1];
  jt = ja[i+1];
  ta =  a[i+1];
  if (ia[i] <= it) goto g265;
  k=i;
g270:
  ia[k+1] = ia[k];
  ja[k+1] = ja[k];
  a[k+1]  =  a[k];
  k = k-1;
  if (it<ia[k]) goto g270;
  ia[k+1] = it;
  ja[k+1] = jt;
  a[k+1]  = ta;
  goto g265;

/* Clean up, if necessary. */
g300:
  if (kflag < 1) 
    for (i=0;i<=nn;i++)
      ia[i] = -ia[i];
  return ierr;
}

/* quick sort with additional integer vector
   Parameters:
     ia: INOUT      integer array of values to be sorted.
     ja: INOUT      integer array to be carried along.
     n: IN          number of values in integer array ia to be sorted.
     kflag: IN      control parameter
                      1: means sort ia in INCREASING order.
		      -1: means sort ia in DECREASING order.
   Status: L1 */
int DDL_qs2i(int *ia, int *ja, int n, int kflag)
{
  int il[22], iu[22];
  int l, m, i, j, k, nn, kk, ij, it, iit, jt, jjt;
  double r, ta, tta;
  int ierr= DDL_SUCCESS;

  nn = n-1;
  if (nn < 0) return -1;
  if (n == 1) return ierr;
  kk = abs(kflag);
  if (kk != 1) return -2;
  /* Alter array IA to get decreasing order if needed. */
  if (kflag < 1) 
    for (i=0;i<=nn;i++)
      ia[i] = -ia[i];

  /* Sort IA and carry JA and A along.
     And now...Just a little black magic... */
  m = 0;
  i = 0;
  j = nn;
  r = .375;
g210:
  if (r <= 0.5898437) r = r+3.90625E-2;
  else r = r-.21875;
g225:
  k = i;

  /* Select a central element of the array and save it in location 
     it, jt, at. */
  ij = i + (double)(j-i)*r;
  it = ia[ij];
  jt = ja[ij];

  /* If first element of array is greater than it, interchange with it. */
  if (ia[i] > it) {
    ia[ij] = ia[i];
    ia[i]  = it;
    it     = ia[ij];
    ja[ij] = ja[i];
    ja[i]  = jt;
    jt     = ja[ij];
  }
  l=j;

  /* If last element of array is less than it, swap with it. */
  if (ia[j] < it) {
    ia[ij] = ia[j];
    ia[j]  = it;
    it     = ia[ij];
    ja[ij] = ja[j];
    ja[j]  = jt;
    jt     = ja[ij];
    /* If first element of array is greater than it, swap with it.*/
    if (ia[i] > it) {
      ia[ij] = ia[i];
      ia[i]  = it;
      it     = ia[ij];
      ja[ij] = ja[i];
      ja[i]  = jt;
      jt     = ja[ij];
    }
  }

  /* Find an element in the second half of the array which is 
     smaller than it. */
g240:
  l=l-1;
  if (ia[l] > it) goto g240;
  /* Find an element in the first half of the array which is 
     greater than it. */
g245:
  k=k+1;
  if (ia[k] < it) goto g245;

  /* Interchange these elements. */
  if (k <= l) {
    iit   = ia[l];
    ia[l] = ia[k];
    ia[k] = iit;
    jjt   = ja[l];
    ja[l] = ja[k];
    ja[k] = jjt;
    goto g240;
  }

  /* Save upper and lower subscripts of the array yet to be sorted. */
  if (l-i > j-k) {
    il[m] = i;
    iu[m] = l;
    i = k;
    m = m+1;
  }
  else {
    il[m] = k;
    iu[m] = j;
    j = l;
    m = m+1;
  }
  goto g260;

/* Begin again on another portion of the unsorted array. */
g255:
  m = m-1;
  if (m < 0) goto g300;
  i = il[m];
  j = iu[m];
g260:
  if (j-i >= 1) goto g225;
  if (i==j) goto g255;
  if(i==0) goto g210;
  i = i-1;
g265:
  i = i+1;
  if (i==j) goto g255;
  it = ia[i+1];
  jt = ja[i+1];
  if (ia[i] <= it) goto g265;
  k=i;
g270:
  ia[k+1] = ia[k];
  ja[k+1] = ja[k];
  k = k-1;
  if (it<ia[k]) goto g270;
  ia[k+1] = it;
  ja[k+1] = jt;
  goto g265;

/* Clean up, if necessary. */
g300:
  if (kflag < 1) 
    for (i=0;i<=nn;i++)
      ia[i] = -ia[i];
  return ierr;
}

/* truncate malloc'd integer vector to length n 
   Parameters:
     ptr: INOUT     pointer to integer vector
     n: IN          size of integer vector
   Status: L1 */
int DDL_Truncate(void *ptr, int n)
{  
  int *tmp;
  tmp = malloc(n*sizeof(int));
  memcpy(tmp, *(void**)ptr, n*sizeof(int));
  free(*(void**)ptr);
  *(void**)ptr = tmp;
  return DDL_SUCCESS;
}

/* get coefficient value a(i,j) (C indices) from sparse matrix
   Parameters:
     ia, ja, a: IN   matrix elements
     type: IN        MPI_Datatype of coefficient elements
     format: IN      DDL format of matrix:
                        DDL_PCK_ROW or DDL_PCK_COL
     i,j: IN         indices of element to find
     coeff: OUT      coefficient value for element a(i,j)
   Status: L1 */
int DDL_Get_coeff(int *ia, int *ja, void *a, MPI_Datatype type, int format, 
		  int i, int j, void *coeff)
{ 
  int k;
  int ierr = DDL_SUCCESS;
  switch (format) {
  case DDL_PCK_ROW:
    ierr = DDL_bsearch(&ja[ia[i]-1], ia[i+1]-ia[i], j+1, &k);   
    k = k+ia[i]-1;
    break;
  case DDL_PCK_COL:
    ierr = DDL_bsearch(&ia[ja[j]-1], ja[j+1]-ja[j], i+1, &k);   
    k = k+ja[j]-1;
    break;
  }
  if (ierr == DDL_SUCCESS)
    if (type = MPI_DOUBLE)
      *(double*)coeff = ((double*)a)[k]; 
    else if (type = MPI_FLOAT)
      *(float*)coeff = ((float*)a)[k]; 
  return ierr;
}

/* search array for particular value using binary search
   Parameters:
     x: IN           array to search
     n: IN           size of array
     val: IN         value to search for
     idx: OUT        index of value if found
   Status: L1 */
int DDL_bsearch(int *x, int n, int val, int *idx)
{
  int i, u, l, v, ierr = DDL_SUCCESS;
  u = n-1;
  l = 0;
  i = (u-l+1)/2+l;
  while (l<=u && x[i] != val) {
    if (x[i] > val)
      u = i-1;
    else
      l = i+1;
    i = (u-l+1)/2+l;
  }
  if (x[i] != val) return DDL_ERROR;
  *idx = i;
  return ierr;
}

/* sequential permutation of **whole** sparse matrix

   very inefficient for now...let's see if we need it anywhere... 
   ia,ja vectors need to be at least size nelt 
   Parameters:
     ia,ja,a: INOUT sparse matrix 
     nelt: IN       number of non-zeros
     n: INOUT       number of rows
     m: INOUT       number of columns
     type: IN       MPI datatype of coefficient vector
     format: IN     DDL sparse matrix format:
                      DDL_PCK_ROW or DDL_PCK_COL
     px: IN         row permutation vector; or NULL
     py: IN         column permutation vector; or NULL
   Status: L1 */
int DDL_Lpermute_sparse(int *ia, int *ja, void *a, int nelt, 
			int n, int m, MPI_Datatype type, int format,
			int *px, int *py)
{
  int  i, j, k, isym = 0, ierr = DDL_SUCCESS;
  /* convert to triad format */
  if (format == DDL_PCK_ROW) 
    ierr = DDL_Rowtotriad(n, nelt, ia, ja, a, type, isym);
  else if (format == DDL_PCK_COL) 
    ierr = DDL_Coltotriad(n, nelt, ia, ja, a, type, isym);
  /* permute rows and columns */
  if (px != NULL)
    for (i=0;i<nelt;i++)
      ia[i] = px[ia[i]-1]+1;
  if (py != NULL)
    for (i=0;i<nelt;i++)
      ja[i] = py[ja[i]-1]+1;
  /* convert back to original format */
  if (format == DDL_PCK_ROW) 
    ierr = DDL_Triadtorow(n, nelt, ia, ja, a, type, isym);
  else if (format == DDL_PCK_COL) 
    ierr = DDL_Triadtocol(n, nelt, ia, ja, a, type, isym);
  i = n;
  n = m;
  m = i;
  return ierr;
}

