/* $Id: ddl_support.c,v 1.4 1995/08/24 10:01:19 tim Exp $ */

/* Copyright 1994 IASC */

/* useful support routines for sparse DDL */

/* Procedures and parameter lists:
   The DDL is continually evolving and this often results in new procedures
   or changes to parameter lists, and sometimes the removal of procedures.
   Each procedure is marked to indicate its longevity and use:
     1-4: indicates longevity of procedure interface 
        1: won't change (ever, honest!)
	2: unlikely to change, but possible
	3: might well be changed
	4: very likely to be changed or removed 
     HL: indicates expected use of procedure
        H: high level procedure, probably documented in written docs,
	   for user programs and library procedures
	L: low level procedure, not documented in written docs,
           for use in Library procedures mainly
   eg. DDL_Create_sparse: H1.
   High level procedures are usually unlikely to change. 
   Parameters are documented:
     IN:      input only parameter; is not changed by procedure
     OUT:     output parameter; any value on entry; changed on exit
     INOUT:   both input and output; read on input; may be changed on exit */

#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

#include "ddl.h"

/* wait for key press; for debugging purposes
   Status: L4 */
int twait(void)
{
  if (DDL_Ishost()) {
    char dummy[80];
    printf("Press any key to continue\n");
    scanf("%s", dummy);
  }
  return MPI_SUCCESS;
}

/* input Harwell-Boeing format file
   Parameters:
     file: INOUT    file handle
     n: OUT         number of rows (columns) of matrix
     nelt: OUT      number of nonzeros in matrix
     ia,ja,a: OUT   sparse matrix in DDL_PCK_COL format
     isym: OUT      whether matrix is symmetric
                        0: not symmetric
                        1: symmetric
     soln: OUT      solution vector to Ax=b, if requested and found
     rhs: OUT       right hand side vector, if requested and found 
     xinit: OUT     initial x-vector, if requested and found
     job: INOUT     A flag indicating on entry what I/O operations to
                    perform:
		      0:  Read only the matrix.
		      1:  Read matrix and rhs (if present). ???
		      2:  Read matrix and soln (if present). ???
		      3:  Read matrix, rhs, xinit and soln (if present).
		    On exit the flag indicates what operations were 
		    actually performed:
		      -3: Unable to parse matrix ``code'' from input file to
		          determine if only the lower triangle 
			  of matrix is stored.
		      -2: Number of non-zeros nelt too large.
		      -1: System size n too large.
		      1:  Read in only the matrix structure, but no 
		          nonzero entries.  hence, a is not referenced 
			  and has the same return values as on input.
		      3:  Read in only the matrix.
		      7:  Read in the matrix and rhs.
		      15: Read in the matrix, rhs, and xinit.
		      23: Read in the matrix, rhs, and soln.
		      31: Read in the matrix, rhs, soln and xinit.
   Status: H1 */
int DDL_Inhb(FILE *file, int *n, int *nelt, int ia[], int ja[], double a[],
	     int *isym, double soln[], double rhs[], double xinit[],
	     int *job)
{
  int jobin, ierr, i, j;
  int idx, ibgn, iend, itemp;
  double temp, zero;
  char title[80], code[3];
  int nline, npls, nrils, nnvls, nrhsls;
  int nrow, ncol, nind, nele;
  int nrhs, nrhsix;
  char rhstyp[3];
  char pntfmt[16], rinfmt[16], nvlfmt[20], rhsfmt[20];

  zero = 0.0;
  ierr = DDL_SUCCESS;
  jobin = *job;

  fscanf(file, "%80c\n", title);
  fscanf(file, "%d%d%d%d%d\n", &nline, &npls, &nrils, &nnvls, &nrhsls);
  fscanf(file, "%3c           %d%d%d%d", 
	 code, &nrow, &ncol, &nind, &nele);
  fscanf(file, "%s%s%s%s\n", pntfmt, rinfmt, nvlfmt, rhsfmt);
  if (nrow > *n) {
    *n = nrow;
    fclose(file);
    *job = -1;
    return DDL_ERROR;
  }
  if (nind > *nelt) {
    *nelt = nind;
    fclose(file);
    *job = -2;
    return DDL_ERROR;
  }
  *n = nrow;
  *nelt = nind;
  if (strncmp(code, "RUA", 3) == 0) {
    *isym = 0;
  }
  else if (strncmp(code, "RSA", 3) == 0) {
    *isym = 1;
  }
  else {
    fclose(file);
    *job = -3;
    return DDL_ERROR;
  }
  /* deal with extra input line for RHS vectors */
  if (nrhsls > 0) {
    fscanf(file, "%3c           %d%d\n", rhstyp, &nrhs, &nrhsix);
  }
  for (i=0;i<*n+1;i++) {
    fscanf(file, "%d", &ja[i]);
  }
  for (i=0;i<*nelt;i++) {
    fscanf(file, "%d", &ia[i]);
  }
  *job = 1;
  if (nnvls > 0) {
    for (i=0;i<*nelt;i++) {
      fscanf(file, "%le", &a[i]);
    }
    *job = *job + 2;
  }
  if (nrhsls > 0 && jobin&1 == 1) {
    /* read in rhs vector */
    if (rhstyp[0] == 'F') {
      for (i=0;i<*n;i++) {
        fscanf(file, "%le", &rhs[i]);
      }
      *job = *job + 4;
    }
    else {
      F77(dfill)(n, rhs, &zero);
    }
    /* read in starting guess */
    if (rhstyp[1] == 'G') {
      for (i=0;i<*n;i++) {
        fscanf(file, "%le", &xinit[i]);
      }
      *job = *job + 8;
    }
    else {
      F77(dfill)(n, xinit, &zero);
    }
    /* read in solution vector */
    if (rhstyp[2] == 'X') {
      for (i=0;i<*n;i++) {
        fscanf(file, "%le", &soln[i]);
      }
      *job = *job + 16;
    }
    else {
      F77(dfill)(n, soln, &zero);
    }
  }

  /* reorder data to column packed---should be like this already */
  /* remember data in FORTRAN format---row/col indices from 1 */
  for (idx=0;idx<*n;idx++) {
    ibgn = ja[idx];
    iend = ja[idx+1]-1;
    if (ibgn < iend) {
      for (i=ibgn-1;i<iend-1;i++) {
	for (j=i+1;j<iend-1;j++) {
	  if (ia[i] > ia[j]) {
	    itemp = ia[i];
	    ia[i] = ia[j];
	    ia[j] = itemp;
	    temp = a[i];
	    a[i] = a[j];
	    a[j] = temp;
	  }
	}
      }
    }
  }
  return ierr;
}

/* sparse matrix I/O for DDL using DDL file format.
   to be called by host process only */

/* output sparse matrix to file
   matrix must be in triad format: DDL_TRIAD_ROW or DDL_TRIAD_COL

   Parameters:
     file: INOUT    file handle
     n: IN          number of rows (columns) of matrix
     nelt: IN       number of nonzeros in matrix
     ia,ja,a: IN    sparse matrix
     type: IN       MPI data type of elements of coefficient vector
     isym: IN       whether matrix is symmetric (not used)
                        0: not symmetric
                        1: symmetric
     format: IN     format of sparse matrix; must be either
                        DDL_TRIAD_ROW: row ordered triad format, or
                        DDL_TRIAD_COL: column ordered triad format
   Status: H1 */
int DDL_Out_sparse(FILE *file, int n, int nelt, 
		   int ia[], int ja[], double a[], MPI_Datatype type,
		   int isym, int format)
{
  int i;
  int ftype;

  /* find type of element */
  DDL_DDLtype(type, &ftype);

  /* Write out the information heading */
  fprintf(file, "%10d\n", format);
  fprintf(file, "%10d%10d%10d%10d\n", ftype, n, nelt, isym);

  /* Write out the matrix non-zeros in triad format */
  for (i=0;i<nelt;i++) {
    fprintf(file, " %5d %5d %16.7le\n", ia[i], ja[i], a[i]);
  }
  return DDL_SUCCESS;
}

/* input sparse matrix from DDL format file.
   matrix will be held in triad format, either DDL_TRIAD_ROW or DDL_TRIAD_COL
   depending on format used in file.

   Parameters:
     file: INOUT    file handle
     n: OUT         number of rows (columns) of matrix
     nelt: OUT      number of nonzeros in matrix
     ia,ja,a: OUT   sparse matrix
     type: OUT      MPI data type of elements of coefficient vector
     isym: OUT      whether matrix is symmetric (not used)
                        0: not symmetric
                        1: symmetric
     format: OUT    format of sparse matrix; will be either
                        DDL_TRIAD_ROW: row ordered triad format, or
                        DDL_TRIAD_COL: column ordered triad format
   Status: H1 */
int DDL_In_sparse(FILE *file, int *n, int *nelt, 
		  int ia[], int ja[], double a[], MPI_Datatype *type,
		  int *isym, int *format)
{
  int i, mxnelt, ftype;

  /* read in the information heading */
  mxnelt = *nelt;
  fscanf(file, "%10d\n", format);
  fscanf(file, "%10d%10d%10d%10d\n", &ftype, n, nelt, isym);
  *nelt = DDL_MIN(*nelt, mxnelt);
  
  /* find type of element */
  DDL_MPItype(ftype, type);

  /* read in the matrix of non-zeros */
  for (i=0;i<*nelt;i++) {
    fscanf(file, "%d%d%le\n", &ia[i], &ja[i], &a[i]);
  }
  return DDL_SUCCESS;
}

/* dense vector I/O for DDL using DDL file format
   to be called by host process only */

/* output vector to file
   Parameters:
     file: INOUT    file handle
     vec: IN        elements of vector
     type: IN       MPI data type of elements of vector
     n: IN          number of elements in vector
   Status: H1 */
int DDL_Out_vector(FILE *file, void *vec, MPI_Datatype type, int *n)
{
  int i, ierr;
  int fformat, ftype;

  ierr = DDL_SUCCESS;
  fformat = DDL_VEC_FMT;

  /* find type of element */
  DDL_DDLtype(type, &ftype);

  /* Write out the information heading */
  fprintf(file, "%10d\n", fformat);
  fprintf(file, "%10d%10d\n", *n, ftype);

  ierr = DDL_Write_vectorblk(file, vec, type, n);

  return ierr;
}

/* input vector from file
   Parameters:
     file: INOUT    file handle
     vec: OUT       elements of vector
     type: OUT      MPI data type of elements of vector
     n: OUT         number of elements in vector
   Status: H1 */
int DDL_In_vector(FILE *file, void *vec, MPI_Datatype *type, int *n)
{
  int ierr;
  int fformat, ftype;
  int fn;

  /* check file parameters */
  ierr = DDL_SUCCESS;
  if (file == NULL) {
    ierr = DDL_Error(DDL_ERR_FILE, DDL_ERR_HARD, DDL_IN_VECTOR);
  }
  /* check file contains vector in correct format */
  else {
    fscanf(file, "%10d\n", &fformat);
    if (fformat != DDL_VEC_FMT) {
      ierr = DDL_Error(DDL_ERR_FILEFORMAT, DDL_ERR_HARD, DDL_IN_VECTOR);
    }
    else {
      /* get vector details from file */
      fscanf(file, "%10d%10d\n", &fn, &ftype);
      if (*n < fn) {
	ierr = DDL_Error(DDL_ERR_TOOSMALL, DDL_ERR_HARD, DDL_IN_VECTOR);
      }
      else {
	*n = fn;
	/* find type of element */
	DDL_MPItype(ftype, type);
	ierr = DDL_Read_vectorblk(file, vec, *type, n);
      }
    }
  }
  return ierr;
}

/* open file (called on host only)
   Parameters:
     file: OUT      file handle
     fname: IN      file name
     mode: IN       access mode:
                      DDL_READ_MODE: open for reading only
                      DDL_WRITE_MODE: open for writing only
   Status: H1 */
int DDL_Open_host(FILE **file, char *fname, int mode)
{
  int ierr;

  ierr = DDL_SUCCESS;

  if (mode == DDL_READ_MODE) {
    /* open file for reading */
    *file = fopen(fname, "r");
  }
  else if (mode == DDL_WRITE_MODE) {
    /* open file for writing */
    *file = fopen(fname, "w");
  }

  /* check file parameter */
  if (*file == NULL) {
    ierr = DDL_Error(DDL_ERR_FILE, DDL_ERR_HARD, DDL_OPEN_HOST);
  }
  return ierr;
}

/* open file (called on all processes in communicator)
   Parameters:
     file: OUT      file handle
     fname: IN      file name
     mode: IN       access mode:
                      DDL_READ_MODE: open for reading only
                      DDL_WRITE_MODE: open for writing only
     comm: IN       MPI communicator of processes involved
   Status: H1 */
int DDL_Open(FILE **file, char *fname, int mode, MPI_Comm comm)
{
  int ierr;

  ierr = DDL_SUCCESS;
  
  /* only open file on host */
  if (DDL_Ishost()) {
    ierr = DDL_Open_host(file, fname, mode);
  }

  /* tell all nodes of error condition */
  MPI_Bcast(&ierr, 1, MPI_INT, 0, comm);

  return ierr;
}

/* close file (called on host only)
   Parameters:
     file: INOUT    file handle
   Status: H1 */
int DDL_Close_host(FILE *file)
{
  int ierr;

  ierr = DDL_SUCCESS;

  /* check file parameter */
  if (file == NULL) {
    ierr = DDL_Error(DDL_ERR_FILE, DDL_ERR_HARD, DDL_CLOSE_HOST);
  }
  else {
    fclose(file);
  }
  return ierr;
}

/* close file  (called on all processes in communicator)
   Parameters:
     file: INOUT    file handle
     comm: IN       MPI communicator of processes involved
   Status: H1 */
int DDL_Close(FILE *file, MPI_Comm comm)
{
  int ierr;

  ierr = DDL_SUCCESS;
  
  /* only close file on host */
  if (DDL_Ishost()) {
    ierr = DDL_Close_host(file);
  }

  /* tell all nodes of error condition */
  MPI_Bcast(&ierr, 1, MPI_INT, 0, comm);

  return ierr;
}

/* file I/O format query routines
   return type and parameters for next object in file
   call these routines to find out the next object in the file
   then call the relevant input routine to input the object */

/* called on host only 
   Parameters:
     file: INOUT    file handle
     format: OUT    format of next object in file; will be one of: 
                       DDL_TRIAD_ROW: row ordered triad format
		       DDL_TRIAD_COL: column ordered triad format
		       DDL_VEC_FMT: vector format
     type: OUT      MPI data type of elements of object
     a: OUT         extra info for object
                    for dense vector:
		       0: number of elements in vector
		    for sparse matrix:  
		       0: number of rows (columns) of square matrix
		       1: number of nonzeros in matrix
   Status: H1 */
int DDL_Fileformat_host(FILE *file, int *format, MPI_Datatype *type, 
			int a[])
{
  int ierr;
  int dummy, ftype, isym;

#ifndef HAVE_FPOS_T /* gcc compiler not know fpos_t */
  long fpos;
#else
  fpos_t fpos;
#endif

  ierr = DDL_SUCCESS;
  
  /* check file parameters */
  if (file == NULL) {
    ierr = DDL_Error(DDL_ERR_FILE, DDL_ERR_HARD, DDL_FILEFORMAT_HOST);
  }
  /* find next object format */
  else {
#ifndef HAVE_FPOS_T /* gcc 2.5.7 compiler not know fgetpos? */
    fpos = ftell(file);
#else
    fgetpos(file, &fpos);
#endif
    fscanf(file, "%10d\n", format);
    /* now read in parameters for objects */
    if (*format == DDL_TRIAD_ROW || *format == DDL_TRIAD_COL) {
      /* get sparse matrix details from file */
      fscanf(file, "%10d%10d%10d%10d\n", &ftype, &a[0], &a[1], &isym);
    }
    if (*format == DDL_VEC_FMT) {
      /* get vector details from file */
      fscanf(file, "%10d%10d\n", &a[0], &ftype);
    }
    /* find MPI_Datatype equivalent to DDL integer ftype */
    DDL_MPItype(ftype, type);
#ifndef HAVE_FPOS_T /* gcc 2.5.7 compiler not know fsetpos? */
    fseek(file, fpos, 0); /* should be SEEK_SET */
#else
    fsetpos(file, &fpos);
#endif
  }
  return ierr;
}

/* called by all processes in communicator
   Parameters:
     file: INOUT    file handle
     format: OUT    format of next object in file; will be one of: 
                       DDL_TRIAD_ROW: row ordered triad format
		       DDL_TRIAD_COL: column ordered triad format
		       DDL_VEC_FMT: vector format
     type: OUT      MPI data type of elements of object
     a: OUT         extra info for object
                    for dense vector:
		       0: number of elements in vector
		    for sparse matrix:  
		       0: number of rows (columns) of square matrix
		       1: number of nonzeros in matrix
     comm: IN       MPI communicator of processes involved
   Status: H1 */
int DDL_Fileformat(FILE *file, int *format, MPI_Datatype *type, 
		   int a[], MPI_Comm comm)
{
  int ierr;
  int ftype;

  ierr = DDL_SUCCESS;
  
  /* only read file on host */
  if (DDL_Ishost()) {
    ierr = DDL_Fileformat_host(file, format, type, a);
  }

  /* tell all nodes whether to continue */
  MPI_Bcast(&ierr, 1, MPI_INT, 0, comm);

  /* exit if error */
  if (ierr != DDL_SUCCESS) return ierr;

  /* OK, we can continue */
  /* tell all nodes the type */
  /* can't communicate MPI_Datatype, so find equivalent DDL integer type */
  DDL_DDLtype(*type, &ftype);
  MPI_Bcast(a, 2, MPI_INT, 0, comm);
  MPI_Bcast(&ftype, 1, MPI_INT, 0, comm);
  MPI_Bcast(format, 1, MPI_INT, 0, comm);
  /* find MPI_Datatype */
  DDL_MPItype(ftype, type);

  return ierr;
}

/* sequential sparse matrix format converters */

/* column packed (DDL_PCK_COL) to column ordered triad (DDL_TRIAD_COL) 
   converter
   Parameters:
     n: IN          number of rows (columns) of matrix
     nelt: IN       number of nonzeros in matrix
     ia,ja,a: INOUT sparse matrix
     type: IN       MPI data type of elements of coefficient vector
     isym: IN       whether matrix is symmetric (not used)
                        0: not symmetric
                        1: symmetric
   Status: H1 */
int DDL_Coltotriad(int n, int nelt, 
		   int ia[], int ja[], void *a, MPI_Datatype type, int isym)
{
  int ierr;
  int i, idx, ibgn, iend, iold;
  int *jaold;
  /* copy ja to jaold */
  jaold = malloc((n+1)*sizeof(int));
  if (jaold == 0) {
    ierr = DDL_Error(DDL_ERR_MEMORY, DDL_ERR_HARD, DDL_COLTOTRIAD);
    return ierr;
  }
  for (i=0;i<=n;i++) {
    jaold[i] = ja[i];
  }
  for (idx=0;idx<n;idx++) {
    ibgn = jaold[idx];
    iend = jaold[idx+1]-1;
    for (i=ibgn-1;i<iend;i++) {
      ja[i] = idx+1;
    }
  }
  free(jaold);
  return DDL_SUCCESS;
}

/* row packed (DDL_PCK_ROW) to row ordered triad (DDL_TRIAD_COL) converter
   Parameters:
     n: IN          number of rows (columns) of matrix
     nelt: IN       number of nonzeros in matrix
     ia,ja,a: INOUT sparse matrix
     type: IN       MPI data type of elements of coefficient vector
     isym: IN       whether matrix is symmetric (not used)
                        0: not symmetric
                        1: symmetric
   Status: H1 */
int DDL_Rowtotriad(int n, int nelt, 
		   int ia[], int ja[], void *a, MPI_Datatype type, int isym)
{
  int ierr;
  /* just call column to triad converter with row column vectors
     interchanged */
  ierr = DDL_Coltotriad(n, nelt, ja, ia, a, type, isym);
  return ierr;
}

/* column ordered triad (DDL_TRIAD_COL) to 
   column packed (DDL_PCK_COL) converter
   Parameters:
     n: IN          number of rows (columns) of matrix
     nelt: IN       number of nonzeros in matrix
     ia,ja,a: INOUT sparse matrix
     type: IN       MPI data type of elements of coefficient vector
     isym: IN       whether matrix is symmetric (not used)
                        0: not symmetric
                        1: symmetric
   Status: H1 */
int DDL_Triadtocol(int n, int nelt, 
		   int ia[], int ja[], double a[], MPI_Datatype type, int isym)
{
  int one = 1;
  int icol, ibgn, iend, i, j, itemp;
  double temp;
  DDL_qs2i1d(ja, ia, a, nelt, 1);
/*  F77(qs2i1d)(ja, ia, a, &nelt, &one);*/

  /* loop over each column to see where the column indices change 
     in the column index array ja.  This marks the beginning of the
     next column. */
  ja[0] = 1;
  for (icol=1;icol<n;icol++) {
    for (j=ja[icol-1];j<=nelt;j++) {
      if(ja[j-1] != icol) {
        ja[icol] = j;
        break;
      }
    }
  }
  ja[n] = nelt+1;

  /* now loop through the ia array
     sorting the column in ascending order. */
  for (icol=0;icol<n;icol++) {
    ibgn = ja[icol];
    iend = ja[icol+1]-1;
    if (ibgn < iend) {
      for (i=ibgn-1;i<iend;i++) {
        for (j=i+1;j<iend;j++) {
          if (ia[i] > ia[j]) {
            itemp = ia[i];
            ia[i] = ia[j];
            ia[j] = itemp;
            temp = a[i];
            a[i] = a[j];
            a[j] = temp;
          }
        }
      }
    }
  }
  return DDL_SUCCESS;
}

/* row ordered triad (DDL_TRIAD_ROW) to row packed (DDL_PCK_ROW) converter
   Parameters:
     n: IN          number of rows (columns) of matrix
     nelt: IN       number of nonzeros in matrix
     ia,ja,a: INOUT sparse matrix
     type: IN       MPI data type of elements of coefficient vector
     isym: IN       whether matrix is symmetric (not used)
                        0: not symmetric
                        1: symmetric
   Status: H1 */
int DDL_Triadtorow(int n, int nelt, 
		   int ia[], int ja[], void *a, MPI_Datatype type, int isym)
{
  int ierr;
  /* just call triad to col converter with row and column vectors 
     interchanged */
  ierr = DDL_Triadtocol(n, nelt, ja, ia, a, type, isym);
  return ierr;
}

/* convert between MPI and DDL element datatypes.
   file format and communications uses DDL datatypes (just integers),
   MPI datatypes used everywhere else. */

/* find DDL datatype for given MPI datatype
   Parameters:
     MPItype: IN    MPI data type
     DDLtype: OUT   DDL data type
   Status: L1 */
int DDL_DDLtype(MPI_Datatype MPItype, int *DDLtype)
{
  if (MPItype == MPI_INT) *DDLtype = DDL_INT;
  else if (MPItype == MPI_FLOAT) *DDLtype = DDL_FLOAT;
  else if (MPItype == MPI_DOUBLE) *DDLtype = DDL_DOUBLE;
  return DDL_SUCCESS;
}

/* find MPI datatype for given DDL datatype
   Parameters:
     DDLtype: IN    DDL data type
     MPItype: OUT   MPI data type
   Status: L1 */
int DDL_MPItype(int DDLtype, MPI_Datatype *MPItype)
{
  if (DDLtype == DDL_INT) *MPItype = MPI_INT;
  else if (DDLtype == DDL_FLOAT) *MPItype = MPI_FLOAT;
  else if (DDLtype == DDL_DOUBLE) *MPItype = MPI_DOUBLE;
  return DDL_SUCCESS;
}

/* skip over sparse matrix data values in input file
   to be called only on the host process
   Parameters:
     file: INOUT    input file
   Status: H1 */
int DDL_Skip_host_sparse(FILE *file)
{
  int i, mxnelt, ftype, n, nelt, type, isym, format;
  int ia, ja;
  double a;

  /* read in the information heading */
  fscanf(file, "%10d\n", &format);
  fscanf(file, "%10d%10d%10d%10d\n", &ftype, &n, &nelt, &isym);
  
  /* read in the matrix of non-zeros */
  for (i=0;i<nelt;i++) {
    fscanf(file, "%d%d%le\n", &ia, &ja, &a);
  }
  return DDL_SUCCESS;
}

/* skip over dense vector data values in input file
   to be called only on the host process
   Parameters:
     file: INOUT    input file
   Status: H1 */
int DDL_Skip_host_vector(FILE *file)
{
  int ierr;
  int fformat, ftype, n, i;
  MPI_Datatype type;
  double data;

  /* check file parameters */
  ierr = DDL_SUCCESS;
  if (file == NULL) {
    ierr = DDL_Error(DDL_ERR_FILE, DDL_ERR_HARD, DDL_SKIP_HOST_VECTOR);
  }
  /* check file contains vector in correct format */
  else {
    fscanf(file, "%10d\n", &fformat);
    if (fformat != DDL_VEC_FMT) {
      ierr = DDL_Error(DDL_ERR_FILEFORMAT, 
		       DDL_ERR_HARD, DDL_SKIP_HOST_VECTOR);
    }
    else {
      /* get vector details from file */
      fscanf(file, "%10d%10d\n", &n, &ftype);
      DDL_MPItype(ftype, &type);
      if (type == MPI_FLOAT) {
	  for (i=0;i<n;i++)
	      fscanf(file, " %e", &data);
      }
      else if (type == MPI_DOUBLE) {
	  for (i=0;i<n;i++)
	      fscanf(file, "%le", &data);
      }
      else if (type == MPI_INT) {
	  for (i=0;i<n;i++)
	      fscanf(file, " %d", &data);
      }
      return DDL_SUCCESS;
    }
  }
  return ierr;
}

/* find time 
   take the difference between two calls to ddl_gettime to
   get time taken in seconds
   Status: H2 */
#include <time.h>
double DDL_Gettime()
{
#ifdef WALL_TIME
  /* wall clock time */
  /* static time set first time gettime is called */
  static time_t ddl_tzero = -1; 
  time_t tt;
  tt = time(NULL);
  if (ddl_tzero == -1) ddl_tzero = tt;  
#ifdef HAVE_DIFFTIME
  return difftime(tt, ddl_tzero);
#else
  return 0;
#endif
#endif

#ifdef ELAPSED_TIME 
  /* elapsed time */
  clock_t ct;
  ct = clock();
#ifdef STDC_HEADERS
  return (ct/(float)CLOCKS_PER_SEC);
#else /* guess CLOCKS_PER_SEC value=1000000 */
  return (ct/(float)1000000);
#endif
#endif
}

/* low level procedures to find distribution information for objects
   without performing communication. */

/* calculate offset of block on process proc
   Parameters:
     n: IN          size of object dimension
     block: IN      blocking factor used to distribute object
     nproc: IN      number of processes for object
     proc: IN       process for which to calculate offset
   Return:          offset of block
   Status: L1 */
int DDL_Find_offset(int n, int block, int nproc, int proc)
{
  int general_size;
  general_size = block*DDL_CD(n/block, nproc);
  return DDL_MIN(proc*general_size,n);
}

/* calculate size of block on process proc
   Parameters:
     n: IN          size of object dimension
     block: IN      blocking factor used to distribute object
     nproc: IN      number of processes for object
     proc: IN       process for which to calculate size of block
   Return:          size of block
   Status: L1 */
int DDL_Find_size(int n, int block, int nproc, int proc)
{
  return DDL_Find_offset(n, block, nproc, proc+1) - 
    DDL_Find_offset(n, block, nproc, proc);
}

/* calculate number of procs used by an object;
   not all may be used by a distribution because of rounding
   Parameters:
     n: IN          size of object dimension
     block: IN      blocking factor used to distribute object
     nproc: IN      maximum number of processes for object
   Return:          number of processes actually used for object
   Status: L1 */
int DDL_Find_proc(int n, int block, int nproc)
{
  int general_size;
  general_size = block*DDL_CD(n/block, nproc);
  return DDL_CD(n, general_size);
}

