/* $Id: ctr.c,v 1.3 1995/08/23 12:04:31 tim Exp $ */

/* Copyright 1995 IASC */

/* sparse triangular solve using permutation vectors on matrix */

#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

#include "ddl.h"

/* constants for sparse data structures */
#define maxn                       5005 
#define mxnelt                     40000
#define np                         10

#define hbfilename      "test4001.hbf"
#define infilename      "indemo"
#define outfilename     "outdemo"

#define MAX_ERR         1.0E-6

int main(int argc, char **argv)
{
  MPI_Comm comm;                 /* communicator for processes */
  int block;
  int ierr;
  int maxnelt;
  int nelt, nr;
  int n, i, j, size, offset;
  int format, itype;
  int sizes[3];                  /* object type parameters from file */
  MPI_Datatype type;
  double *data, *a;
  int *ia, *ja;
  double res, alpha, beta, err;
  FILE *infile, *outfile;
  int fformat;
  int dummy, isym;
  int *px, *ipx, *py, *ipy;
  double *x, *b;
  int nx, ny, nz, tmp;
  DDL_Vec x_ddl, b_ddl, true_ddl, px_ddl, py_ddl;
  DDL_Spa a_ddl;
  int xperm, yperm, upper, rot, k;

  /* limit CPU run-time for each node 
     optional, but probably wise! 
     (time given in seconds) 
     */
  DDL_Set_cpu_limit(1*60);

  /* initialise MPI and sparse DDL*/
  MPI_Init(&argc, &argv);
  DDL_Init_sparse(MPI_COMM_WORLD, &comm);

  if (DDL_Ishost()) {
    CreateInput(&xperm, &yperm, &upper);
  }

  /* tell all processes if there are permutations and upper/lower */
  ierr = MPI_Bcast(&xperm, 1, MPI_INT, 0, comm);
  ierr = MPI_Bcast(&yperm, 1, MPI_INT, 0, comm);
  ierr = MPI_Bcast(&upper, 1, MPI_INT, 0, comm);

  px_ddl = DDL_VECTOR_NULL;
  py_ddl = DDL_VECTOR_NULL;

  /* open input and output files */
  ierr = DDL_Open(&infile, infilename, DDL_READ_MODE, comm);
  ierr = DDL_Open(&outfile, outfilename, DDL_WRITE_MODE, comm);

  /* read in the information heading */
  ierr = DDL_Fileformat(infile, &fformat, &type, sizes, comm);
  n = sizes[0];
  nelt = sizes[1];
  format = DDL_PCK_ROW;
  block = 1;

  /* set up size vector */
  sizes[0] = n;
  sizes[1] = n;
  sizes[2] = mxnelt;

  /* create DDL sparse matrix */
  a_ddl = DDL_SPARSE_NULL;
  ierr = DDL_Create_sparse(sizes, type, format, block, comm, &a_ddl);

  /* read data from file to sparse object */
  ierr = DDL_Read(a_ddl, infile);

  /* create rhs vector */
  ierr = DDL_Fileformat(infile, &fformat, &type, sizes, comm);
  n = sizes[0];
  b_ddl = DDL_VECTOR_NULL;
  ierr = DDL_Create_vector(n, type, block, comm, &b_ddl);
  /* fill rhs vector with data from file */
  ierr = DDL_Read(b_ddl, infile);

  /* create true solution vector */
  ierr = DDL_Fileformat(infile, &fformat, &type, sizes, comm);
  n = sizes[0];
  true_ddl = DDL_VECTOR_NULL;
  ierr = DDL_Create_vector(n, type, block, comm, &true_ddl);
  /* fill vector with data from file */
  ierr = DDL_Read(true_ddl, infile);

  /* create solution vector */
  x_ddl = DDL_VECTOR_NULL;
  ierr = DDL_Create_vector(n, type, block, comm, &x_ddl);

  /* create row permutation vector */
  if (xperm != 0) {
    ierr = DDL_Fileformat(infile, &fformat, &type, sizes, comm);
    n = sizes[0];
    ierr = DDL_Create_vector(n, type, block, comm, &px_ddl);
    /* fill vector with data from file */
    ierr = DDL_Read(px_ddl, infile);
  }

  /* create column permutation vector */
  if (yperm != 0) {
    ierr = DDL_Fileformat(infile, &fformat, &type, sizes, comm);
    n = sizes[0];
    ierr = DDL_Create_vector(n, type, block, comm, &py_ddl);
    /* fill vector with data from file */
    ierr = DDL_Read(py_ddl, infile);
  }

  /* copy permutation vectors to sparse object permutation vectors */
  if (xperm != 0 || yperm != 0) 
    ierr = DDL_Setperm_sparse(a_ddl, px_ddl, py_ddl);

  /* say matrix is triangular */
  if (upper == 1) ierr = DDL_Setsubtype_sparse(a_ddl, DDL_UPPER);
  else  ierr = DDL_Setsubtype_sparse(a_ddl, DDL_LOWER);

  /* create triangular solve comms. data */
  ierr = DDL_Trisolve_comm(a_ddl);

  /* now perform triangular solve */
  ierr = DDL_Trisolve(a_ddl, x_ddl, b_ddl);

  /* write result to file */
  ierr = DDL_Write(x_ddl, outfile);

  /* check result vector is correct */
  ierr = DDL_Check_err(x_ddl, true_ddl, &err);
  if (DDL_Ishost())
    if (err > MAX_ERR) 
      printf("Error: maximum relative error is %le.\n", err);
    else
      printf("Result correct to within %le.\n", MAX_ERR);

  /* remove DDL objects */
  DDL_Free(a_ddl);
  DDL_Free(x_ddl);
  DDL_Free(b_ddl);
  DDL_Free(true_ddl);
  if (xperm != 0) DDL_Free(px_ddl);
  if (yperm != 0) DDL_Free(py_ddl);

  /* close input and output files */
  ierr = DDL_Close(infile, comm);
  ierr = DDL_Close(outfile, comm);

  /* tidy up and exit */
  DDL_Finalize_sparse(comm);
  MPI_Finalize();    
  return;
}

/* created input file for triangular solve;
   file contains:
   sparse matrix
   rhs vector
   soln vector
   x permutation vector
   y permutation vector
   */
int CreateInput(int *xperm, int *yperm, int *upper)
{
  FILE *infile, *outfile;
  int format;
  MPI_Datatype type;
  int ia[mxnelt], ja[mxnelt];
  double a[mxnelt];
  int nelt, job;
  int n, i, j, isym, size;
  double soln[maxn], rhs[maxn], xinit[maxn];
  int *px, *ipx, *py, *ipy;
  double *x, *b;
  int nx, ny, nz;
  int pt, rot, k;
  int ierr = DDL_SUCCESS;
  char ch[2];

  n = 125; 
  nx = 5;
  ny = 5;
  nz = 5;
  type = MPI_DOUBLE;

  ipx = NULL;
  ipy = NULL;

  /* make vectors on host--easier sequentially! */
  /* set up data so that result is always the same */
  px = malloc(sizeof(int)*n);
  py = malloc(sizeof(int)*n);
  x = malloc(sizeof(double)*n);
  b = malloc(sizeof(double)*n);
  /* set up permutation vectors */
  printf("Enter permutation number for px (0-5): ");
  scanf("%d", xperm);
  if (*xperm != 0) {
    ipx = malloc(sizeof(int)*n);
    set_perm(*xperm, px, ipx, n, nx, ny, nz);
  }
  printf("Enter permutation number for py (0-5): ");
  scanf("%d", yperm);
  if (*yperm != 0) {
    ipy = malloc(sizeof(int)*n);
    set_perm(*yperm, py, ipy, n, nx, ny, nz);
  }
  /* set up matrix */
  printf("Upper or lower triangular matrix (U/L): ");
  scanf("%1s", ch);
  if (ch[0] == 'L' || ch[0] == 'l' ) {
    *upper = 0;
    /* lower triangular matrix with a(i,j)=n/j */
    /* triad format */
    k = 0;
    for (i=0;i<n;i++) {
      for (j=0;j<=i;j++) {
	ia[k] = i+1; /* ia stored as local F indices */
	ja[k] = j+1;
	a[k] = (double)n/(j+1);
	k = k+1;
      }
    }
    nelt = k;
  }
  else {
    /* upper triangular matrix with a(i,j)=n/j */
    /* triad format */
    *upper = 1;
    k = 0;
    for (i=0;i<n;i++) {
      for (j=i;j<n;j++) {
	ia[k] = i+1; /* ia stored as local F indices */
	ja[k] = j+1;
	a[k] = (double)n/(n-j);
	k = k+1;
      }
    }
    nelt = k;
  }
  /* convert to row packed format */
  isym = 0;
  ierr = DDL_Triadtorow(n, nelt, ia, ja, a, type, isym);
  format = DDL_PCK_ROW;
  /* permute matrix---very inefficiently!! */
  if (*xperm != 0 || *yperm != 0)
    ierr = DDL_Lpermute_sparse(ia, ja, a, nelt, n, n, type, format, ipx, ipy);

  /* convert row to triad */
  ierr = DDL_Rowtotriad(n, nelt, ia, ja, a, type, isym);
  /* initialise data vectors */
  if (*upper == 1) {
    for (i=0;i<n;i++) x[n-1-i] = i+1; /* solution should be this*/
    for (i=0;i<n;i++) b[n-1-i] = n*(i+1); /* rhs */
  }
  else {
    for (i=0;i<n;i++) x[i] = i+1; /* solution should be this*/
    for (i=0;i<n;i++) b[i] = n*(i+1); /* rhs */
  }
  /* write out matrix as row ordered triad format */
  /* this data file will be used for rest of demo */
  format = DDL_TRIAD_ROW;
  ierr = DDL_Open_host(&outfile, infilename, DDL_WRITE_MODE);
  ierr = DDL_Out_sparse(outfile, n, nelt, ia, ja, a, type,
			isym, format);
  ierr = DDL_Out_vector(outfile, b, MPI_DOUBLE, &n);
  ierr = DDL_Out_vector(outfile, x, MPI_DOUBLE, &n);
  
  if (*xperm != 0) ierr = DDL_Out_vector(outfile, px, MPI_INT, &n);
  if (*yperm != 0) ierr = DDL_Out_vector(outfile, py, MPI_INT, &n);
  ierr = DDL_Close_host(outfile);

  /* free memory */
  free(px);
  free(py);
  if (*xperm != 0) free(ipx);
  if (*yperm != 0) free(ipy);
  free(x);
  free(b);

  return ierr;
}

int set_perm(int pt, int *pv, int *ipv, int n, int nx, int ny, int nz)
{
  int i, rot, ierr;
  switch (pt) {
  case 1:
    /* identity permutation */
    for (i=0;i<n;i++)
      pv[i] = i;
    break;
  case 2:
    /* red/black permutation */
    ierr = DDL_calc_perm(nx, ny, nz, pv);
      break;
  case 3:
    /* rotate by rot */
    rot = 1;
    for (i=0;i<n;i++)
      pv[i] = (i+rot)%n;
    break;
  case 4:
    /* invert vector */
    for (i=0;i<n;i++)
      pv[i] = n-i-1;
    break;
  case 5:
    /* swap single pair (extremes) */
    for (i=0;i<n;i++)
      pv[i] = i;
    pv[0] = n-1;
    pv[n-1] = 0;
    break;
  default:
    /* identity permutation */
    for (i=0;i<n;i++)
      pv[i] = i;
    break;
  }      
  DDL_Calc_invperm(pv, ipv, n);
}

