/* $Id: thread1.c,v 1.2 1995/08/21 10:24:10 tim Exp $ */

/* DDL parallel thread demo 1 */
/* for Solaris */

/* Copyright 1994 IASC */

#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <thread.h>
#include <math.h>

#include "ddl.h"

/* constants for sparse data structures */
#define maxn                       5005 
#define mxnelt                     40000
#define np                         10

#define hbfilename      "test4001.hbf"
#define infilename      "indemo"
#define outfilename     "outdemo"

#define block 1 /* blocking not implemented */

/* function header */
int work(void);

/* global data for threads */
DDL_Spa m, m2;
DDL_Vec x, y, x2, y2;
double alpha, beta;

int main(int argc, char **argv)
{
  MPI_Comm comm;                 /* communicator for processes */
  int ierr;
  int maxnelt;
  int nelt;
  int n, i, j, size, offset;
  int format, itype;
  int sizes[3];                  /* object type parameters from file */
  MPI_Datatype type;
  double *data, *a;
  int *ia, *ja;
  double res, error;
  FILE *infile, *outfile;
  int fformat;
  int dummy;

  /* thread variables */
  thread_t tid;
  int arg = NULL;
  long thr_flags = NULL;
  int *status;

  /* initialise MPI and sparse DDL*/
  MPI_Init(&argc, &argv);
  DDL_Init_sparse(MPI_COMM_WORLD, &comm);

  if (DDL_Ishost()) {
    CreateInput();
  }

  /* open input and output files */
  ierr = DDL_Open(&infile, infilename, DDL_READ_MODE, comm);
  ierr = DDL_Open(&outfile, outfilename, DDL_WRITE_MODE, comm);

  /* read in the information heading */
  ierr = DDL_Fileformat(infile, &fformat, &type, sizes, comm);
  n = sizes[0];
  nelt = sizes[1];
  format = DDL_PCK_ROW;
  
  /* set up size vector */
  sizes[0] = n;
  sizes[1] = n;
  sizes[2] = mxnelt;

  /* create DDL sparse matrix */
  m = DDL_SPARSE_NULL;
  ierr = DDL_Create_sparse(sizes, type, format, block, comm, &m);
  m2 = DDL_SPARSE_NULL;
  ierr = DDL_Create_sparse(sizes, type, format, block, comm, &m2);

  /* read data from file to sparse object */
  ierr = DDL_Read(m, infile);

  /* create vector */
  ierr = DDL_Fileformat(infile, &fformat, &type, sizes, comm);
  n = sizes[0];
  x = DDL_VECTOR_NULL;
  ierr = DDL_Create_vector(n, type, block, comm, &x);
  y = DDL_VECTOR_NULL;
  ierr = DDL_Create_vector(n, type, block, comm, &y);
  x2 = DDL_VECTOR_NULL;
  ierr = DDL_Create_vector(n, type, block, comm, &x2);
  y2 = DDL_VECTOR_NULL;
  ierr = DDL_Create_vector(n, type, block, comm, &y2);

  /* fill vector with data from file */
  ierr = DDL_Read(x, infile);

  /* initialize vector y */
  ierr = DDL_Get_vector(y, &data);
  ierr = DDL_Size_vector(y, &size);
  for (i=0;i<size;i++) data[i] = 1.0;

  /* initialize constants on all pocesses */
  alpha = 1.0;
  beta = 0.0;

  /* copy distributed objects */
  ierr = DDL_Copy(m, m2);
  ierr = DDL_Copy(x, x2);
  ierr = DDL_Copy(y, y2);

  /* run parallel threads to perform matrix vector multiply */
  ierr = thr_create(NULL, 0, work, &arg, thr_flags, &tid);
  printf ("%d\n", ierr);

  /* create matrix-vector multiply data */
  ierr = DDL_Smv_comm(m);

  /* now perform sparse matrix vector multiply */
  ierr = DDL_Smv(m, x, y, alpha, beta); 

  /* write data to file */
/*  ierr = DDL_Write(m, outfile);*/
  ierr = DDL_Write(y, outfile);

  /* remove DDL objects */
  DDL_Free(m);
  DDL_Free(x);
  DDL_Free(y);

  /* close input and output files */
  ierr = DDL_Close(infile, comm);
  ierr = DDL_Close(outfile, comm);

  /* synchronise threads */
  ierr = thr_join(tid, NULL, &status);
  printf ("%d\n", ierr);
  printf ("%d\n", *status);

  /* check results */
  check(infilename, outfilename, &error);
  printf("Max error: %le\n",error);

  /* tidy up and exit */
  DDL_Finalize_sparse(comm);
  MPI_Finalize();    
  return;
}

/* convert Harwell-Boeing file to row-packed triad file */
int CreateInput(void)
{
  FILE *infile, *outfile;
  int ierr, format;
  MPI_Datatype type;
  int ia[mxnelt], ja[mxnelt];
  double a[mxnelt];
  int nelt, job;
  int n, i, isym, size;
  double soln[maxn], rhs[maxn], xinit[maxn];

  /* read in sparse matrix from file */
  job = 3;
  n = maxn;
  nelt = mxnelt;
  /* read in as column gathered */
  ierr = DDL_Open_host(&infile, hbfilename, DDL_READ_MODE);
  ierr = DDL_Inhb(infile, &n, &nelt, ia, ja, a, 
		  &isym, soln, rhs, xinit, &job);
  ierr = DDL_Close_host(infile);
  type = MPI_DOUBLE;
  /* convert to triad */
  ierr = DDL_Coltotriad(n, nelt, ia, ja, a, type, isym);
  /* convert triad to row */
  ierr = DDL_Triadtorow(n, nelt, ia, ja, a, type, isym);
  /* convert row to triad */
  ierr = DDL_Rowtotriad(n, nelt, ia, ja, a, type, isym);
  /* write out as row ordered triad format */
  /* this data file will be used for rest of demo */
  format = DDL_TRIAD_ROW;
  ierr = DDL_Open_host(&outfile, infilename, DDL_WRITE_MODE);
  ierr = DDL_Out_sparse(outfile, n, nelt, ia, ja, a, type,
			isym, format);
  ierr = DDL_Out_vector(outfile, rhs, MPI_DOUBLE, &n);
  ierr = DDL_Close_host(outfile);
}

/* check result vector */
int check(char *ddlfilename, char *solnfilename, 
	  double *maxerr)
{
  FILE *infile, *outfile;
  int ierr, format;
  MPI_Datatype type;
  int ia[mxnelt], ja[mxnelt];
  double a[mxnelt];
  int nelt, job;
  int n, i, isym, size;
  double soln[maxn], correct[maxn], x[maxn], error;
  /* get calculated solution vector */
  ierr = DDL_Open_host(&infile, solnfilename, DDL_READ_MODE);
  n = maxn;
  ierr = DDL_In_vector(infile, soln, &type, &n);
  ierr = DDL_Close_host(infile);
  /* read in sparse matrix from file */
  ierr = DDL_Open_host(&infile, ddlfilename, DDL_READ_MODE);
  n = maxn;
  nelt = mxnelt;
  ierr = DDL_In_sparse(infile, &n, &nelt, ia, ja, a, 
		       &type, &isym, &format);
  ierr = DDL_Triadtorow(n, nelt, ia, ja, a, type, isym);
  format = DDL_PCK_ROW;
  /* fill vector with data from file */
  n = maxn;
  ierr = DDL_In_vector(infile, x, &type, &n);
  ierr = DDL_Close_host(infile);

  /* set vector to 1 */
  for (i=0;i<n;i++) correct[i] = 1.0;

  /* calculate sparse matrix-vector multiply */
  ierr = DDL_Smv_seq(n, n, ia, ja, 
		     a, type,
		     nelt, format, x, correct,
		     alpha, beta);

  /* check for differences */
  *maxerr = 0.0;
  for (i=0;i<n;i++) {
      error = fabs(fabs(soln[i])-fabs(correct[i]));
      if (error > *maxerr) {
	  *maxerr = error;
      }
  }
  return DDL_SUCCESS;
}

/* thread procedure */
int work(void)
{
  MPI_Comm comm;
  int ierr;
  FILE *outfile;
#define outfilename2 "outdemo2\0"

  ierr = DDL_Comm(m2, &comm);
  /* open output file */
  ierr = DDL_Open(&outfile, outfilename2, DDL_WRITE_MODE, comm);

  /* create matrix-vector multiply data */
  ierr = DDL_Smv_comm(m2);

  /* now perform sparse matrix vector multiply */
  ierr = DDL_Smv(m2, x2, y2, alpha, beta); 

  /* write data to file */
  ierr = DDL_Write(y2, outfile);

  /* remove DDL objects */
  DDL_Free(m2);
  DDL_Free(x2);
  DDL_Free(y2);

  /* close output files */
  ierr = DDL_Close(outfile, comm);

  thr_exit(&ierr);
}
