/* $Id: cdemo.c,v 1.6 1995/08/23 11:11:47 tim Exp $ */

/* Copyright 1994 IASC */

/* A simple example of Sparse-DDL user programs in C */

/* some standard header files... */
#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

/* always include the Sparse-DDL header file */
#include "ddl.h"

/* constants for sparse data structures */
#define maxn                       5005 
#define mxnelt                     40000
#define np                         10

/* file names for I/O */
#define hbfilename      "../parsim/data/test4001.hbf"
#define infilename      "indemo"
#define outfilename     "outdemo"

#define block 1 /* block size for distribution */

int main(int argc, char **argv)
{
  MPI_Comm comm;
  DDL_Spa m;                     /* sparse matrix object */
  DDL_Vec x, y;                  /* vector objects */
  int ierr;
  int maxnelt;
  int nelt;
  int n, i, j, size, offset;
  int format, itype;
  int sizes[3];
  MPI_Datatype type;
  double *data, *a;
  int *ia, *ja;
  double res, alpha, beta;
  FILE *infile, *outfile;
  int fformat;
  int dummy;

  /* initialise MPI and Sparse-DDL 
     these two lines must be included at the start of all user programs
     DDL_Init_sparse sets up an MPI communicator, comm, compatible with 
     the DDL from MPI_COMM_WORLD. */
  MPI_Init(&argc, &argv);
  DDL_Init_sparse(MPI_COMM_WORLD, &comm);

  /* create DDL input file on host process */
  if (DDL_Ishost()) {
    CreateInput();
  }

  /* open input and output files (on all processes) */
  ierr = DDL_Open(&infile, infilename, DDL_READ_MODE, comm);
  ierr = DDL_Open(&outfile, outfilename, DDL_WRITE_MODE, comm);

  /* read in the header information about first object in file
     which is a sparse matrix */
  ierr = DDL_Fileformat(infile, &fformat, &type, sizes, comm);
  n = sizes[0]; /* number of rows of matrix */
  format = DDL_PCK_ROW; /* we want this format for distributed object */
  
  /* set up size vector for DDL_Create_sparse */
  sizes[0] = n;
  sizes[1] = n;
  sizes[2] = mxnelt; /* maximum number of nonzeros to allocate space for */

  /* create DDL sparse matrix object */
  ierr = DDL_Create_sparse(sizes, type, format, block, comm, &m);

  /* read data from file to sparse object */
  ierr = DDL_Read(m, infile);

  /* create vector object for next vector in data file */
  ierr = DDL_Fileformat(infile, &fformat, &type, sizes, comm);
  n = sizes[0];
  ierr = DDL_Create_vector(n, type, block, comm, &x);

  /* create second identical vector */
  ierr = DDL_Create_vector(n, type, block, comm, &y);

  /* fill first vector with data from file */
  ierr = DDL_Read(x, infile);

  /* initialize vector y */
  ierr = DDL_Get_vector(y, &data); /* get pointer to vector data values */
  ierr = DDL_Size_vector(y, &size); /* find out size of local vector block */ 
  for (i=0;i<size;i++) data[i] = 0.0; /* initialise values to zero */

  /* initialize constants on all pocesses */
  alpha = 1.0;
  beta = 1.0;
  
  /* demonstration of axpy operation on distributed objects... */
  ierr = DDL_Axpy(alpha, x,y);

  /* create matrix-vector multiply communication pattern data */
  ierr = DDL_Smv_comm(m);

  /* now perform sparse matrix vector multiply */
  ierr = DDL_Smv(m, x, y, alpha, beta); 

  /* write result vector to file */
  ierr = DDL_Write(y, outfile);

  /* remove DDL objects */
  DDL_Free(m);
  DDL_Free(x);
  DDL_Free(y);

  /* close input and output files */
  ierr = DDL_Close(infile, comm);
  ierr = DDL_Close(outfile, comm);

  /* print OK message */
  if (DDL_Ishost()) {
    printf("Demonstration program finished OK (phew!)\n");
  }

  /* tidy up and exit
     these calls must be included at the end of every DDL program */
  DDL_Finalize_sparse(comm);
  MPI_Finalize();    
  return;
}

/* convert Harwell-Boeing file to row-packed triad file 
   only executed by host process with file I/O capability */
int CreateInput(void)
{
  FILE *infile, *outfile;
  int ierr, format;
  MPI_Datatype type;
  int ia[mxnelt], ja[mxnelt];
  double a[mxnelt];
  int nelt, job;
  int n, i, isym, size;
  double soln[maxn], rhs[maxn], xinit[maxn];

  /* open Harwell-Boeing data file */
  ierr = DDL_Open_host(&infile, hbfilename, DDL_READ_MODE);

  /* read in Harwell-Boeing file format to DDL_PCK_COL, 
     column packed format matrix
     (also reads in rhs vector, soln and xinit) */
  job = 3;
  n = maxn;
  nelt = mxnelt;
  ierr = DDL_Inhb(infile, &n, &nelt, ia, ja, a, 
		  &isym, soln, rhs, xinit, &job);

  /* close input file */
  ierr = DDL_Close_host(infile);

  /* data type of coefficient elements */
  type = MPI_DOUBLE;

  /* convert column packed format, DDL_PCK_COL, 
     to row gathered triad, DDL_TRIAD_ROW, for output to DDL data file */ 
  /* first, convert to column gathered triad, DDL_TRIAD_COL */
  ierr = DDL_Coltotriad(n, nelt, ia, ja, a, type, isym);
  /* then, convert column gathered triad to row packed format, DDL_PCK_ROW */
  ierr = DDL_Triadtorow(n, nelt, ia, ja, a, type, isym);
  /* finally, convert row packed format to row gathered triad, DDL_TRIAD_ROW */
  ierr = DDL_Rowtotriad(n, nelt, ia, ja, a, type, isym);

  /* write out matrix as row ordered triad format */
  /* this data file will be used for rest of demo */
  format = DDL_TRIAD_ROW;
  ierr = DDL_Open_host(&outfile, infilename, DDL_WRITE_MODE);
  ierr = DDL_Out_sparse(outfile, n, nelt, ia, ja, a, type,
			isym, format);
  /* write out rhs vector as well */
  ierr = DDL_Out_vector(outfile, rhs, type, &n);
  ierr = DDL_Close_host(outfile);
}
