/* $Id: cmv.c,v 1.2 1995/08/23 11:18:11 tim Exp $ */

/* Copyright 1995 IASC */

/* sparse matrix-vector multiply using permutation vectors for matrix */

#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

#include "ddl.h"

/* constants for sparse data structures */
#define maxn                       5005 
#define mxnelt                     40000
#define np                         10

#define hbfilename      "test4001.hbf"
#define infilename      "indemo"
#define outfilename     "outdemo"

#define MAX_ERR         1.0E-6

int main(int argc, char **argv)
{
  MPI_Comm comm;                 /* communicator for processes */
  int block;
  int ierr;
  int maxnelt;
  int nelt, nr;
  int n, i, j, size, offset;
  int format, itype;
  int sizes[3];                  /* object type parameters from file */
  MPI_Datatype type;
  double *data, *a;
  int *ia, *ja;
  double res, alpha, beta, err;
  FILE *infile, *outfile;
  int fformat;
  int dummy, isym;
  int *px, *ipx, *py, *ipy;
  double *x, *y;
  int nx, ny, nz, tmp;
  DDL_Vec x_ddl, y_ddl, true_ddl, px_ddl, py_ddl;
  DDL_Spa a_ddl;
  int pt, rot, k;

  /* limit CPU run-time for each node 
     optional, but probably wise! 
     (time given in seconds) 
     */
  DDL_Set_cpu_limit(1*60);

  /* initialise MPI and sparse DDL*/
  MPI_Init(&argc, &argv);
  DDL_Init_sparse(MPI_COMM_WORLD, &comm);

  if (DDL_Ishost()) {
    CreateInput();
  }

  /* open input and output files */
  ierr = DDL_Open(&infile, infilename, DDL_READ_MODE, comm);
  ierr = DDL_Open(&outfile, outfilename, DDL_WRITE_MODE, comm);

  /* read in the information heading */
  ierr = DDL_Fileformat(infile, &fformat, &type, sizes, comm);
  n = sizes[0];
  nelt = sizes[1];
  format = DDL_PCK_ROW;
  block = 1;

  /* set up size vector */
  sizes[0] = n;
  sizes[1] = n;
  sizes[2] = mxnelt;

  /* create DDL sparse matrix */
  a_ddl = DDL_SPARSE_NULL;
  ierr = DDL_Create_sparse(sizes, type, format, block, comm, &a_ddl);

  /* read data from file to sparse object */
  ierr = DDL_Read(a_ddl, infile);

  /* create x vector */
  ierr = DDL_Fileformat(infile, &fformat, &type, sizes, comm);
  n = sizes[0];
  x_ddl = DDL_VECTOR_NULL;
  ierr = DDL_Create_vector(n, type, block, comm, &x_ddl);
  /* fill x vector with data from file */
  ierr = DDL_Read(x_ddl, infile);

  /* create y vector */
  ierr = DDL_Fileformat(infile, &fformat, &type, sizes, comm);
  n = sizes[0];
  y_ddl = DDL_VECTOR_NULL;
  ierr = DDL_Create_vector(n, type, block, comm, &y_ddl);
  /* fill y vector with data from file */
  ierr = DDL_Read(y_ddl, infile);

  /* create true solution vector */
  ierr = DDL_Fileformat(infile, &fformat, &type, sizes, comm);
  n = sizes[0];
  true_ddl = DDL_VECTOR_NULL;
  ierr = DDL_Create_vector(n, type, block, comm, &true_ddl);
  /* fill vector with data from file */
  ierr = DDL_Read(true_ddl, infile);

  /* create row permutation vector */
  ierr = DDL_Fileformat(infile, &fformat, &type, sizes, comm);
  n = sizes[0];
  px_ddl = DDL_VECTOR_NULL;
  ierr = DDL_Create_vector(n, type, block, comm, &px_ddl);
  /* fill vector with data from file */
  ierr = DDL_Read(px_ddl, infile);

  /* create column permutation vector */
  ierr = DDL_Fileformat(infile, &fformat, &type, sizes, comm);
  n = sizes[0];
  py_ddl = DDL_VECTOR_NULL;
  ierr = DDL_Create_vector(n, type, block, comm, &py_ddl);
  /* fill vector with data from file */
  ierr = DDL_Read(py_ddl, infile);

  /* copy permutation vectors to sparse object permutation vectors */
  ierr = DDL_Setperm_sparse(a_ddl, px_ddl, py_ddl);

  /* initialize constants on all pocesses */
  alpha = 1.0;
  beta = 1.0;
  
  /* create matrix-vector multiply data */
  ierr = DDL_Smv_comm(a_ddl);

  /* now perform sparse matrix vector multiply */
  ierr = DDL_Smv(a_ddl, x_ddl, y_ddl, alpha, beta);

  /* write result to file */
  ierr = DDL_Write(y_ddl, outfile);

  /* check result vector is correct */
  ierr = DDL_Check_err(y_ddl, true_ddl, &err);
  if (DDL_Ishost())
    if (err > MAX_ERR) 
      printf("Error: maximum relative error is %le.\n", err);
    else
      printf("Result correct to within %le.\n", MAX_ERR);

  /* remove DDL objects */
  DDL_Free(a_ddl);
  DDL_Free(x_ddl);
  DDL_Free(y_ddl);
  DDL_Free(true_ddl);
  DDL_Free(px_ddl);
  DDL_Free(py_ddl);

  /* close input and output files */
  ierr = DDL_Close(infile, comm);
  ierr = DDL_Close(outfile, comm);

  /* tidy up and exit */
  DDL_Finalize_sparse(comm);
  MPI_Finalize();    
  return;
}

/* created input file for sparse matrix vector multiply
   file contains:
   a sparse matrix
   x vector
   initial y vector
   result y vector
   x permutation vector
   y permutation vector
   */
int CreateInput(void)
{
  FILE *infile, *outfile;
  int format;
  MPI_Datatype type;
  int ia[mxnelt], ja[mxnelt];
  double a[mxnelt];
  int nelt, job;
  int n, i, j, isym, size;
  double soln[maxn], rhs[maxn], xinit[maxn];
  int *px, *ipx, *py, *ipy;
  double *x, *y, *res;
  int nx, ny, nz;
  int pt, rot, k;
  int ierr = DDL_SUCCESS;

  n = 125;  
  nx = 5;
  ny = 5;
  nz = 5;
  type = MPI_DOUBLE;

  /* make vectors on host--easier sequentially! */
  /* set up vectors so that result is always the same except it will
     have been permuted */
  px = malloc(sizeof(int)*n);
  py = malloc(sizeof(int)*n);
  ipx = malloc(sizeof(int)*n);
  ipy = malloc(sizeof(int)*n);
  x = malloc(sizeof(double)*n);
  y = malloc(sizeof(double)*n);
  res = malloc(sizeof(double)*n);
  /* initialise data vectors */
  for (i=0;i<n;i++) x[i] = 1.0/(i+1);
  for (i=0;i<n;i++) y[i] = (i+1); /* initial y values */
  for (i=0;i<n;i++) res[i] = 10*(i+1)*n+(i+1);
  /* set up permutation vectors */
  printf("Enter permutation number for px (1-5): ");
  scanf("%d", &pt);
  set_perm(pt, px, ipx, n, nx, ny, nz);
  printf("Enter permutation number for py (1-5): ");
  scanf("%d", &pt);
  set_perm(pt, py, ipy, n, nx, ny, nz);

  /* full matrix with a(i,j)=j*10*i */
  /* triad format */
  k = 0;
  for (i=0;i<n;i++) {
    for (j=0;j<n;j++) {
      ia[k] = i+1; /* ia stored as local F indices */
      ja[k] = j+1;
      a[k] = (j+1)*10*(i+1);
      k = k+1;
    }
  }
  nelt = k;
  /* convert to row packed format */
  isym = 0;
  ierr = DDL_Triadtorow(n, nelt, ia, ja, a, type, isym);
  format = DDL_PCK_ROW;
  /* permute matrix---very inefficiently!! */
  ierr = DDL_Lpermute_sparse(ia, ja, a, nelt, n, n, type, format, ipx, ipy);

  /* convert row to triad */
  ierr = DDL_Rowtotriad(n, nelt, ia, ja, a, type, isym);
  /* write out as row ordered triad format */
  /* this data file will be used for rest of demo */
  format = DDL_TRIAD_ROW;
  ierr = DDL_Open_host(&outfile, infilename, DDL_WRITE_MODE);
  ierr = DDL_Out_sparse(outfile, n, nelt, ia, ja, a, type,
			isym, format);
  ierr = DDL_Out_vector(outfile, x, MPI_DOUBLE, &n);
  ierr = DDL_Out_vector(outfile, y, MPI_DOUBLE, &n);
  ierr = DDL_Out_vector(outfile, res, MPI_DOUBLE, &n);
  ierr = DDL_Out_vector(outfile, px, MPI_INT, &n);
  ierr = DDL_Out_vector(outfile, py, MPI_INT, &n);
  ierr = DDL_Close_host(outfile);

  /* free memory */
  free(px);
  free(py);
  free(ipx);
  free(ipy);
  free(x);
  free(y);
  free(res);

  return ierr;
}

int set_perm(int pt, int *pv, int *ipv, int n, int nx, int ny, int nz)
{
  int i, rot, ierr;
  switch (pt) {
  case 1:
    /* identity permutation */
    for (i=0;i<n;i++)
      pv[i] = i;
    break;
  case 2:
    /* red/black permutation */
    ierr = DDL_calc_perm(nx, ny, nz, pv);
      break;
  case 3:
    /* rotate by rot */
    rot = 1;
    for (i=0;i<n;i++)
      pv[i] = (i+rot)%n;
    break;
  case 4:
    /* invert vector */
    for (i=0;i<n;i++)
      pv[i] = n-i-1;
    break;
  case 5:
    /* swap single pair (extremes) */
    for (i=0;i<n;i++)
      pv[i] = i;
    pv[0] = n-1;
    pv[n-1] = 0;
    break;
  default:
    /* identity permutation */
    for (i=0;i<n;i++)
      pv[i] = i;
    break;
  }      
  DDL_Calc_invperm(pv, ipv, n);
}
