      subroutine iludsol(n,b,x,nz,ia,ja,a,isym,rwork,iwork)
      integer n, nz, ia(*), ja(*), isym, iwork(*)
      double precision b(*), x(*), a(*), rwork(*)
*
*     This routine solves the system of linear equations:
*     (I + L*D**(-1))*(D + U) x = b
*     where D is a block diagonal matrix where each block is intended to
*     correspond to the diagonal entry of a cell in a multi-component
*     PDE and where L and U are the corresponding lower and upper block
*     triangular matrices.
*
*     NOTE: As the number of components per cell is assumed small (say
*     smaller than 20), the submatrices of D are stored as dense
*     matrices and systems involving D**(-1) therefore are solved 
*     using the BLAS-2 matrix-vector routine >dgemv<.
*
****DESCRIPTION
* *Usage:
*      INTEGER   n, nz, ia(nz), ja(n), isym, iwork(*)
*      DOUBLE PRECISION b(n), x(n), a(nz)
*      DOUBLE PRECISION rwork(*)
*
*      call iludsol(n, b, x, nz, ia, ja, a, isym, rwork, iwork)
*
* *Arguments:
* n      :IN       Integer.
*         Order of the Matrix.
* b      :IN       Double Precision b(n).
*         Right-hand side vector.
* x      :OUT      Double Precision x(n).
*         The solution to the system of linear equations being solved
* nz     :IN       Integer.
*         Number of Non-Zeros stored in A.
* ia     :IN       Integer ia(nz).
* ja     :IN       Integer ja(n).
* a      :IN       Double Precision a(nz).
*         These arrays should hold the matrix A in the SLAP
*         Column format.
* isym   :IN       Integer.
*         Flag to indicate symmetric storage format.
*         If isym=0, all nonzero entries of the matrix are stored.
*         If isym=1, the matrix is symmetric, and only the upper
*         or lower triangle of the matrix is stored.
*         NOTE: At present, only isym = 0 (nonsymmetric) is
*         supported
* rwork  :IN       Double Precision rwork(nc*n+nc)
*         Work vector assumed to contain the inverse block matrices of
*         D, obtained by a previous call to >formilu<. If there are nc
*         components per cell, then rwork should contain AT LEAST nc*(n+1)
*         elements. It is assumed that the inverse block matices of D
*         are stored starting at rwork(iwork(2)). The additional nc components
*         at the end are used as a temporary workspace in >ilusol<.
* iwork  :IN       Integer iwork(4+2*n)
*         Work vector assumed to contain:
*         I. Pointers to the start of:
*            1) the number of blocks (n/nc)  in iwork(1)
*            2) the starting location for D (in inverse form) (iwork(2))
*            3) the starting location for first entries of each column of
*               L stored implicitly in (ia, ja, a)
*            4) the starting location for last entries of each column of U
*               stored implicitly in (ia, ja, a).
*         II.  The first entries of each column of L when  stored
*              implicitly  in (ia, ja, a), starting at location
*              iwork(iwork(4)).
*         III. The last entries of each column of U when  stored
*              implicitly  in (ia, ja, a), starting at location
*              iwork(iwork(5)).
*
* *Other important points to note:
*
*  The calling sequence (and usage of rwork and iwork) is consistent
*  with the calling sequence expected from a preconditioner solver in the
*  SLAP package. Drivers for this precondtioner can therefore be easily
*  written for any of the itertative solvers for nonsymmetric systems.
*
*  This routine can be used as the main solver for any form of block
*  incomplete solver where the size of the blocks of D are smaller than
*  around 20 by 20 (larger than this and a sparse representation probably is
*  faster). HOWEVER, THIS ROUTINE REQUIRES THAT EACH BLOCK BE THE SAME SIZE.
*
*  This preconditioner is almost guaranteed not to work if the values for
*  the start of columns in L and the end of columns in U are not correct.
*  These positions are stored so that using A is as efficient as using
*  separate copies of L and U.
*
* *Routines called:
*  dcopy  (Level 1 BLAS)
*  dgemv  (Level 1 BLAS)
*  partriup (Sparse package support routine)
*
* *Update history
*
*  Initial version coded: 08/05/94
*  Revised:		  03/06/94
*  Author:                C. Addison
*  Address:               IASC
*                         University of Liverpool
*
*******************
*
      integer i, numblks, iptrblk, rem, div, maxblk,
     1        dblk, jl, ju, low, high, nc, ncsq, temp
      numblks = iwork(1)
      dblk =    iwork(2)
      jl =      iwork(3)
      ju =      iwork(4)
      call dcopy(n,b(1),1,x(1),1)
      div = n / numblks
      nc = div
      temp = dblk + n*nc
      ncsq = nc*nc
      rem = 0
*
*     Stage 1: Solve D(low:high,low:high) temp = x(low:high)
*           1a: x(high+1:n) = x(high+1:n)-L(high+1:n,low:high)*temp
*
      low = 1 - nc
      high = 0
      do i=0,numblks-2
        low = low + nc
        high = high + nc
	call dgemv('n',nc,nc,1.d0,rwork(i*ncsq+dblk),nc,x(low),1,
     1		   0.d0,rwork(temp),1)
        call partriup(n,rwork(temp),x,-1.d0,iwork(jl+low-1),
     1                 ia,ja,a,'l',low,high)
      end do
*
*     Stage 2: Solve the upper triangular part
*
      low = n + 1
      high = n + nc
      do i=numblks-1,0,-1
        low = low - nc
        high = high - nc
	call dgemv('n',nc,nc,1.d0,rwork(i*ncsq+dblk),nc,x(low),1,
     1		   0.d0,rwork(temp),1)
	call dcopy(nc,rwork(temp),1,x(low),1)
        if (i .gt. 0) then
           call partriup(n,rwork(temp),x,-1.d0,iwork(ju+low-1),
     1                 ia,ja,a,'u',low,high)
        endif
      end do
      end

