      subroutine formilu(n,nz,ia,ja,a,nc,
     1           d,jl,ju,list,ipvt,den_l,den_u,iprob)
      integer n, nz, ia(*), ja(*), nc, jl(*), ju(*), list(2,*), ipvt(*),
     1        iprob
      double precision a(*), d(*), den_l(*), den_u(*)
*
*     This routine forms the diagonal matrix of an ilu(0) approximation
*     to the LU factors of A where the incomplete LU factorisation is
*     performed over nc by nc sub-blocks, corresponding to there being
*     n/nc cells with nc components (equations) per cell in a discretisated
*     spatial domain.
*
*     The matrix D corresponding to the block diagonals of A is formed and
*     then updated using a conventional ilu approach - updating the diagonal
*     blocks of D so that the incomplete factorisation: (L+D)D**(-1)(U+D) has
*     the same block diagonal as the original matrix A.
*
****DESCRIPTION
* *Usage:
*      INTEGER  n, nz, ia(nz), ja(n), jl(n), ju(n), list(2,numblks), nc,
*     +          ivpt(n)
*      DOUBLE PRECISION d(nc*n), a(nz), den_l(nc*nc), den_u(nc*nc)
*
*      call formilu(n,nz,ia,ja,a,nc,d,jl,ju,list,ipvt,den_l,den_u)
*
* *Arguments:
* n      :IN       Integer.
*         Order of the Matrix.
* nz     :IN       Integer.
*         Number of Non-Zeros stored in A.
* ia     :IN       Integer ia(nz).
* ja     :IN       Integer ja(n).
* a      :IN       Double Precision a(nz).
*         These arrays should hold the matrix A in the SLAP
*         Column format.
* nc     :IN       Integer
*         The number of components in each cell, n MUST be divisible
*         by nc
* d      :OUT      Double Precision d(n*nc)
*         Contains the factored form of the ILU(0) updated block diagonal
*         matrix
* jl     :OUT      Integer jl(n)
*         Contains the first entries of each column of the lower triangular
*         matrix L stored implicitly in (ia, ja, a)
* ju     :OUT      Integer ju(n)
*         Contains the last entries of each column of the upper triangular
*         matrix U stored implicitly in (ia, ja, a)
* list   :WORK     Integer list(2,n/nc)
*         Workspace used internally as a linked list for possible sub-blocks
*         L(j,k) to use in forming D(j,j) - L(j,k) D(k,k)**(-1) U(k,j)
*         (Both L(j,k) and U(k,j) must be non-zero to bother.)
* ipvt   :OUT      Integer ipvt(n)
*         Contains pivot information used subsequently in the solve step.
* den_l  :WORK     Double Precision den_l(nc,nc)
*         Temporary matrix in which to store a dense copy of L(j,k).
* den_u  :WORK     Double Precision den_u(nc,nc)
*         Temporary matrix in which to store a dense copy of U(k,j).
* iprob  :OUT      Integer
*         iprob = 0 normally, iprob > 0 indicates the number of blocks
*         which produced a singular LU factorisation. These were fudged by
*         setting the diagonal zero value to 1. Performance is almost
*         certainly to be degraded if iprob is positive.
*
* *Routines called:
*  dfill: Level 1 BLAS-like
*  dgetrf: LAPACK matrix factorisation
*  dgetrs: LAPACK matrix solve
*  dgemm: Level 3 BLAS
*
* *Update history
*     Initial version
*     10/04/94
*
*     Revised to eliminate explicit forming of L and U (and make
*     searches slightly shorter)
*     08/05/94
*
*     C. Addison
*     IASC
*******************
*
      parameter (zero = 0.d0)
      integer i, j, k, rem, div,
     1        mylow, myhigh, num_d, numblks, ist, row,
     2        first_row, last_row, floor, col_low, col_high, row_low,
     3        row_high, col, head, free, start, ptr, old_ptr,
     4        ncsq, offset, update, info, low, high
      logical found
      logical first
*
*     As the block size is the number of components per cell, the problem
*     must divide evenly into n/nc blocks or else something is wrong!.
*
      rem = 0
      iprob = 0
      numblks = n / nc
      div = nc
      ncsq = nc * nc
      call dfill(n*nc,d,zero)
*
      num_d = 0
      kold = -1
      do j = 1,n
         k = (j-1)/nc
         if (k .ne. kold) then
            mylow =  k*nc
            myhigh = mylow  + nc
*           write(*,*) '>>> ',k,mylow,myhigh,j
            kold = k
            offset = nc * mylow - mylow
         end if
*
*        Special case for ja(j) - diagonal entries
*
         col= (j-1-mylow)*nc
         row = j + offset
         num_d = num_d + 1
         d(row+col) = a(ja(j))
         low = n+1
*
*        Rather than store L and U explicitly, we only store the
*        start point for each column of L within A and the end point
*        for each column of U. Of course sometimes, these are empty.
*
         ju(j) = 0
         jl(j) = nz+1
         high = 0
         do i = ja(j)+1, ja(j+1)-1
            if (ia(i) .gt. mylow .and.
     1          ia(i) .le. myhigh) then
              num_d = num_d + 1
              row = ia(i) + offset
              d(row+col) = a(i)
            else if (ia(i) .le. mylow .and. ia(i) .gt. high) then
              high = ia(i)
              ju(j) = i
            else if (ia(i) .gt. myhigh .and. ia(i) .lt. low) then
              low = ia(i)
              jl(j) = i
            end if
         end do
      end do
*
*     Initial D, L and U formed; now start to update D properly
*
      do k = 0, numblks-2
         floor = k*nc +1
         call dgetrf(nc,nc,d(k*ncsq+1),nc,ipvt,info)
         if (info .gt. 0) then
            iprob = iprob + 1
            do j=(info-1)*(nc+1)+1,ncsq,nc+1
               d(k*ncsq+j) = 1.d0
            end do
         end if
*
*        Modification to produce explicit D^{-1}
*
         call dfill(ncsq,den_l,zero)
         do j=1,ncsq,nc+1
            den_l(j) = 1.d0
         end do
         call dgetrs('n',nc,nc,d(k*ncsq+1),nc,
     1                 ipvt,den_l,nc,info)
         call dcopy(ncsq,den_l,1,d(k*ncsq+1),1)
*
         head = 1
         free = 2
         list(1,head) = 0
         list(2,head) = -1
*
*        Range of row values to trap for
*
         row_low = floor + nc
         row_high = n+1
*
*        Scan down lower triangular L matrix; form a linked list of possible
*        block column starting positions with which to scan U.
*
         do j = floor,(k+1)*nc
            first_row = ja(j)+1
            last_row = ja(j+1) - 1
*
*           Set ist to the first appropriate element of L (found earlier)
*
            ist = jl(j)
            if (ist .ne. nz+1) then
               do i = ist, last_row
                  start = ((ia(i)-1)/nc)*nc + 1
                  ptr = head
                  do while(start .gt. list(1,ptr) .and.
     1                                   list(2,ptr) .ne. -1)
                     old_ptr = ptr
                     ptr = list(2,ptr)
                  end do
                  if (list(2,ptr) .eq. -1 .and.
     1                              start .gt. list(1,ptr)) then
                     list(2,ptr) = free
                     list(1,free) = start
                     list(2,free) = -1
                     free = free + 1
                  else if(list(1,ptr) .gt. start) then
                     list(1,free) = start
                     list(2,free) = ptr
                     list(2,old_ptr) = free
                     free = free + 1
                  endif
               enddo
            endif
         end do
*
*        Scan the upper triangular portion of A to see which blocks across
*        rows match with lower triangular blocks down columns and then
*        update the appropriate portion of D accordingly. (Sounds simple
*        doesn't it!
*
         ptr =  list(2,head)
         do while(ptr .ne. -1)
            col_low = list(1,ptr)
            col_high = col_low + nc
            row_low = floor
            row_high = floor + nc
            found = .false.
            col = col_low
            do while(.not. found .and. col .lt. col_high)
               first_row = ja(col)+1
               last_row = ja(col+1)-1
               ist = first_row
               do while(ia(ist) .lt. row_low .and. ist .lt. last_row)
                  ist = ist + 1
               end do
               if (ia(ist) .lt. row_high .and. ia(ist) .ge. row_low
     1                                   .and. ist .le. last_row) then
                  found = .true.
                  call dfill(ncsq,den_u,zero)
*
*                 Fill non-zeros of den_u; then matrix multiply for den_l
*
                  row_low = list(1,ptr)
                  row_high = row_low + nc
                  do j = 0,nc-1
                     col = floor + j
                     first_row = jl(col)
                     last_row = ja(col+1)-1
                     offset = j*nc - row_low + 1
                     do i = first_row,last_row
                        if (ia(i) .ge. row_low .and.
     1                      ia(i) .lt. row_high) then
                           den_u(ia(i)+offset) = a(i)
                        end if
                     end do
                  end do
                  call dgemm('n','n',nc,nc,nc,1.d0,den_u,nc,
     1                     d(k*ncsq+1),nc,0.d0,den_l,nc)
*
*                 den_u:
*
                  call dfill(ncsq,den_u,zero)
                  row_low = floor
                  row_high = floor + nc
                  do j = 0,nc-1
                     col = col_low + j
                     first_row = ja(col)+1
*                    last_row = ja(col+1)-1
                     last_row = ju(col)
                     offset = j*nc - row_low + 1
                     do i =first_row,last_row
                        if (ia(i) .ge. row_low .and.
     1                      ia(i) .lt. row_high) then
                           den_u(ia(i)+offset) = a(i)
                        end if
                     end do
                  end do
*
*                 Update the appropriate block of d
*
                  update = (col_low-1)/nc
                  call dgemm('n','n',nc,nc,nc,-1.d0,den_l,nc,den_u,nc,
     1                       1.d0, d(update*ncsq+1),nc)
               else
                  col= col+1
               endif
            enddo
            ptr = list(2,ptr)
         end do
      end do
*
*     All blocks have been processed, now factor the last block
*
      k = numblks - 1
      call dgetrf(nc,nc,d(k*ncsq+1),nc,ipvt,info)
      if (info .gt. 0) then
         iprob = iprob + 1
         do j=(info-1)*(nc+1)+1,ncsq,nc+1
            d(k*ncsq+j) = 1.d0
         end do
      end if
*
*     Modification to produce explicit D^{-1}
*
      call dfill(ncsq,den_l,zero)
      do j=1,ncsq,nc+1
        den_l(j) = 1.d0
      end do
*
      call dgetrs('n',nc,nc,d(k*ncsq+1),nc,
     1            ipvt,den_l,nc,info)
      call dcopy(ncsq,den_l,1,d(k*ncsq+1),1)
*
      end
